{
This is part of AY Emulator project
AY-3-8910/12 Emulator
Version 2.9 for Windows and Linux
Author Sergey Vladimirovich Bulba
(c)1999-2016 S.V.Bulba
}

unit MainWin;

{$mode objfpc}{$H+}
{$ASMMODE intel}

interface

uses
  LCLIntf, LCLType,
  {$IFDEF Windows}
   Windows, MMSystem, WaveOutAPI,
  {$ELSE Windows}
   LMessages,
  {$ENDIF Windows}
  Types, SysUtils, FileUtil, Classes, Graphics, Controls, Forms, Dialogs, About,
  LH5, UniReader, AY, WinVersion, lazutf8, ExtCtrls;

const

//Version related constants
 VersionString = '2.9';
 VersionMajor = 2;
 VersionMinor = 9;
 CompilYs = '2016';
 CompilY = 2016;
 CompilM = 4;
 CompilD = 8;

{$ifdef beta}
BetaNumber = 'beta 16';
{$endif beta}

//Default mixer parameters
 SampleRateDef  = 48000;
 SampleBitDef   = 16;
 FrqZ80Def      = 3494400;
 AY_FreqDef     = 1773400;
 IntOffsetDef   = 0;
 BeeperMaxDef   = 128;
 MaxTStatesDef  = 69888;
 Interrupt_FreqDef = 50000;
 Index_ALDef    = 255;
 Index_ARDef    = 13;
 Index_BLDef    = 170;
 Index_BRDef    = 170;
 Index_CLDef    = 13;
 Index_CRDef    = 255;
 NumOfChanDef   = 2;
 ChanModeDef    = 1;
 MFPTimerModeDef = 0;
 MFPTimerFrqDef = 2457600;

 NumberOfBuffersDef = 3;
 BufLen_msDef = 200;

 //todo: select default CP in Linux due system language settings
 //todo: Ansi for Vortex Tracker, OEM for other trackers?
 CodePageDef:string = {$IFDEF Windows}'Ansi'{$ELSE Windows}'CP1251'{$ENDIF Windows};

//User defined windows messages
 WM_LINEPARAM      = WM_USER;
 WM_PLAYNEXTITEM   = WM_USER + 1;
 WM_PLAYERROR      = WM_USER + 2;
 WM_FINALIZEWO     = WM_USER + 5;
 WM_HIDEMINIMIZE   = WM_USER + 6;
 WM_GETTIMELENGTH  = WM_USER + 8;
 WM_BASSMETADATA   = WM_USER + 9;

//Metrics of some controls
//Main window
 MWWidth = 358;
 MWHeight = 123;

 //Spectrum analizer
 spa_num = 91 - 26 - 2;
 spa_width = spa_num + 2; spa_height = 20;
 spa_x = 26; spa_y = 34;

 //Amplitude analizer
 amp_width = 17; amp_height = 15;
 amp_x = 50; amp_y = 18;

 max_width2 = spa_width; max_height2 = spa_height;

 //Scrolling title
 scr_lineheight = 24;
 scr_x = 108;scr_y = 48;
 scr_width = 197; scr_height = scr_lineheight;

 //Time label
 time_x = 24; time_y = 65;
 time_width = 93-24;time_height = 20;

 max_height = scr_height;

 //Offsets of background bitmaps for controls
 spa_src = 0;
 amp_src = spa_width;
 time_src = spa_width + amp_width;
 scr_src = spa_width + amp_width + time_width;
 max_src = scr_src + scr_width;

//Skin 2.0 identificator
SkinId:string = 'Ay_Emul 2.0 Skin File'#13#10#26;
SkinIdLen = 24;

VTPath:string = 'VT.exe';

Zero:integer = 0; //:-)

type

//Spectrum analizer values
 TSpa = array[0..spa_num - 1] of integer;
 PSpa = ^TSpa;

//Own sens object
  PSensZone = ^TSensZone;
  TSensZone = class(TObject)
  constructor Create(ps:PSensZone;x,y,w,h:integer;pr:TNotifyEvent);
  function Touche(x,y:integer):boolean;
  public
  Next:PSensZone;
  zx,zy,zw,zh:integer;
  Clicked:boolean;
  Action:TNotifyEvent;
  end;

//Own button object
  PButtZone = ^TButtZone;
  TButtZone = class(TObject)
  constructor Create(ps:PButtZone;x,y,w,h,rh:integer;Bmp:TBitmap;x1,y1,x2,y2:integer;pr:TNotifyEvent);
  procedure Free;
  function Touche(x,y:integer):boolean;
  procedure Push;
  procedure UnPush;
  procedure Switch_On;
  procedure Switch_Off;
  procedure Redraw(OnCanvas:boolean);
  public
  Next:PButtZone;
  zx,zy,zw,zh,RgnHandle:integer;
  Clicked:integer;
  Is_On,Is_Pushed:boolean;
  Bmp1,Bmp2:TBitmap;
  Action:TNotifyEvent;
  end;

//Own led object
  PLedZone = ^TLedZone;
  TLedZone = class(TObject)
  constructor Create(ps:PLedZone;x,y,w,h:integer;Bmp:TBitmap;x1,y1,x2,y2:integer);
  procedure Free;
  procedure Redraw(OnCanvas:boolean);
  public
  Next:PLedZone;
  zx,zy,zw,zh:integer;
  State:boolean;
  Bmp1,Bmp2:TBitmap;
  end;

//Own mouse moving object
  PMoveZone = ^TMoveZone;
  TMoveZone = class(TObject)
  constructor Create(ps:PMoveZone;x,y,w,h,y1,h1,rh:integer;pr:TNotifyEvent);
  procedure Free;
  function Touche(x,y:integer):boolean;
  function ToucheBut(x,y:integer):boolean;
  procedure AddBitmaps(Bmp:TBitmap;x1,y1,bw,bh:integer;m:boolean);
  procedure Redraw(OnCanvas:boolean);
  procedure HideBmp;
  public
  Next:PMoveZone;
  zx,zy,zw,zh,zy1,zh1,Delt,OldX,OldY,PosX,PosY,RgnHandle,
  bm1h,bm1w:integer;
  Clicked,State,Bmps:boolean;
  Bmp1,Bmp2:TBitmap;
  Action:TNotifyEvent;
  end;

 FIDO_Status = (FIDO_Nothing,FIDO_Playing,FIDO_Exit);

//Main window form

  { TForm1 }

  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    TrayIcon1: TTrayIcon;
    procedure ButOpenClick(Sender: TObject);
    procedure DoMovingWindow(Sender: TObject);
    procedure DoMovingVol(Sender: TObject);
    procedure DoMovingProgr(Sender: TObject);
    procedure DoMovingScroll(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormDeactivate(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure FormPaint(Sender: TObject);
    {$IFNDEF Windows}
    procedure DoSetRgn(Sender: TObject);
    {$ENDIF Windows}
    procedure FormShow(Sender: TObject);
    procedure PlayClick(Sender: TObject);
    procedure SetDefault;
    procedure ButPauseClick(Sender: TObject);
    procedure ButStopClick(Sender: TObject);
    procedure CommandLineInterpreter(CL:string;Start:boolean);
    procedure TrayIcon1DblClick(Sender: TObject);
    procedure TrayIcon1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TrayIcon1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure WMPLAYNEXTITEM(var Msg: TMsg);message WM_PLAYNEXTITEM;
    procedure WMBASSMETADATA(var Msg: TMsg);message WM_BASSMETADATA;
    procedure WMPLAYERROR(var Msg: TMsg);message WM_PLAYERROR;
{$IFDEF Windows}
    procedure WMLINEPARAM(var Msg: TMsg);message WM_LINEPARAM;
    procedure WMFINALIZEWO(var Msg: TMsg);message WM_FINALIZEWO;
{$ENDIF Windows}
    procedure HideMinimize(var Msg: TMsg);message WM_HIDEMINIMIZE;
    procedure DoMinimize;
    procedure DoRestore;
    procedure DoVisualisation;
    procedure MessageSkipper;
    procedure NewMessageSkipper;
    procedure SwapLan;
    procedure Set_Chip_Frq(Fr:integer);
    procedure Set_Player_Frq(Fr:integer);
    procedure ButMixerClick(Sender: TObject);
    procedure ButMinClick(Sender: TObject);
    procedure ButCloseClick(Sender: TObject);
    procedure ButAboutClick(Sender: TObject);
    procedure ButAmpClick(Sender: TObject);
    procedure ButTimeClick(Sender: TObject);
    procedure ButSpaClick(Sender: TObject);
    procedure ShowAllParams;
    procedure RestoreAllParams;
    procedure ButListClick(Sender: TObject);
    procedure ButNextClick(Sender: TObject);
    procedure ButPrevClick(Sender: TObject);
    procedure Set_Mode(It:Integer);
    procedure Set_Mode_Manual(AL,AR,BL,BR,CL,CR:byte);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ButToolsClick(Sender: TObject);
    procedure ButLoopClick(Sender: TObject);
    procedure Set_Z80_Frq(NewF:integer);
    procedure Set_N_Tact(NewF:integer);
    procedure CommandLineAndRegCheck;
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure AppRestore(Sender: TObject);
    procedure AppMinimize(Sender: TObject);
    procedure AppEndSession(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure RemoveTrayIcon;
    procedure AddTrayIcon;
    function LoadSkin(FName:string;First:boolean):boolean;
    procedure SetMainBmp(p:pointer;size:integer);
    procedure BmpFree;
    procedure CopyBmpSources;
    procedure Set_MFP_Frq(Md,Fr:integer);
    procedure ShowApp(Tray:boolean);
    procedure FIDO_SaveStatus(Status:FIDO_Status);
    procedure JumpToTime;
    procedure CallHelp;
    procedure VolUp;
    procedure VolDown;
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure SaveParams;
    procedure SetVisTimerPeriod(VTP:integer);
    procedure Set_Sample_Rate2(SR:integer);
    procedure Set_Sample_Bit2(SB:integer);
    procedure Set_Stereo2(St:integer);
    procedure SetBuffers(len,num:integer);
{$IFDEF Windows}
    procedure Set_WODevice2(WOD:integer;NM:string);
    procedure Set_MIDIDevice2(MD:integer;NM:string);
{$ENDIF Windows}
    procedure Set_BufLen_ms2(BL:integer);
    procedure Set_NumberOfBuffers2(NB:integer);
    procedure Set_Chip2(Ch:ChTypes);
    procedure Set_Z80_Frq2(NewF:integer);
    procedure Set_Chip_Frq2(Fr:integer);
    procedure Set_Player_Frq2(Fr:integer);
    procedure Set_IntOffset2(InO:integer);
    procedure Set_N_Tact2(NT:integer);
    procedure Set_N_TactS(t:string);
    procedure Set_Language2(Rus:boolean);
    procedure Set_Loop2(Lp:boolean);
    procedure Set_TrayMode2(TM:integer);
    procedure Set_MFP_Frq2(Md,Fr:integer);
    procedure SetAutoSaveDefDir2(ASD:boolean);
    procedure SetAutoSaveWindowsPos2(ASW:boolean);
    procedure SetAutoSaveVolumePos2(ASV:boolean);
    procedure SetChan2(u,i:integer);
    procedure SetFilter(FQ:integer);
    procedure SetFilter2(FQ:integer);
    procedure CalcFiltKoefs;
    procedure SaveAllParams;
    procedure DoCloseActions;
    procedure SelectTrayIcon(n:integer);
    procedure SelectAppIcon(n:integer);
    {$IFDEF Windows}
    procedure SetPriority2(NP:integer);
    {$ENDIF Windows}
    procedure VisTimerEvent(Sender: TObject);
    procedure Ay_Emul_ShowExceptionA(Sender : TObject; E : Exception);
  private
    { Private declarations }
  public
    { Public declarations }

  SkinFileName,SkinAuthor,SkinComment:string;

  DefaultDirectory,SkinDirectory:string;

  LastTimeComLine:DWORD;
  end;

function IntelWord(Wrd:word):word;
function IntelDWord(DWrd:dword):dword;
function DivMul(q1,q2,q3:int64):dword;

procedure PlayCurrent;
procedure StopAndFreeAll;
procedure StopPlaying;
procedure RestoreControls;

procedure AYVisualisation(smp:DWORD);

procedure RedrawVisChannels(ca,cb,cc,mh:integer);
procedure RedrawVisSpectrum(CP:PVisPoint;MaxVal:integer);

procedure ShowProgress(a1:integer);

procedure Set_Sample_Rate(SR:integer);
procedure Set_Sample_Bit(SB:integer);
procedure Set_Stereo(St:integer);

procedure Rewind(newpos,maxpos:integer);

procedure SetScrollString(const scrstr:string);
procedure ReprepareScroll;

{$IFDEF Windows}
procedure GetSysVolume;
{$ENDIF Windows}
procedure SetSysVolume;
procedure RedrawVolume;

procedure Ay_Emul_ShowException (Msg : ShortString);

procedure SetCommandLine(clstr:string);

procedure AdjustFormOnDesktop(Frm:TForm);

var
  Form1: TForm1;

  MFPTimerFrq,MFPTimerMode:integer;

  VisTimer:TTimer;
  VisTimerPeriod:integer = 30;

  Spa_points:array[0..spa_num] of integer;
  Spa_piks,Spa_prev:TSpa;
  PSpa_Piks,PSpa_prev:PSpa;

  Scr_Left:boolean = False;
  ScrFlg:boolean = True;
  Scr_Pause:integer = 1;
  Scroll_Distination:integer = -1;
  Item_Displayed:integer = -1;
  HorScrl_Offset:integer = 0;
  Scroll_Offset:integer = scr_lineheight;
  ClearTimeInd:boolean;
  TimeMode:integer = 0;
  CurrTime_Rasch:integer;
  BaseSample:DWORD;

  BMP_Sources,BMP_Time,BMP_Vis:TBitmap;
  BMP_VScroll,BMP_Scroll:TBitmap;
  
  BMP_DBuffer:TBitmap;

  VProgrPos:integer;
  ProgrMax,ProgrPos:longword;
  ProgrWidth:word;

  IntOffset:integer;
  OUTZXAYConv_TotalTime:integer;

  FrqZ80,Interrupt_Freq,AY_Freq:integer;

  Russian_Interface:boolean = False;
  PSG_Skip:word;
  Real_End_All,May_Quit:boolean;
  Real_End:array[0..1] of boolean;
  Do_Scroll:boolean = True;
  Do_Loop:boolean = False;
  Time_ms:integer = 0;
  TimeShown:integer = -MaxInt;

  ButPlay,ButNext,ButPrev,ButOpen,ButStop,ButPause,ButAbout,
  ButLoop,ButMixer,ButTools,ButList,ButMinimize,ButClose:TButtZone;
  SensSpa,SensAmp,SensTime:TSensZone;
  MoveWin,MoveVol,MoveProgr,MoveScr:TMoveZone;
  Led_AY,Led_YM,Led_Stereo:TLedZone;
  MyFormRgn,RgnClose,RgnMin,RgnMixer,RgnTools,RgnPList,
  RgnLoop,RgnBack,RgnPlay,RgnNext,RgnStop,RgnPause,RgnOpen,
  RgnVol,RgnProgr:HRGN;

  IndicatorChecked:boolean = True;
  SpectrumChecked:boolean = True;
  AutoSaveDefDir:boolean = True;

  //Tray Icon Data
  TrayMode:integer = 0;
  TrayIconNumber:integer = 11;
  TrayIconClicked:boolean = False;

  RecurseDirs:boolean = True;
  RecurseDirsDoDetect:boolean = False;

  AppIsModal:boolean = False;

  MenuIconNumber:integer = 0;
  AppIconNumber:integer = 0;
  MusIconNumber:integer = 10;
  SkinIconNumber:integer = 3;
  ListIconNumber:integer = 2;
  BASSIconNumber:integer = 4;

  FIDO_Descriptor_Enabled:boolean = False;
  FIDO_Descriptor_WinEnc:boolean = False;
  FIDO_Descriptor_KillOnNothing:boolean = False;
  FIDO_Descriptor_KillOnExit:boolean = True;
  FIDO_Descriptor_Prefix:string = '... Ay_Emul: ';
  FIDO_Descriptor_Suffix:string = '';
  FIDO_Descriptor_Nothing:string = 'no song playing...';
  FIDO_Descriptor_Filename:string;
  FIDO_Descriptor_String:string = '';

  ToolsX:integer = MaxInt;
  ToolsY:integer;
  AutoSaveWindowsPos:boolean = True;
  AutoSaveVolumePos:boolean = {$IFDEF Windows}False{$ELSE Windows}True{$ENDIF Windows};
  Uninstall:boolean = False;

const
  ButtZoneRoot:PButtZone = nil;
  SensZoneRoot:PSensZone = nil;
  MoveZoneRoot:PMoveZone = nil;
  LedZoneRoot:PLedZone = nil;
  CLFast = 800;
  InitialScan:boolean = False;

var
  AfterScan:array of string;
  VolumeCtrl,VolumeCtrlMax:integer;
  VolLinear:boolean = False;

 IsPlaying:boolean = False;
 Paused:boolean;
 NumberOfChannels,SampleRate,SampleBit:integer;
 NumberOfBuffers,BufferLength,BuffLen,BufLen_ms:integer;
 NOfTicks:DWORD;
 {$IFNDEF Windows}
 Timer1:TTimer; //для обхода глюков GTK
 {$ENDIF Windows}

 
implementation

uses Mixer, PlayList, Tools, Z80, JmpTime, Players, Options,
     lightBASS, BASScode, BassTags
     {$IFDEF Windows}
     , ProgBox, CDviaMCI, Midi
     {$ENDIF Windows}
     , Convs, Languages;

{$R *.lfm}

var
  CloseActionsDone:boolean = False;

  sw:integer = scr_width;
  sj,sw1,sj1,sw2,sj2:integer;
  ss,ss1,ss2:string;

  {$IFDEF Windows}
  PrevWndProc: WNDPROC;
  PrevAWndProc: WNDPROC;
  {$ENDIF Windows}

function IntelWord(Wrd:word):word;
begin
asm
xchg al,ah
end;
end;

function IntelDWord(DWrd:dword):dword;
begin
asm
xchg al,ah
ror eax,16
xchg al,ah
end;
end;

function DivMul(q1,q2,q3:int64):dword;
begin
Result := q1 * q2 div q3;
end;

procedure GetStringWnJ(const s:string; var w,j:integer);
begin
w := BMP_VScroll.Canvas.TextWidth(s);
j := 0;
if scr_width > w then
 j := (scr_width - w) div 2;
end;

procedure RedrawScroll;
begin
BMP_Scroll.Canvas.CopyMode:=cmSrcCopy;
BMP_Scroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_height),BMP_Sources.Canvas,Bounds(scr_src,0,scr_width,scr_height));
BMP_VScroll.Canvas.TextOut(-HorScrl_Offset + sj,scr_height,ss);
BMP_Scroll.Canvas.CopyMode:=cmSrcAnd;
BMP_Scroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_height),BMP_VScroll.Canvas,Bounds(0,scr_height,scr_width,scr_height));
Form1.Canvas.CopyMode:=cmSrcCopy;
Form1.Canvas.CopyRect(Bounds(scr_x,scr_y,scr_width,scr_height),BMP_Scroll.Canvas,Rect(0,0,scr_width,scr_height));
end;

procedure RedrawTime;
var
 CurTimeJ:integer;
 CurrTimeStr:string;
 sig:string;
begin
if TimeMode = 1 then sig := '-' else sig := '';
CurrTimeStr := sig + TimeSToStr(abs(TimeShown));
CurTimeJ := time_width - BMP_Time.Canvas.TextWidth(CurrTimeStr);
if CurTimeJ > 0 then CurTimeJ := CurTimeJ div 2;
BMP_Time.Canvas.CopyMode:=cmSrcCopy;
BMP_Time.Canvas.CopyRect(Rect(0,0,time_width,time_height),BMP_Sources.Canvas,Bounds(time_src,0,time_width,time_height));
BMP_Time.Canvas.TextOut(CurTimeJ,0,CurrTimeStr);
Form1.Canvas.CopyMode:=cmSrcCopy;
Form1.Canvas.CopyRect(Bounds(time_x,time_y,time_width,time_height),BMP_Time.Canvas,Rect(0,0,time_width,time_height));
end;

procedure CalculateSpectrumPoints;
var
 i:integer;
begin
Spa_points[0] := $FFF;
for i := 1 to spa_num do
 Spa_points[i] := round($FFF * exp(-ln(16 * 22050 * $FFF/AY_Freq)*i/spa_num))
end;

procedure TForm1.DoVisualisation;
var
 Y_Stp:integer;
 Points_To_Scroll:integer;
 Temp,Temp1:integer;
begin
 begin
{$IFDEF Windows}
 WOVisualisation;
{$ENDIF Windows}
 BASSVisualisation;
{$IFDEF Windows}
 CDVisualisation;
 MIDIVisualisation;
{$ENDIF Windows}
 if ClearTimeInd then
  begin
   BMP_Time.Canvas.CopyMode:=cmSrcCopy;
   BMP_Time.Canvas.CopyRect(Rect(0,0,time_width,time_height),BMP_Sources.Canvas,Bounds(time_src,0,time_width,time_height));
   Canvas.CopyMode:=cmSrcCopy;
   Canvas.CopyRect(Bounds(time_x,time_y,time_width,time_height),BMP_Time.Canvas,Rect(0,0,time_width,time_height));
   TimeShown := -MaxInt;
   ClearTimeInd := False;
  end;
 if Time_ms <> 0 then
  begin
   case TimeMode of
   0: Temp := round(CurrTime_Rasch / 1000);
   1:
    begin
     Temp := Time_ms - CurrTime_Rasch;
     if Temp < 0 then Temp := 0;
     Temp := -round(Temp / 1000);
    end;
   else
    begin
     Temp := round(Time_ms / 1000);
     if Temp < 0 then Temp := 0;
    end;
   end;
   if Temp <> TimeShown then
    begin
     TimeShown := Temp;
     RedrawTime;
    end;
  end;
 Temp := Item_Displayed;
 Temp1 := Scroll_Distination;
 if Abs(Temp1 - Temp) > 16 then
  begin
   if Temp1 > Temp then
    Temp := Temp1 - 16
   else
    Temp := Temp1 + 16;
   Item_Displayed := Temp;
   ss := GetPlayListString(PlaylistItems[Temp]);
   GetStringWnJ(ss,sw,sj);
   BMP_VScroll.Canvas.TextOut(sj,scr_lineheight,ss);
  end;
 Points_To_Scroll := scr_lineheight*(Temp1 - Temp + 1) - Scroll_Offset;
 if Points_To_Scroll <> 0 then
  begin
   ScrFlg := False;
   Y_Stp := (Abs(Points_To_Scroll) - 1) div scr_lineheight + 1;
   if Y_Stp >= scr_lineheight then Y_Stp := scr_lineheight - 1;
   if Points_To_Scroll > 0 then
    begin
     if (Scroll_Offset >= scr_lineheight) then
      begin
       BMP_VScroll.Canvas.FillRect(0,scr_lineheight*2,scr_width,scr_lineheight*3);
       if Temp + 1 < Length(PlaylistItems) then
        begin
         ss2 := GetPlayListString(PlaylistItems[Temp + 1]);
         GetStringWnJ(ss2,sw2,sj2);
         BMP_VScroll.Canvas.TextOut(sj2,scr_lineheight*2,ss2);
        end
      end;
     Inc(Scroll_Offset,Y_Stp);
     if Scroll_Offset >= 2*scr_lineheight then
      begin
       HorScrl_Offset := 0;
       ss := ss2; sw := sw2; sj := sj2;
       BMP_VScroll.Canvas.CopyMode:=cmSrcCopy;
       BMP_VScroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_lineheight*2),BMP_VScroll.Canvas,Bounds(0,scr_lineheight,scr_width,scr_lineheight*2));
       Dec(Scroll_Offset,scr_lineheight);
       Inc(Temp);
       Item_Displayed := Temp;
      end;
    end
   else
    begin
     if (Scroll_Offset <= scr_lineheight) then
      begin
       BMP_VScroll.Canvas.FillRect(0,0,scr_width,scr_lineheight);
       if Temp - 1 >= 0 then
        begin
         ss1 := GetPlayListString(PlaylistItems[Temp - 1]);
         GetStringWnJ(ss1,sw1,sj1);
         BMP_VScroll.Canvas.TextOut(sj1,0,ss1);
        end
      end;
     Dec(Scroll_Offset,Y_Stp);
     if Scroll_Offset <= 0 then
      begin
       HorScrl_Offset := 0;
       ss := ss1; sw := sw1; sj := sj1;
       BMP_VScroll.Canvas.CopyMode:=cmSrcCopy;
       BMP_VScroll.Canvas.CopyRect(Bounds(0,scr_lineheight,scr_width,scr_lineheight*2),BMP_VScroll.Canvas,Rect(0,0,scr_width,scr_lineheight*2));
       Inc(Scroll_Offset,scr_lineheight);
       Dec(Temp);
       Item_Displayed := Temp;
      end;
    end;
   BMP_Scroll.Canvas.CopyMode:=cmSrcCopy;
   BMP_Scroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_height),BMP_Sources.Canvas,Bounds(scr_src,0,scr_width,scr_height));
   BMP_Scroll.Canvas.CopyMode:=cmSrcAnd;
   BMP_Scroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_height),BMP_VScroll.Canvas,Bounds(0,Scroll_Offset,scr_width,scr_height));
   Canvas.CopyMode:=cmSrcCopy;
   Canvas.CopyRect(Bounds(scr_x,scr_y,scr_width,scr_height),BMP_Scroll.Canvas,Rect(0,0,scr_width,scr_height));
  end;
 if Do_Scroll and ScrFlg and (sw > scr_width) and
    not MoveScr.Clicked then
  begin
   Dec(Scr_Pause);
   if Scr_Pause = 0 then
    begin
     Inc(Scr_Pause);
     if Scr_Left then
      begin
       Dec(HorScrl_Offset);
       if HorScrl_Offset < 0 then
        begin
         Scr_Left := False;
         HorScrl_Offset := 0;
         Scr_Pause := 50;
        end
       else
        RedrawScroll;
      end
     else
      begin
       Inc(HorScrl_Offset);
       if HorScrl_Offset > sw - scr_width then
        begin
         Scr_Left := True;
         HorScrl_Offset := sw - scr_width;
         Scr_Pause := 50;
        end
       else
        RedrawScroll;
      end;
    end;
  end
 else
  ScrFlg := True;
 end;
end;

procedure TForm1.VisTimerEvent(Sender: TObject);
begin
if WindowState <> wsMinimized then DoVisualisation;
end;

procedure RedrawVisChannels(ca,cb,cc,mh:integer);
begin
if IndicatorChecked then
 begin
  BMP_Vis.Canvas.CopyMode:=cmSrcCopy;
  BMP_Vis.Canvas.CopyRect(Rect(0,0,amp_width,amp_height),BMP_Sources.Canvas,Bounds(amp_src,0,amp_width,amp_height));
  if ca > 0 then
   begin
    BMP_Vis.Canvas.MoveTo(1,amp_height);
    BMP_Vis.Canvas.LineTo(1,amp_height + 1 - ca * amp_height div mh);
   end;
  if cb > 0 then
   begin
    BMP_Vis.Canvas.MoveTo(8,amp_height);
    BMP_Vis.Canvas.LineTo(8,amp_height + 1 - cb * amp_height div mh);
   end;
  if cc > 0 then
   begin
    BMP_Vis.Canvas.MoveTo(15,amp_height);
    BMP_Vis.Canvas.LineTo(15,amp_height + 1 - cc * amp_height div mh);
   end;
  Form1.Canvas.CopyMode:=cmSrcCopy;
  Form1.Canvas.CopyRect(Bounds(amp_x,amp_y,amp_width,amp_height),BMP_Vis.Canvas,Rect(0,0,amp_width,amp_height));
 end;
end;

procedure RedrawVisSpectrum(CP:PVisPoint;MaxVal:integer);
var
 p:pointer;
 i,j,n:integer;
begin
if SpectrumChecked then
 begin
  p := PSpa_prev;
  PSpa_prev := PSpa_piks;
  PSpa_piks := p;
  if CP <> nil then
   begin
    FillChar(PSpa_piks^,SizeOf(TSpa),0);
    for n := 0 to 1 do
     begin
      for i := 0 to spa_num - 1 do
       begin
        if (CP^.R[n].TnA > Spa_Points[i + 1]) and (CP^.R[n].TnA <= Spa_Points[i]) then
         if PSpa_piks^[i] < CP^.R[n].AmpA then
          PSpa_piks^[i] := CP^.R[n].AmpA;
        if (CP^.R[n].TnB > Spa_Points[i + 1]) and (CP^.R[n].TnB <= Spa_Points[i]) then
         if PSpa_piks^[i] < CP^.R[n].AmpB then
          PSpa_piks^[i] := CP^.R[n].AmpB;
        if (CP^.R[n].TnC > Spa_Points[i + 1]) and (CP^.R[n].TnC <= Spa_Points[i]) then
         if PSpa_piks^[i] < CP^.R[n].AmpC then
          PSpa_piks^[i] := CP^.R[n].AmpC;
       end;
      if not TSMode then break;
     end;
   end;
  BMP_Vis.Canvas.CopyMode:=cmSrcCopy;
  BMP_Vis.Canvas.CopyRect(Rect(0,0,spa_width,spa_height),BMP_Sources.Canvas,Bounds(spa_src,0,spa_width,spa_height));
  for i := 0 to spa_num - 1 do
   begin
    if PSpa_Piks^[i] > 0 then
     begin
      BMP_Vis.Canvas.MoveTo(i + 1,spa_height);
      BMP_Vis.Canvas.LineTo(i + 1,(MaxVal - PSpa_Piks^[i])*spa_height div MaxVal);
     end;
    if PSpa_Prev^[i] > PSpa_Piks^[i] then
     begin
      PSpa_Piks^[i] := PSpa_Prev^[i];
      if PSpa_Piks^[i] > 0 then
       begin
        j := (MaxVal - PSpa_Piks^[i])*spa_height div MaxVal;
        BMP_Vis.Canvas.Pixels[i,j]:=$0a0a0a;
        BMP_Vis.Canvas.Pixels[i+1,j]:=$0a0a0a;
        BMP_Vis.Canvas.Pixels[i+2,j]:=$0a0a0a;
       end;
      Dec(PSpa_Piks^[i],(MaxVal + 1) div 16);
     end
   end;
  Form1.Canvas.CopyMode:=cmSrcCopy;
  Form1.Canvas.CopyRect(Bounds(spa_x,spa_y,spa_width,spa_height),BMP_Vis.Canvas,Rect(0,0,spa_width,spa_height));
 end;
end;

procedure ShowProgress(a1:integer);
var
 x:word;
begin
if (ProgrMax = 0) or MoveProgr.Clicked then exit;
if ProgrMax <> longword(-1) then
 ProgrPos := a1
else
 ProgrPos := 0;
if ProgrMax < ProgrPos then ProgrPos := ProgrMax;
x := DivMul(ProgrWidth,ProgrPos,ProgrMax);
if MoveProgr.PosX <> x then
 begin
  MoveProgr.HideBmp;
  OffsetRgn(MoveProgr.RgnHandle,x - MoveProgr.PosX,0);
  MoveProgr.PosX := x;
  MoveProgr.Redraw(False);
 end;
end;

procedure TForm1.SetDefault;
var
 IsPl:boolean;
begin
{$IFDEF Windows}
IsPl := WOThreadActive;
{$ELSE Windows}
IsPl := IsPlaying;
{$ENDIF Windows}
PreAmp := PreAmpDef;
BeeperMax := BeeperMaxDef;
Set_Z80_Frq(FrqZ80Def);
Set_Player_Frq(Interrupt_FreqDef);
if not IsPl then Set_Sample_Rate(SampleRateDef);
Set_Chip_Frq(AY_FreqDef);
Set_MFP_Frq(MFPTimerModeDef,MFPTimerFrqDef);
IntOffset := IntOffsetDef;
Set_N_Tact(MaxTStatesDef);
if not IsPl then
 begin
  Set_Sample_Bit(SampleBitDef);
  Set_Stereo(NumOfChanDef);
  SetBuffers(BufLen_msDef,NumberOfBuffersDef);
{$IFDEF Windows}
  WODevice := WODeviceDef;
{$ENDIF Windows}
 end;
Set_Mode_Manual(Index_ALDef,Index_ARDef,Index_BLDef,Index_BRDef,
                Index_CLDef,Index_CRDef);
ChType := YM_Chip;
SetFilter(1);
BASSFFTType := BASS_DATA_FFT8192;
BASSFFTNoWin := 0;//BASS_DATA_FFT_NOWINDOW;
BASSFFTRemDC := BASS_DATA_FFT_REMOVEDC;
BASSAmpMin := 0.003;
BASSNetAgent := '';
BASSNetUseProxy := True;
BASSNetProxy := '';
Calculate_Level_Tables;
RedrawPlaylist(ShownFrom,False);
CalculateTotalTime(False);
end;

procedure TForm1.ButOpenClick(Sender: TObject);
begin
ButOpen.UnPush;
if GetKeyState(VK_SHIFT) and 128 <> 0 then
 Form3.Add_Directory_Dialog(False)
{$IFDEF Windows}
else if GetKeyState(VK_CONTROL) and 128 <> 0 then
 Form3.Add_CD_Dialog(False)
{$ENDIF Windows}
else
 Form3.Add_Item_Dialog(False);
end;

procedure TForm1.PlayClick(Sender: TObject);
begin
if IsPlaying then exit;
if not FileAvailable then
 begin
  ButPlay.UnPush;
  exit;
 end;
PlayCurrent;
end;

procedure TForm1.ButPauseClick(Sender: TObject);
begin
if not IsPlaying then
 begin
  ButPause.UnPush;
  exit;
 end;
case CurFileType of
BASSFileMin..BASSFileMax{$IFNDEF Windows}, MinAYChipFile..MaxAYChipFile{$ENDIF Windows}:
 begin
  Paused := True;
  SwitchPause;
  Paused := BASSPaused;
 end;
{$IFDEF Windows}
CDAFile:
 begin
  CDSwitchPause(CurCDNum,Handle);
  Paused := CDPlayingPaused
 end;
MidFileMin..MidFileMax:
 begin
  MIDIParams^.paused := not MIDIParams^.paused;
  Paused := MIDIParams^.paused;
 end
else
 WOPauseRestart
 {$ENDIF Windows}
end;
if not Paused then
 begin
  FIDO_SaveStatus(FIDO_Playing);
  ButPause.Switch_Off;
 end
else
 begin
  FIDO_SaveStatus(FIDO_Nothing);
  ButPause.Switch_On;
 end;
end;

procedure TForm1.ButStopClick(Sender: TObject);
begin
try
 StopAndFreeAll
finally
 ButStop.UnPush;
end;
end;

procedure RestoreControls;
begin
 Form1.FIDO_SaveStatus(FIDO_Nothing);
 ButPlay.Switch_Off;

//Commented due bugs in Lazarus in Win98
// Form2.GroupBox3.Enabled := True;
Form2.RadioButton29.Enabled := True;
Form2.RadioButton24.Enabled := True;
Form2.RadioButton23.Enabled := True;
Form2.RadioButton8.Enabled := True;
Form2.RadioButton9.Enabled := True;
Form2.RadioButton10.Enabled := True;
Form2.RadioButton27.Enabled := True;
Form2.Edit31.Enabled := True;
Form2.SpeedButton1.Enabled := True;

 Form2.GroupBox4.Enabled := True;
 Form2.Buff.Enabled := True;
 Form2.GroupBox10.Enabled := True;
 Form2.GroupBox13.Enabled := True;
 Form2.RadioButton13.Enabled := True;
 Form2.RadioButton14.Enabled := True;

 ButStop.UnPush;
 ButPause.Switch_Off;
 Form2.Edit12.Text := ''; Form2.Edit13.Text := ''; Form2.Edit14.Text := '';
 Form2.Edit15.Text := ''; Form2.Edit16.Text := ''; Form2.Edit17.Text := '';
 Form2.Edit18.Text := ''; Form2.Edit23.Text := ''; Form2.Edit26.Text := '';
 Form2.CheckBox4.Checked := False;
 Form2.CheckBox5.Checked := False;
 Form2.CheckBox6.Checked := False;
 Form2.CheckBox7.Checked := False;
end;

procedure PlayCurrent;
var
 BDLL:TBassDll;
begin
case ChType of
AY_Chip:
 begin
  Led_AY.State := False;
  Led_YM.State := True
 end;
YM_Chip:
 begin
  Led_AY.State := True;
  Led_YM.State := False
 end
end;
Led_Stereo.State := NumberOfChannels = 1;
Led_AY.Redraw(False);
Led_YM.Redraw(False);
Led_Stereo.Redraw(False);

ButPlay.Switch_On;
Form1.ShowAllParams;
ButPause.Switch_Off;
ButStop.UnPush;

//Commented due bugs in Lazarus in Win98
//Form2.GroupBox3.Enabled := False;
Form2.RadioButton29.Enabled := False;
Form2.RadioButton24.Enabled := False;
Form2.RadioButton23.Enabled := False;
Form2.RadioButton8.Enabled := False;
Form2.RadioButton9.Enabled := False;
Form2.RadioButton10.Enabled := False;
Form2.RadioButton27.Enabled := False;
Form2.Edit31.Enabled := False;
Form2.SpeedButton1.Enabled := False;

Form2.GroupBox4.Enabled := False;
Form2.Buff.Enabled := False;
Form2.GroupBox10.Enabled := False;
Form2.GroupBox13.Enabled := False;
Form2.RadioButton13.Enabled := False;
Form2.RadioButton14.Enabled := False;

Form1.FIDO_SaveStatus(FIDO_Playing);

try
 InitForAllTypes(True);
case CurFileType of
BASSFileMin..BASSFileMax{$IFNDEF Windows}, MinAYChipFile..MaxAYChipFile{$ENDIF Windows}:
 begin
  case CurFileType of
  WMAFile: BDLL := BASSWMADLL;
  APEFile: BDLL := BASSAPEDLL;
  FLACFile: BDLL := BASSFLACDLL;
  WVFile: BDLL := BASSWVDLL;
  AC3File: BDLL := BASSAC3DLL;
  else BDLL := BASSDLL;
  end;
  StartBASS(BDLL);
 end;
{$IFDEF Windows}
CDAFile:
 StartCD(CurCDNum,CurCDTrk);
MIDFileMin..MIDFileMax:
 StartMIDIThread;
else
 StartWOThread;
{$ENDIF Windows}
end;
except
 RestoreControls;
 ShowException(ExceptObject, ExceptAddr);
end;
end;

procedure TForm1.MessageSkipper;
{$IFDEF Windows}
var
 masg:TMsg;
{$ENDIF Windows}
begin
{$IFDEF Windows}
if PeekMessage(masg,Handle,WM_LBUTTONDOWN,WM_LBUTTONDOWN,PM_REMOVE) then
 May_Quit:=True;
while PeekMessage(masg,0,0,0,PM_REMOVE) do
 case masg.message of
 WM_KEYDOWN:
  if masg.wparam=VK_ESCAPE then May_Quit:=True;
 WM_PAINT:
  begin
   TranslateMessage(Masg);
   DispatchMessage(Masg);
  end;
 end;
{$ELSE Windows}
Application.ProcessMessages; //TODO: Non Win32 MessageSkipper
{$ENDIF Windows}
end;


procedure TForm1.NewMessageSkipper;
{$IFNDEF Windows}
begin
Application.ProcessMessages; //TODO: non Win32 NewMessageSkipper
end;
{$ELSE Windows}
var
 masg:TMsg;
begin
if PrgBox then
 while PeekMessage(masg,Form4.Handle,WM_MOUSEFIRST,WM_MOUSELAST,PM_REMOVE) do
  begin
   TranslateMessage(Masg);
   DispatchMessage(Masg);
  end;
while PeekMessage(masg,0,0,0,PM_REMOVE) do
 case masg.message of
 WM_KEYDOWN:
  if masg.wparam=VK_ESCAPE then May_Quit:=True;
 WM_PAINT:
  begin
   TranslateMessage(Masg);
   DispatchMessage(Masg);
  end;
 end;
end;
{$ENDIF Windows}

procedure SetCommandLine(clstr:string);
begin
if InitialScan then
 Form1.CommandLineInterpreter(clstr,False)
else
 begin
  SetLength(AfterScan,Length(AfterScan) + 1);
  AfterScan[Length(AfterScan) - 1] := clstr;
 end;
end;

{$IFDEF Windows}

procedure TForm1.WMLINEPARAM(var Msg: TMsg);
var
 HBlock:DWORD;
 HAddr:PChar;
begin
HBlock := OpenFileMapping(FILE_MAP_ALL_ACCESS,False,'Ay_Emul Command Line Area');
if HBlock <> 0 then
 try
  pointer(HAddr) := MapViewOfFile(HBlock,FILE_MAP_ALL_ACCESS,0,0,arealen);
  if HAddr <> nil then
   try
    SetCommandLine(HAddr);
   finally
    UnmapViewOfFile(pointer(HAddr));
   end;
 finally
  CloseHandle(HBlock);
 end;
end;
{$ENDIF Windows}

procedure TForm1.CommandLineInterpreter(CL:string;Start:boolean);
var
 CLPos,CLLen,fileadp:integer;
 Param:string;
 fileex,quote,Fast,fileadd:boolean;
 ParamFiles:TStringList;

 procedure CommandLineParameter(CLP:string);
 var
  Ch:char;
  ErrPos,NewFrq,i,j:integer;
  usils:array[0..5]of byte;
  TempStr:string;
 begin
  if CLP = '' then exit;
  if CLP[1] = '/' then
   begin
    if Length(CLP) < 2 then exit;
    case char(byte(CLP[2]) or $20) of
    's':
      begin
       Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
       if ErrPos = 0 then
        Set_Sample_Rate2(NewFrq)
      end;
    'b':
     begin
      Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
      if ErrPos = 0 then
       Set_Sample_Bit2(NewFrq)
     end;
    'z':
     begin
      Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
      if ErrPos = 0 then Set_Z80_Frq2(NewFrq)
     end;
    'y':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      Val(CLP,NewFrq,ErrPos);
      if ErrPos = 0 then Set_Chip_Frq2(NewFrq)
      else if CLP = 'list' then Form2.CheckBox3.Checked := True
      else if CLP = 'mixer' then Form2.CheckBox3.Checked := False
     end;
    'q':
     begin
      CLP := Trim(Copy(CLP,3,Length(CLP) - 2));
      if CLP = '' then
       Set_MFP_Frq2(0,0)
      else
       begin
        Val(CLP,NewFrq,ErrPos);
        if ErrPos = 0 then
         Set_MFP_Frq2(1,NewFrq)
       end
     end;
    't':
     begin
      Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
      if ErrPos = 0 then Set_IntOffset2(Newfrq)
     end;
    'a':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then
       IndicatorChecked := True
      else if CLP = 'off' then
       IndicatorChecked := False
      else if CLP = 'dd' then
       begin
        fileadd := True;
        fileadp := -1
       end
      else if CLP = 'dp' then
       fileadd := True
     end;
    'f':
     begin
      TempStr := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if TempStr = 'on' then
       SpectrumChecked := True
      else if TempStr = 'off' then
       SpectrumChecked := False
      else if (Length(TempStr) > 2) and (TempStr[1] = 'd') then
       begin
        CLP := Copy(CLP,5,Length(CLP) - 4);
        case TempStr[2] of
        'f':FIDO_Descriptor_FileName := CLP;
        'n':FIDO_Descriptor_Nothing := CLP;
        's':FIDO_Descriptor_Suffix := CLP;
        'p':FIDO_Descriptor_Prefix := CLP;
        'e':FIDO_Descriptor_Enabled := CLP <> '0';
        'k':FIDO_Descriptor_KillOnNothing := CLP <> '0';
        'x':FIDO_Descriptor_KillOnExit := CLP <> '0';
        'w':FIDO_Descriptor_WinEnc := CLP <> '0'
        end
       end
     end;
    'i':
     Set_N_TactS(Copy(CLP,3,Length(CLP) - 2));
    'l':
     if Length(CLP) = 3 then
      begin
       Ch := char(byte(CLP[3]) or $20);
       if Ch in ['e','r'] then
        Set_Language2(Ch = 'r')
      end;
    'n':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      Val(CLP,NewFrq,ErrPos);
      if ErrPos = 0 then
       Set_Player_Frq2(NewFrq)
      else if CLP = 'list' then
       Form2.CheckBox9.Checked := True
      else if CLP = 'mixer' then
       Form2.CheckBox9.Checked := False
     end;
    'c':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then
       Set_Loop2(True)
      else if CLP = 'off' then
       Set_Loop2(False)
     end;
    {$IFDEF Windows}
    'r':
     if Length(CLP) = 3 then
      begin
       Ch := char(byte(CLP[3]) or $20);
       if Ch in ['i','n','h'] then
        begin
         case Ch of
         'i':SetPriority2(IDLE_PRIORITY_CLASS);
         'n':SetPriority2(NORMAL_PRIORITY_CLASS)
         else SetPriority2(HIGH_PRIORITY_CLASS)
         end
        end;
      end;
    {$ENDIF Windows}
    'h':
     begin
      CLP := UpperCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'MONO' then NewFrq := 0
      else if CLP = 'AYABC' then NewFrq := 1
      else if CLP = 'AYACB' then NewFrq := 2
      else if CLP = 'AYBAC' then NewFrq := 3
      else if CLP = 'AYBCA' then NewFrq := 4
      else if CLP = 'AYCAB' then NewFrq := 5
      else if CLP = 'AYCBA' then NewFrq := 6
      else if CLP = 'YMABC' then NewFrq := 7
      else if CLP = 'YMACB' then NewFrq := 8
      else if CLP = 'YMBAC' then NewFrq := 9
      else if CLP = 'YMBCA' then NewFrq := 10
      else if CLP = 'YMCAB' then NewFrq := 11
      else if CLP = 'YMCBA' then NewFrq := 12
      else if CLP = 'LIST' then
       begin
        Form2.CheckBox1.Checked := True;
        NewFrq := -1
       end
      else if CLP = 'MIXER' then
       begin
        Form2.CheckBox1.Checked := False;
        NewFrq := -1
       end
      else
       begin
        i := 1;
        CLP := CLP + ',';
        for j := 0 to 5 do
         begin
          TempStr := '';
          while (i <= Length(CLP)) and (CLP[i] <> ',') do
           begin
            TempStr := TempStr + CLP[i];
            Inc(i)
           end;
          Inc(i);
          if i - 1 > Length(CLP) then break;
          Val(TempStr,usils[j],ErrPos);
          if ErrPos <> 0 then break
         end;
        if (i - 1 <= Length(CLP)) and (ErrPos = 0) then
         with Form2 do
          for j := 0 to 5 do
           SetChan2(usils[j],j);
        NewFrq := -1
       end;
      if NewFrq >= 0 then Form2.SetChanIndexes(NewFrq)
     end;
    'd':
     begin
      CLP := UpperCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'MONO' then Set_Stereo2(1)
      else if CLP = 'STEREO' then Set_Stereo2(2)
      else if CLP = 'LIST' then Form2.CheckBox8.Checked := True
      else if CLP = 'MIXER' then Form2.CheckBox8.Checked := False
     end;
    'e':
     begin
      CLP := UpperCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'AY' then Set_Chip2(AY_Chip)
      else if CLP = 'YM' then Set_Chip2(YM_Chip)
      else if CLP = 'LIST' then Form2.CheckBox2.Checked := True
      else if CLP = 'MIXER' then Form2.CheckBox2.Checked := False
     end;
    'g':
     begin
      CLP := Copy(CLP,3,Length(CLP) - 2);
      if CLP = '0' then
       Set_TrayMode2(0)
      else if CLP = '1' then
       Set_TrayMode2(1)
      else if CLP = '2' then
       Set_TrayMode2(2)
     end;
    'j':
     begin
      CLP := Copy(CLP,3,Length(CLP) - 2);
      if CLP = '0' then
       TimeMode := 0
      else if CLP = '1' then
       TimeMode := 1
      else if CLP = '2' then
       TimeMode := 2;
      TimeShown := -MaxInt;
     end;
    'k':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then Do_Scroll := True
      else if CLP = 'off' then Do_Scroll := False
     end;
    'p':
     LoadSkin(ExpandFileNameUTF8({ValidateFileName(}Copy(CLP,3,Length(CLP) - 2)),False);
    'w':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then
       SetAutoSaveDefDir2(True)
      else if CLP = 'off' then
       SetAutoSaveDefDir2(False)
      else if (Length(CLP) > 2) and (CLP[1] = 'o') then
       begin
        Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
        if ErrPos = 0 then
         case CLP[2] of
         'n':Set_NumberOfBuffers2(NewFrq);
         'l':Set_BufLen_ms2(NewFrq);
{$IFDEF Windows}
         'd':Set_WODevice2(NewFrq,'')
{$ENDIF Windows}
         end
       end
     end;
    'u':
     begin
      Val(Copy(CLP,3,Length(CLP) - 2),NewFrq,ErrPos);
      if ErrPos = 0 then SetChan2(NewFrq,6)
     end;
    'v':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'hide' then
       PostMessage(Handle,WM_HIDEMINIMIZE,0,0)
      else if CLP = 'show' then
       ShowApp(False);
     end;
    'x':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then
       SetAutoSaveWindowsPos2(True)
      else if CLP = 'off' then
       SetAutoSaveWindowsPos2(False)
     end;
    '!':
     begin
      CLP := LowerCase(Copy(CLP,3,Length(CLP) - 2));
      if CLP = 'on' then
       SetAutoSaveVolumePos2(True)
      else if CLP = 'off' then
       SetAutoSaveVolumePos2(False)
     end
    end
   end
  else
   begin
    fileex := True;
    ParamFiles.Add(ExpandFileNameUTF8(CLP));
   end;
 end;

var
 First:integer;
 dir:string;

begin
if AppIsModal or not IsWindowEnabled(Handle) then { TODO : AppIsModal всегда False, так как нет аналога Delphi-функции OnModal }
 exit; //ignore command line scan for closing
//modal window to finish playlist operations

Fast := GetTickCount - LastTimeComLine < CLFast;
dir := GetCurrentDirUTF8;
ParamFiles := TStringList.Create;
fileex := False; fileadd := False; fileadp := Length(PlayListItems);
CLPos := 1;
CLLen := Length(CL);
First := 0;
while CLPos <= CLLen do
 begin
  quote := False;
  Param := '';
  while (CLPos <= CLLen) and (quote or (CL[CLPos] > ' ')) do
   begin
    if CL[CLPos] = '"' then
     quote := not quote
    else
     Param := Param + CL[CLPos];
    Inc(CLPos);
   end;
  case First of
  0:
   begin
    First := 1;
    SetCurrentDirUTF8({ValidateDirName(}Param);
   end;
  1:
   First := 2;
  else
   CommandLineParameter(Param);
  end;
  Inc(CLPos);
 end;
SetCurrentDirUTF8(dir);
if FileEx then
 begin
  try
   if not fileadd and not Start and not Fast then
    begin
     StopPlaying;
     ClearPlayList;
    end
   else if fileadd and (fileadp >= 0) then
    StopPlaying;
   Form3.Add_Files(ParamFiles);
   CalculateTotalTime(False)
  finally
   CreatePlayOrder;
   ParamFiles.Free;
   if (fileadp >= 0) and (fileadp < Length(PlayListItems)) then
    RedrawPlaylist(fileadp,True)
   else
    RedrawPlaylist(0,True);
  end;
  First := -1; if not Start then First := 0;
  if not fileadd then
   begin
    if not Fast then PlayItem(0,First)
   end 
  else if (fileadp >= 0) and (fileadp < Length(PlayListItems)) then
   PlayItem(PlayListItems[fileadp]^.Tag,First);
 end;
LastTimeComLine := GetTickCount;
end;

procedure TForm1.SwapLan;
var
 temp:string;
begin

temp := Form2.GroupBox12.Caption;
Form2.GroupBox12.Caption := Lan_Mixer_Resampler;
Lan_Mixer_Resampler := temp;

temp := Form2.RadioButton26.Caption;
Form2.RadioButton26.Caption := Lan_Mixer_FIRFilter;
Lan_Mixer_FIRFilter := temp;

temp := Form2.RadioButton28.Caption;
Form2.RadioButton28.Caption := Lan_Mixer_Averager;
Lan_Mixer_Averager := temp;

temp := Form2.Label7.Caption;
Form2.Label7.Caption := Lan_Mixer_FstInt;
Lan_Mixer_FstInt := temp;

temp := Form2.Caption;
Form2.Caption := Lan_Mixer_BoxTitle;
Lan_Mixer_BoxTitle := temp;

temp := Form2.KUsil.Caption;
Form2.KUsil.Caption := Lan_Mixer_ChansAmpl;
Lan_Mixer_ChansAmpl := temp;

temp := Form2.GroupBox2.Caption;
Form2.GroupBox2.Caption := Lan_Mixer_FrqAY;
Lan_Mixer_FrqAY := temp;

temp := Form2.GroupBox1.Caption;
Form2.GroupBox1.Caption := Lan_Mixer_ChType;
Lan_Mixer_ChType := temp;

temp := Form2.GroupBox3.Caption;
Form2.GroupBox3.Caption := Lan_Mixer_SamRate;
Lan_Mixer_SamRate := temp;

temp := Form2.GroupBox4.Caption;
Form2.GroupBox4.Caption := Lan_Mixer_BitRate;
Lan_Mixer_BitRate := temp;

temp := Form2.GroupBox5.Caption;
Form2.GroupBox5.Caption := Lan_Mixer_Chans;
Lan_Mixer_Chans := temp;

temp := Form2.GroupBox7.Caption;
Form2.GroupBox7.Caption := Lan_Mixer_FrqPlr;
Lan_Mixer_FrqPlr := temp;

temp := Form2.CheckBox1.Caption;
Form2.CheckBox1.Caption := Lan_Mixer_Get;
Form2.CheckBox2.Caption := Lan_Mixer_Get;
Form2.CheckBox3.Caption := Lan_Mixer_Get;
Form2.CheckBox8.Caption := Lan_Mixer_Get;
Form2.CheckBox9.Caption := Lan_Mixer_Get;
Lan_Mixer_Get := temp;

temp := Form2.RadioButton7.Caption;
Form2.RadioButton7.Caption := Lan_Mixer_Another;
Form2.RadioButton16.Caption := Lan_Mixer_Another;
Form2.RadioButton20.Caption := Lan_Mixer_Another;
Form2.RadioButton25.Caption := Lan_Mixer_Another;
Form2.RadioButton27.Caption := Lan_Mixer_Another;
Lan_Mixer_Another := temp;

temp := Form2.GroupBox9.Caption;
Form2.GroupBox9.Caption := Lan_Mixer_FrqZ80;
Lan_Mixer_FrqZ80 := temp;

temp := Form2.Button1.Caption;
Form2.Button1.Caption := Lan_Mixer_Restore;
Lan_Mixer_Restore := temp;

temp := Form2.Button2.Caption;
Form2.Button2.Caption := Lan_Mixer_Close;
Lan_Mixer_Close := temp;

if Russian_Interface then
 begin
  Form2.Label22.Caption := 'Hz';
  Form2.Label23.Caption := 'Hz';
  Form2.Label24.Caption := 'Hz';
  Form2.Label25.Caption := 'Hz';
  Form2.Label26.Caption := 'Hz';
  Form2.Label27.Caption := 'Hz';
  Form2.RadioButton11.Caption := '16 bit';
  Form2.RadioButton12.Caption := '8 bit'
 end
else
 begin
  Form2.Label22.Caption := 'Гц';
  Form2.Label23.Caption := 'Гц';
  Form2.Label24.Caption := 'Гц';
  Form2.Label25.Caption := 'Гц';
  Form2.Label26.Caption := 'Гц';
  Form2.Label27.Caption := 'Гц';
  Form2.RadioButton11.Caption := '16 бит';
  Form2.RadioButton12.Caption := '8 бит'
 end;

temp := Form3.Caption;
Form3.Caption := Lan_List_BoxTitle;
Lan_List_BoxTitle := temp;

temp := Form3.SpeedButton1.Caption;
Form3.SpeedButton1.Caption := Lan_List_AddItems;
Lan_List_AddItems := temp;

temp := Form3.SpeedButton2.Caption;
Form3.SpeedButton2.Caption := Lan_List_ClearList;
Lan_List_ClearList := temp;

temp := Form3.SpeedButton3.Caption;
Form3.SpeedButton3.Caption := Lan_List_SaveList;
Lan_List_SaveList := temp;

temp := Form3.SpeedButton4.Caption;
Form3.SpeedButton4.Caption := Lan_List_Tools;
Lan_List_Tools := temp;

temp := Form3.MenuItemAdjusting.Caption;
Form3.MenuItemAdjusting.Caption := Lan_List_ItemAdj;
Lan_List_ItemAdj := temp;

temp := Form3.MenuConvert.Caption;
Form3.MenuConvert.Caption := Lan_List_Convert;
Lan_List_Convert := temp;

temp := Form3.MenuSaveAs.Caption;
Form3.MenuSaveAs.Caption := Lan_List_SaveItem;
Lan_List_SaveItem := temp;

temp := Form3.MenuOpenInEditor.Caption;
Form3.MenuOpenInEditor.Caption := Lan_List_OpenInEditor;
Lan_List_OpenInEditor := temp;

Russian_Interface := not Russian_Interface;
end;


procedure TForm1.Set_Chip_Frq2(Fr: integer);
begin
if Fr <> AY_Freq then
 begin
  Set_Chip_Frq(Fr);
  Form2.FrqAYTemp := AY_Freq;
  Form2.Set_Frqs
 end
end;

procedure TForm1.Set_Chip_Frq(Fr:integer);
begin
if (Fr >= 1000000) and (Fr <= 3546800) then
 begin
  SuspendPlaying;
  try
  AY_Freq := Fr;
  CalculateSpectrumPoints;
  if MFPTimerMode = 0 then
   MFPTimerFrq := round(AY_Freq * 16 / 13);
  Delay_In_Tiks := round(8192/SampleRate * AY_Freq);
  FrqAyByFrqZ80 := round(AY_Freq/FrqZ80/8 * 4294967296);
  Tik.Re := Delay_In_Tiks;
  AY_Tiks_In_Interrupt := round(AY_Freq/(Interrupt_Freq/1000 * 8));
  YM6TiksOnInt := AY_Freq/(Interrupt_Freq/1000 * 8);
  SetFilter(FilterQuality);
  if IsPlaying then
   begin
    Form2.Edit18.Text := IntToStr(AY_Freq);
    Form2.Edit26.Text := IntToStr(MFPTimerFrq)
   end
  finally
{$IFDEF Windows}
   WOUnresetPlaying;
{$ENDIF Windows}
  end;
 end;
end;

procedure TForm1.Set_MFP_Frq(Md, Fr: integer);
begin
SuspendPlaying;
try
if Md = 0 then
 begin
  MFPTimerMode := 0;
  MFPTimerFrq := round(AY_Freq * 16 / 13)
 end
else
 if (Fr >= 1000000) and (Fr <= 4365292) then
  begin
   MFPTimerMode := 1;
   MFPTimerFrq := Fr;
  end;
if IsPlaying then
 Form2.Edit26.Text := IntToStr(MFPTimerFrq);
finally
{$IFDEF Windows}
 WOUnresetPlaying;
{$ENDIF Windows}
end;
end;

procedure TForm1.ButMixerClick(Sender: TObject);
begin
if ButMixer.Is_On then
 ButMixer.Switch_Off
else
 ButMixer.Is_On := True;
Form2.Visible := ButMixer.Is_On;
{$IFNDEF Windows}
if Form2.WindowState = wsMinimized then //mask GTK error (minimize button is always visible (do bug report?)
 Form2.WindowState := wsNormal;
{$ENDIF Windows}
end;

procedure TForm1.Set_Z80_Frq(NewF: integer);
begin
if (NewF >= 1000000) and (NewF <= 8000000) then
 begin
  SuspendPlaying;
  try
  if (FrqZ80 <> NewF) and FileAvailable and
     (CurFileType in [OUTFile,ZXAYFile,AYFile,AYMFile,EPSGFile]) then
   begin
    Time_ms := trunc(Time_ms/NewF*FrqZ80+0.5);
    ProgrMax := trunc(Time_ms/1000*SampleRate+0.5);
    VProgrPos := trunc(VProgrPos/NewF*FrqZ80+0.5)
   end;
  FrqZ80 := NewF;
  FrqAyByFrqZ80 := trunc(AY_Freq/FrqZ80/8*4294967296+0.5);
  finally
{$IFDEF Windows}
   WOUnresetPlaying
{$ENDIF Windows}
  end;
  RedrawPlaylist(ShownFrom,False);
  CalculateTotalTime(False);
 end;
end;

procedure TForm1.Set_Z80_Frq2(NewF: integer);
begin
if NewF <> FrqZ80 then
 begin
  Set_Z80_Frq(NewF);
  Form2.Set_Z80Frqs
 end 
end;

procedure TForm1.Set_N_Tact(NewF:integer);
begin
if (NewF > 9999) and (NewF <= 200000) then
 begin
  SuspendPlaying;
  try
  if (MaxTStates <> NewF) and FileAvailable and
     (CurFileType in [AYFile,AYMFile]) then
   begin
    Time_ms := trunc(Time_ms/MaxTStates*NewF+0.5);
    ProgrMax := trunc(Time_ms/1000*SampleRate+0.5);
    VProgrPos := trunc(VProgrPos/MaxTStates*NewF+0.5)
   end;
  MaxTStates := NewF;
  if IntOffset >= MaxTStates then
   begin
    IntOffset := MaxTStates - 1;
    Form2.FTact.Text := IntToStr(IntOffset)
   end
  finally
{$IFDEF Windows}
   WOUnresetPlaying;
{$ENDIF Windows}
  end;
  RedrawPlaylist(ShownFrom,False);
  CalculateTotalTime(False);
 end;
end;

procedure TForm1.Set_N_Tact2(NT: integer);
begin
if NT <> MaxTStates then
 begin
  Set_N_Tact(NT);
  Form2.Edit19.Text := IntToStr(MaxTStates)
 end
end;

procedure TForm1.Set_N_TactS(t: string);
var
 V,ErrPos:integer;
begin
Val(t,V,ErrPos);
if ErrPos = 0 then Set_N_Tact2(V)
end;

procedure TForm1.SetVisTimerPeriod(VTP:integer);
begin
if (VTP > 9) and (VTP < 101) then
 begin
  VisTimerPeriod := VTP;
  VisTimer.Interval:=VTP;
 end;
end;

procedure TForm1.Set_Sample_Rate2(SR:integer);
begin
if (SR <> SampleRate) and not IsPlaying then
 begin
  Set_Sample_Rate(SR);
  Form2.SetSRs
 end
end;

procedure Set_Sample_Rate(SR:integer);
begin
if IsPlaying then exit;
if not ((SR >= 8000) and (SR < 300000)) then exit;
SampleRate := SR;
VisStep := round(SampleRate/100);
BufferLength := round(BufLen_ms * SampleRate / 1000);
VisPosMax := round(BufferLength * NumberOfBuffers / VisStep) + 1;
VisTickMax := VisStep * VisPosMax;
SetLength(VisPoints,VisPosMax);
Delay_In_Tiks := round(8192/SampleRate*AY_Freq);
Form1.SetFilter(FilterQuality);
end;

procedure SetSynthesizer;
begin
  if NumberOfChannels = 2 then
   begin
    if SampleBit = 8 then
     Synthesizer := @Synthesizer_Stereo8
    else
     Synthesizer := @Synthesizer_Stereo16;
   end
  else if SampleBit = 8 then
   Synthesizer := @Synthesizer_Mono8
  else
   Synthesizer := @Synthesizer_Mono16;
Calculate_Level_Tables;
end;

procedure TForm1.Set_Sample_Bit2(SB: integer);
begin
if (SampleBit <> SB) and ((SB = 16) or (SB = 8)) and not IsPlaying then
 begin
  Set_Sample_Bit(SB);
  case SB of
  16:Form2.RadioButton11.Checked := True;
  8:Form2.RadioButton12.Checked := True
  end
 end 
end;

procedure Set_Sample_Bit(SB:integer);
begin
if IsPlaying then exit;
SampleBit := SB;
SetSynthesizer
end;

procedure TForm1.Set_Stereo2(St: integer);
begin
if (St <> NumberOfChannels) and (St in [1,2]) and not IsPlaying then
 begin
  Set_Stereo(St);
  case St of
  1:Form2.RadioButton14.Checked := True;
  2:Form2.RadioButton13.Checked := True
  end
 end
end;

procedure Set_Stereo(St:integer);
begin
if IsPlaying then exit;
NumberOfChannels := St;
SetSynthesizer
end;

procedure TForm1.ShowAllParams;
begin
Form2.Edit12.Text := IntToStr(Index_AL);
Form2.Edit13.Text := IntToStr(Index_AR);
Form2.Edit14.Text := IntToStr(Index_BL);
Form2.Edit15.Text := IntToStr(Index_BR);
Form2.Edit16.Text := IntToStr(Index_CL);
Form2.Edit17.Text := IntToStr(Index_CR);
Form2.Edit18.Text := IntToStr(AY_Freq);
Form2.Edit23.Text := FloatToStrF(Interrupt_Freq/1000,ffFixed,7,3);
Form2.Edit26.Text := IntToStr(MFPTimerFrq);
if ChType = AY_Chip then
 Form2.CheckBox4.Checked := True
else
 Form2.CheckBox5.Checked := True;
if NumberOfChannels = 2 then
 Form2.CheckBox7.Checked := True
else
 Form2.CheckBox6.Checked := True;
end;

procedure TForm1.RestoreAllParams;
begin
with Form2 do
 begin
  if RadioButton2.Checked then
   ChType := YM_Chip
  else
   ChType := AY_Chip;
  SetChan2(TrackBar1.Position,0);
  SetChan2(TrackBar2.Position,1);
  SetChan2(TrackBar3.Position,2);
  SetChan2(TrackBar4.Position,3);
  SetChan2(TrackBar5.Position,4);
  SetChan2(TrackBar6.Position,5);
  Set_Chip_Frq(FrqAYTemp);
  Set_Player_Frq(FrqPlTemp);
  if RadioButton13.Checked then
   Set_Stereo(2)
  else
   Set_Stereo(1);
 end;
end;

procedure TForm1.ButListClick(Sender: TObject);
begin
if ButList.Is_On then
 ButList.Switch_Off
else
 ButList.Is_On := True;
Form3.Visible := ButList.Is_On;
{$IFNDEF Windows}
if Form3.WindowState = wsMinimized then //mask GTK error (minimize button is always visible (do bug report?)
 Form3.WindowState := wsNormal;
{$ENDIF Windows}
end;

procedure TForm1.ButNextClick(Sender: TObject);
begin
ButNext.UnPush;
Form3.PlayNextItem;
end;

procedure TForm1.ButPrevClick(Sender: TObject);
begin
ButPrev.UnPush;
Form3.PlayPreviousItem;
end;

procedure TForm1.WMPLAYNEXTITEM(var Msg: TMsg);
var
 Flg:boolean;
begin
{$IFDEF Windows}
if CurFileType <> CDAFile then
 StopPlaying
else
 begin
  IsPlaying := False;
  Reseted := 0;
  Paused := False;
  RestoreControls
 end;
{$ELSE Windows}
StopPlaying;
{$ENDIF Windows}
Flg := (Direction = 3) and (not ListLooped);
if not Flg then
 begin
  if Direction <> 3 then
   begin
    Form3.PlayNextItem;
    Flg := PlayingItem >= Length(PlayListItems) - 1
   end
  else
   PlayCurrent;
 end;
if not IsPlaying and Flg then
 begin
  FreeBASS;
  UnloadBASS;
  {$IFDEF Windows}
  CloseCDDevice(CurCDNum);
  {$ENDIF Windows}
 end;
end;

procedure TForm1.WMBASSMETADATA(var Msg: TMsg);
var
 Tags:TTags;
 s:string;
begin
if (MusicHandle <> 0) and MusicIsStream and TAGS_Read_Meta(MusicHandle,Tags) then
 begin
  ForceScrollForDisplay;
  s := FormatScrollString(Tags.Artist,Tags.Title,'',Unknown);
  if s = '' then ReprepareScroll else SetScrollString(s);
  CurItem.PLStr := ss;
  if not Paused then FIDO_SaveStatus(FIDO_Playing);
  TrayIcon1.Hint := ss;
 end;
end;

procedure TForm1.Set_Mode_Manual(AL,AR,BL,BR,CL,CR:byte);
begin
Index_AL := AL; Index_AR := AR;
Index_BL := BL; Index_BR := BR;
Index_CL := CL; Index_CR := CR;
Calculate_Level_Tables
end;

procedure TForm1.Set_Mode(It:Integer);
var
 Echo:integer;
begin
if It > 0 then
 begin
  if ChType = AY_Chip then Echo := 85 else Echo := 13;
  case It of
  1: begin
      Index_AL := 255; Index_AR := Echo;
      Index_BL := 170; Index_BR := 170;
      Index_CL := Echo; Index_CR := 255;
     end;
  2: begin
      Index_AL :=255; Index_AR := Echo;
      Index_BL :=Echo; Index_BR := 255;
      Index_CL :=170; Index_CR := 170;
     end;
  3: begin
      Index_AL :=170; Index_AR := 170;
      Index_BL :=255; Index_BR := Echo;
      Index_CL :=Echo; Index_CR := 255;
     end;
  4: begin
      Index_AL :=Echo; Index_AR := 255;
      Index_BL :=255; Index_BR := Echo;
      Index_CL :=170; Index_CR := 170;
     end;
  5: begin
      Index_AL := 170; Index_AR := 170;
      Index_BL := Echo; Index_BR := 255;
      Index_CL := 255; Index_CR := Echo;
     end;
  6: begin
      Index_AL := Echo; Index_AR := 255;
      Index_BL := 170; Index_BR := 170;
      Index_CL := 255; Index_CR := Echo;
     end;
   end;
 end
else
 begin
  Index_AL := 255; Index_AR := 255;
  Index_BL := 255; Index_BR := 255;
  Index_CL := 255; Index_CR := 255;
 end;
Calculate_Level_Tables;
end;

procedure TForm1.Set_Player_Frq2(Fr:integer);
begin
if Fr <> Interrupt_Freq then
 begin
  Set_Player_Frq(Fr);
  Form2.FrqPlTemp := Interrupt_Freq;
  Form2.Set_Pl_Frqs;
  RedrawPlaylist(ShownFrom,False);
  CalculateTotalTime(False);
 end;
end;

procedure TForm1.Set_Player_Frq(Fr:integer);
begin
if (Fr >= 1000) and (Fr <= 2000000) and (Interrupt_Freq <> Fr) then
 begin
  SuspendPlaying;
  try
  if (Interrupt_Freq <> Fr) and FileAvailable and
                       (CurFileType in [MinVBLType..MaxVBLType])then
   begin
    Time_ms := trunc(Time_ms/Fr*Interrupt_Freq+0.5);
    ProgrMax := trunc(Time_ms/1000*SampleRate+0.5);
    VProgrPos := trunc(VProgrPos/Fr*Interrupt_Freq+0.5)
   end;
  Interrupt_Freq := Fr;
  if IsPlaying then
   Form2.Edit23.Text := FloatToStrF(Interrupt_Freq/1000,ffFixed,70,3);
  AY_Tiks_In_Interrupt := trunc(AY_Freq/(Interrupt_Freq/1000*8)+0.5);
  YM6TiksOnInt := AY_Freq/(Interrupt_Freq/1000*8);
  finally
{$IFDEF Windows}
   WOUnresetPlaying;
{$ENDIF Windows}
  end;
 end;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
 p:PSensZone;
 p1:PMoveZone;
 p2:PButtZone;
 OfsR:integer;
//{$IFDEF Windows}
// r:TRect;
//{$ENDIF Windows}
begin
if ssDouble in Shift then
 if (X >= scr_x) and (X < scr_x + scr_width) and
    (Y >= scr_y) and (Y < scr_y + scr_height) then
  begin
   Do_Scroll := not Do_Scroll;
   exit
  end;

if Button = mbLeft then
 begin
  if MoveWin.Touche(X,Y) then
   begin
//{$IFDEF Windows}
//    SystemParametersInfo(SPI_GETWORKAREA,0,@r,0);
//    ClipCursor(@r);
//{$ELSE Windows}
{$IFNDEF Windows}
//    BeginAutoDrag;
   BeginDrag(False);
{$ENDIF Windows}
   end;
  p := SensZoneRoot;
  while p <> nil do
   begin
    if p^.Touche(X,Y) then
     p^.Clicked := True;
    p := p^.Next;
   end;
  p2 := ButtZoneRoot;
  while p2 <> nil do
   begin
    if (p2^.Clicked = 0) and p2^.Touche(X,Y) then
     begin
      p2^.Clicked := 1;
      p2^.Push;
     end;
    p2 := p2^.Next
   end;
  p1 := MoveZoneRoot;
  while p1 <> nil do
   begin
    if p1^.Bmps then
     begin
      if p1^.ToucheBut(X,Y) then
       begin
        p1^.OldX := X;
        p1^.Delt := X - p1^.posX;
        p1^.Clicked := True
       end
      else if p1^.Touche(X,Y) then
       begin
        p1^.Clicked := True;
        OfsR := X - p1^.zx - p1^.bm1w div 2;
        if OfsR > p1^.zw - p1^.bm1w then
         OfsR := p1^.zw - p1^.bm1w
        else if OfsR < 0 then
         OfsR := 0;
        if OfsR <> p1^.PosX then
         begin
          p1^.HideBmp;
          OffsetRgn(p1^.RgnHandle,OfsR - p1^.PosX,0);
          p1^.PosX := OfsR;
          p1^.Redraw(False);
          p1^.Action(Self)
         end;
        p1^.OldX := X;
        p1^.Delt := X - p1^.posX;
       end;
     end
    else if p1^.Touche(X,Y) then
     begin
      p1^.OldX := X;
      p1^.OldY := Y;
      p1^.Clicked := True;
     end;
    p1 := p1^.Next;
   end;
 end;
end;

procedure TForm1.ButToolsClick(Sender: TObject);
begin
if not ButTools.Is_On then
 begin
  ButTools.Is_On := True;
  FinderWorksNow := False;
  Form6 := TForm6.Create(Self);
 end
else if not FinderWorksNow then
{$IFDEF Windows}
 PostMessage(Form6.Handle,WM_CLOSE,0,0);
{$ELSE Windows}
 Form6.Close;
{$ENDIF Windows}
end;

procedure MainWinRepaint;
var
 p:PButtZone;
 p1:PMoveZone;
 p2:PLedZone;
begin
if LedZoneRoot <> nil then
 begin
  p2 := LedZoneRoot;
  repeat
   p2^.Redraw(True);
   p2 := p2^.Next;
  until p2 = nil;
 end;
if ButtZoneRoot <> nil then
 begin
  p := ButtZoneRoot;
  repeat
   p^.Redraw(True);
   p := p^.Next;
  until p = nil;
 end;
if MoveZoneRoot <> nil then
 begin
  p1 := MoveZoneRoot;
  repeat
   p1^.Redraw(True);
   p1 := p1^.Next;
  until p1 = nil;
 end;

BMP_DBuffer.Canvas.CopyMode:=cmSrcCopy;
BMP_DBuffer.Canvas.CopyRect(Bounds(scr_x,scr_y,scr_width,scr_height),BMP_Scroll.Canvas,Rect(0,0,scr_width,scr_height));
BMP_DBuffer.Canvas.CopyRect(Bounds(time_x,time_y,time_width,time_height),BMP_Time.Canvas,Rect(0,0,time_width,time_height));
Form1.Canvas.CopyMode:=cmSrcCopy;
Form1.Canvas.CopyRect(Rect(0,0,MWWidth,MWHeight),BMP_DBuffer.Canvas,Rect(0,0,MWWidth,MWHeight));
end;

procedure TForm1.ButLoopClick(Sender: TObject);
begin
if ButLoop.Is_On then
 ButLoop.Switch_Off
else
 ButLoop.Is_On := True;
Do_Loop := ButLoop.Is_On;
BASS_SetLoop;
{$IFDEF Windows}
MIDI_SetLoop;
{$ENDIF Windows}
end;

constructor TSensZone.Create(ps:PSensZone;x,y,w,h:integer;pr:TNotifyEvent);
var
 p:PSensZone;
begin
inherited Create;
zx := x; zy := y; zw := w; zh := h;
if SensZoneRoot = nil then
 SensZoneRoot := ps
else
 begin
  p := SensZoneRoot;
  while p^.Next <> nil do p := p^.Next;
  p^.Next := ps;
 end;
Next := nil;
Clicked := False;
Action := pr;
end;

function TSensZone.Touche(x,y:integer):boolean;
begin
Result := (x >= zx) and (x < zx + zw) and (y >= zy) and (y < zy + zh)
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
 p:PButtZone;
 p1:PMoveZone;
 OfsR:integer;
begin
if ssShift in Shift then Shift := Shift - [ssShift];
if [ssLeft] = Shift then
 begin
  p := ButtZoneRoot;
  while p <> nil do
   begin
    if (p^.Clicked = 1) and not p^.Is_On then
     if p^.Touche(X,Y) then
      p^.Push
     else
      p^.UnPush;
    p := p^.Next
   end;
  p1 := MoveZoneRoot;
  while p1 <> nil do
   begin
    if p1^.Clicked then
     begin
      if p1^.Bmps then
       begin
        OfsR := p1^.posX + X - p1^.OldX;
        p1^.OldX := X;
        if OfsR < 0 then
         begin
          p1^.OldX := p1^.Delt;
          OfsR := 0;
         end
        else if OfsR > p1^.zw - p1^.bm1w then
         begin
          OfsR := p1^.zw - p1^.bm1w;
          p1^.OldX := OfsR + p1^.Delt;
         end;
        if OfsR <> p1^.PosX then
         begin
          p1^.HideBmp;
          OffsetRgn(p1^.RgnHandle,OfsR - p1^.PosX,0);
          p1^.PosX := OfsR;
          p1^.Redraw(False);
          p1^.Action(Self);
         end;
       end
      else
       begin
        p1^.PosX := X - p1^.OldX;
        p1^.PosY := Y - p1^.OldY;
        p1^.Action(Self);
       end;
     end;
    p1 := p1^.Next;
   end;
 end;
end;

procedure TForm1.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
 p:PSensZone;
 p1:PMoveZone;
 p2:PButtZone;
begin
if Button = mbLeft then
 begin
//{$IFDEF Windows}
//  ClipCursor(nil);
//{$ENDIF Windows}
  p := SensZoneRoot;
  while p <> nil do
   begin
    if p^.Clicked then
     begin
      if p^.Touche(X,Y) then
       p^.Action(Self);
      p^.Clicked := False;
     end;
    p := p^.Next;
   end;
  p2 := ButtZoneRoot;
  while p2 <> nil do
   begin
    if p2^.Clicked = 1 then
     begin
      if p2^.Touche(X,Y) then
       p2^.Action(Self);
      if ButtZoneRoot = nil then break; //FPC
      p2^.Clicked := 0;
     end;
    p2 := p2^.Next;
   end;
  p1 := MoveZoneRoot;
  while p1 <> nil do
   begin
    if p1^.Clicked then
     begin
      p1^.Clicked := False;
      p1^.Action(Self);
     end;
    p1 := p1^.Next;
   end;
 end;
end;

constructor TButtZone.Create(ps:PButtZone;x,y,w,h,rh:integer;Bmp:TBitmap;x1,y1,x2,y2:integer;pr:TNotifyEvent);
var
 p:PButtZone;
begin
inherited Create;
zx := x; zy := y; zw := w; zh := h;
RgnHandle := rh;
if ButtZoneRoot = nil then
 ButtZoneRoot := ps
else
 begin
  p := ButtZoneRoot;
  while p^.Next <> nil do p := p^.Next;
  p^.Next := ps
 end;
Next := nil;
Is_On := False;
Is_Pushed := False;
Clicked := 0;
Action := pr;
Bmp1 := TBitmap.Create; Bmp1.Width:=zw; Bmp1.Height:=zh;
Bmp1.Canvas.CopyMode:=cmSrcCopy;
Bmp1.Canvas.CopyRect(Rect(0,0,zw,zh),Bmp.Canvas,Bounds(x1,y1,zw,zh));
Bmp2 := TBitmap.Create; Bmp2.Width:=zw; Bmp2.Height:=zh;
Bmp2.Canvas.CopyMode:=cmSrcCopy;
Bmp2.Canvas.CopyRect(Rect(0,0,zw,zh),Bmp.Canvas,Bounds(x2,y2,zw,zh));
end;

function TButtZone.Touche(x,y:integer):boolean;
begin
if RgnHandle <> 0 then
 Result := PtInRegion(RgnHandle,x,y)
else
 Result := (x >= zx) and (x < zx + zw) and (y >= zy) and (y < zy + zh)
end;

procedure TButtZone.Free;
begin
Bmp1.Free;
Bmp2.Free;
inherited;
end;

procedure TButtZone.Redraw(OnCanvas:boolean);
begin
if OnCanvas then
 begin
 BMP_DBuffer.Canvas.CopyMode:=cmSrcCopy;
 if not Is_Pushed then
  BMP_DBuffer.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp1.Canvas,Rect(0,0,zw,zh))
 else
  BMP_DBuffer.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp2.Canvas,Rect(0,0,zw,zh));
 end
else
 begin
 Form1.Canvas.CopyMode:=cmSrcCopy;
 if not Is_Pushed then
  Form1.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp1.Canvas,Rect(0,0,zw,zh))
 else
  Form1.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp2.Canvas,Rect(0,0,zw,zh));
 end;
end;

procedure TButtZone.Push;
begin
if not Is_Pushed then
 begin
  Is_Pushed := True;
  Redraw(False);
 end
end;

procedure TButtZone.UnPush;
begin
if Is_Pushed then
 begin
  Is_Pushed := False;
  Redraw(False);
 end;
end;

procedure TButtZone.Switch_On;
begin
if not Is_On then
 Is_On := True;
Push;
end;

procedure TButtZone.Switch_Off;
begin
if Is_On then
 Is_On := False;
UnPush;
end;

constructor TLedZone.Create(ps:PLedZone;x,y,w,h:integer;Bmp:TBitmap;x1,y1,x2,y2:integer);
var
 p:PLedZone;
begin
inherited Create;
zx := x; zy := y; zw := w; zh := h;
if LedZoneRoot = nil then
 LedZoneRoot := ps
else
 begin
  p := LedZoneRoot;
  while p^.Next <> nil do p := p^.Next;
  p^.Next := ps;
 end;
Next := nil;
State := False;
Bmp1 := TBitmap.Create; Bmp1.Width:=zw; Bmp1.Height:=zh;
Bmp1.Canvas.CopyMode:=cmSrcCopy;
Bmp1.Canvas.CopyRect(Rect(0,0,zw,zh),Bmp.Canvas,Bounds(x1,y1,zw,zh));
Bmp2 := TBitmap.Create; Bmp2.Width:=zw; Bmp2.Height:=zh;
Bmp2.Canvas.CopyMode:=cmSrcCopy;
Bmp2.Canvas.CopyRect(Rect(0,0,zw,zh),Bmp.Canvas,Bounds(x2,y2,zw,zh));
end;

procedure TLedZone.Redraw(OnCanvas:boolean);
begin
if OnCanvas then
 begin
 BMP_DBuffer.Canvas.CopyMode:=cmSrcCopy;
 if not State then
  BMP_DBuffer.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp1.Canvas,Rect(0,0,zw,zh))
 else
  BMP_DBuffer.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp2.Canvas,Rect(0,0,zw,zh));
 end
else
 begin
 Form1.Canvas.CopyMode:=cmSrcCopy;
 if not State then
  Form1.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp1.Canvas,Rect(0,0,zw,zh))
 else
  Form1.Canvas.CopyRect(Bounds(zx,zy,zw,zh),Bmp2.Canvas,Rect(0,0,zw,zh));
 end;
end;

procedure TLedZone.Free;
begin
Bmp1.Free;
Bmp2.Free;
inherited;
end;

procedure TForm1.ButCloseClick(Sender: TObject);
begin
Close;
end;

procedure TForm1.ButAboutClick(Sender: TObject);
begin
  with TAboutBox.Create(Self) do
  try
   {$ifdef beta}
   AbDBuffer.Canvas.TextOut(122-AbDBuffer.Canvas.TextWidth(BetaNumber) div 2,236,BetaNumber);
   {$endif beta}
   ShowModal;
  finally
   Free;
   ButAbout.UnPush;
  end;
end;

procedure TForm1.ButSpaClick(Sender: TObject);
begin
SpectrumChecked := not SpectrumChecked
end;

procedure TForm1.ButAmpClick(Sender: TObject);
begin
IndicatorChecked := not IndicatorChecked
end;

procedure TForm1.ButTimeClick(Sender: TObject);
begin
Inc(TimeMode);
if TimeMode > 2 then TimeMode := 0;
TimeShown := -MaxInt;
end;

constructor TMoveZone.Create(ps:PMoveZone;x,y,w,h,y1,h1,rh:integer;
                                                        pr:TNotifyEvent);
var
 p:PMoveZone;
begin
inherited Create;
zx := x; zy := y; zw := w; zh := h;
zy1 := y1; zh1 := h1;
RgnHandle := rh;
PosX := 0;
if MoveZoneRoot = nil then
 MoveZoneRoot := ps
else
 begin
  p := MoveZoneRoot;
  while p^.Next <> nil do p := p^.Next;
  p^.Next := ps;
 end;
Next := nil;
Bmps := False;
State := False;
Clicked := False;
Action := pr;
end;

function TMoveZone.ToucheBut(x,y:integer):boolean;
begin
Result :=
PtInRegion(RgnHandle,x,y);
//((x >= PosX + zx) and (x < PosX + zx + Bm1w) and (y >= zy) and (y < zy + Bm1h));
end;

function TMoveZone.Touche(x,y:integer):boolean;
begin
Result := ((x >= zx) and (x < zx + zw) and (y >= zy + zy1) and (y < zy + zy1 + zh1));
end;

procedure TForm1.DoMovingWindow(Sender: TObject);
begin
(*{$IFNDEF Windows}
Left := Left + MoveWin.PosX;
Top := Top + MoveWin.PosY;
{$ENDIF Windows}*)
end;

procedure TForm1.DoMovingScroll(Sender: TObject);
begin
Inc(MoveScr.OldX,MoveScr.PosX);
if sw <= scr_width then exit;
if Scroll_Distination <> Item_Displayed then exit;
Dec(HorScrl_Offset,MoveScr.PosX);
if HorScrl_Offset < 0 then
 HorScrl_Offset := 0
else if HorScrl_Offset > sw - scr_width then
 HorScrl_Offset := sw - scr_width;
RedrawScroll;
end;

procedure TMoveZone.AddBitmaps(Bmp:TBitmap;x1,y1,bw,bh:integer;m:boolean);
begin
Bmps := True;
Bmp1 := TBitmap.Create; Bmp1.Width:=bw; Bmp1.Height:=bh;
Bm1w := bw;
Bm1h := bh;
Bmp1.Canvas.CopyMode:=cmSrcCopy;
Bmp1.Canvas.CopyRect(Rect(0,0,bw,bh),Bmp.Canvas,Bounds(x1,y1,bw,bh));
if m then
 begin
  Bmp1.TransparentColor:=Bmp1.Canvas.Pixels[0,0];
  Bmp1.Transparent:=True;
  Bmp1.TransparentMode:=tmFixed;
 end;
Bmp2 := TBitmap.Create; Bmp2.Width:=zw; Bmp2.Height:=zh;
Bmp2.Canvas.CopyMode:=cmSrcCopy;
Bmp2.Canvas.CopyRect(Rect(0,0,zw,zh),Bmp.Canvas,Bounds(zx,zy,zw,zh));
end;

procedure TMoveZone.Free;
begin
if Bmps then
 begin
  Bmp1.Free;
  Bmp2.Free;
 end;
inherited;
end;

procedure TMoveZone.Redraw(OnCanvas:boolean);
begin
if Bmps then
 begin
    BMP_DBuffer.Canvas.Draw(zx + PosX,zy,Bmp1);
  if not OnCanvas then
   begin
    Form1.Canvas.CopyMode:=cmSrcCopy;
    Form1.Canvas.CopyRect(Bounds(zx,zy,zw,zh),BMP_DBuffer.Canvas,Bounds(zx,zy,zw,zh));
   end;
 end;
end;

procedure TMoveZone.HideBmp;
begin
if Bmps then
 begin
  BMP_DBuffer.Canvas.CopyMode:=cmSrcCopy;
  BMP_DBuffer.Canvas.CopyRect(Bounds(zx + PosX,zy,Bm1w,Bm1h),Bmp2.Canvas,Bounds(PosX,0,Bm1w,Bm1h));
 end;
end;

procedure TForm1.DoMovingVol(Sender: TObject);
begin
VolumeCtrl := MoveVol.PosX;
SetSysVolume;
end;

procedure Rewind(newpos,maxpos:integer);
var
 i,d:longword;
{$IFDEF Windows}
 MSF:packed record
  case boolean of
  True: (MSF:DWORD);
  False:(M,S,F:byte);
 end;
{$ENDIF Windows}
begin
if not IsPlaying or Paused or MoveProgr.Clicked
{$IFDEF Windows}
   or ((CurFileType in [MIDFileMin..MIDFileMax]) and MIDIParams^.seeking)
{$ENDIF Windows}
 then exit;
if ProgrMax = longword(-1) then exit;
if newpos < 0 then
 newpos := 0
else if newpos > maxpos then
 newpos := maxpos;
i := round(newpos/maxpos * ProgrMax);
ShowProgress(i);
case CurFileType of
BASSFileMin..BASSFileMax:
 begin
  //Mask BASS error: if seeking to the end on some mod-music, sync_end message is not sent
  if not Do_Loop and (i >= ProgrMax) then
   PostMessage(Form1.Handle,WM_PLAYNEXTITEM,0,0)
  else
   begin
    d := 0;
    if (CurFileType in [StreamFileMin..StreamFileMax]) and (StreamPlayFrom > 0) then
     d := StreamPlayFrom;
    //BASS cannot seek to the end of some mod-music, but I havn't any idea how to fix it
    if BASS_ChannelSetPosition(MusicHandle,BASS_ChannelSeconds2Bytes(MusicHandle,(i + d) / 1000),BASS_POS_BYTE) then
     CurrTime_Rasch := i
    //Temporary solving problem so:
    else if i >= 50 then
     begin
      dec(i,50);
      if BASS_ChannelSetPosition(MusicHandle,BASS_ChannelSeconds2Bytes(MusicHandle,(i + d) / 1000),BASS_POS_BYTE) then
       CurrTime_Rasch := i;
     end;
   end;
 end;
{$IFDEF Windows}
CDAFile:
 begin
  CurrTime_Rasch := round(i / 75 * 1000);
  MSF.F := i mod 75;
  i := i div 75;
  MSF.S := i mod 60;
  MSF.M := i div 60;
  CDSetPosition(CurCDNum,CurCDTrk,MSF.MSF,Form1.Handle)
 end;
MIDFileMin..MIDFileMax:
 begin
  MIDIParams^.seek_to := i;
  CurrTime_Rasch := i;
  MIDIParams^.seeking := True;
 end
{$ENDIF Windows}
else
 begin
  {$IFDEF Windows}
  WOResetPlaying(True);
  {$ELSE Windows}
  BASS_ChannelStop(MusicHandle);
  {$ENDIF Windows}
  try
   RerollMusic(newpos,maxpos);
  finally
  {$IFDEF Windows}
   WOUnresetPlaying;
  {$ELSE Windows}
   BASS_ChannelPlay(MusicHandle,True);
  {$ENDIF Windows}
  end;
 end;
end;
end;

procedure TForm1.DoMovingProgr(Sender: TObject);
begin
Rewind(MoveProgr.PosX,ProgrWidth);
end;

procedure TForm1.DoMinimize;
begin
Application.Minimize;
{$IFNDEF Windows}
Form1.WindowState:=wsMinimized; //prevent GTK widgetset error
AppMinimize(Self); //second bug - onminimize is not raised
{$ENDIF Windows}
end;


procedure TForm1.ButMinClick(Sender: TObject);
begin
ButMinimize.UnPush;
DoMinimize;
end;

procedure TForm1.AppMinimize(Sender: TObject);
begin
case TrayMode of
{$IFDEF Windows}
1:
 ShowWindow(GetParent(Form1.Handle),SW_HIDE); //LCL cannot minimize tool window app
{$ENDIF Windows}
2:
 begin
  AddTrayIcon;
  RemoveTaskbarButton;
 end;
end;
end;

procedure TForm1.AppRestore(Sender: TObject);
begin
if TrayMode = 2 then
 begin
  RemoveTrayIcon;
  AddTaskbarButton;
 end;
end;

procedure TForm1.DoRestore;
begin
Application.Restore; //not work in gtk, bug?
{$IFNDEF Windows}
Form1.WindowState:=wsNormal; //prevent GTK widgetset error
AppRestore(Self); //second bug - onrestore is not raised
{$ENDIF Windows}
end;

procedure TForm1.ShowApp(Tray:boolean);
begin
if WindowState = wsMinimized then
 DoRestore
else if Tray then
 Application.BringToFront
else
 begin //real bring to front instead of taskbar button flashing
  if TrayMode = 2 then TrayMode := -1;
  Application.Minimize;
  Application.Restore;
  if TrayMode = -1 then TrayMode := 2;
 end;
end;


{$IFDEF Windows}
function IsApplicationForm(h:THandle):boolean;
var
 i:integer;
begin
Result := False;
for i := 0 to Screen.CustomFormCount - 1 do
 if Screen.CustomForms[i].Handle = h then
  begin
   Result := True;
   exit;
  end;
end;

function Overlapped:boolean;
var
 R1,R2:TRect;
// f:textfile;
 h:THandle;
begin
Result := False;
{AssignFile(f,'c:\tst.txt');
Rewrite(f);
try}
h := Form1.Handle;
if not GetWindowRect(h,R1) then exit;
(*Writeln(f,'Left: ' + IntToStr(R1.Left) + '; ' +
        'Right: ' + IntToStr(R1.Right) + '; ' +
        'Top: ' + IntToStr(R1.Top) + '; ' +
        'Bottom: ' + IntToStr(R1.Bottom) + ' ' +
         {GetWindowTitle(h) +} ' ' + IntToStr(GetParent(h)));
*)
repeat
h := GetNextWindow(h,GW_HWNDPREV); if h = 0 then exit;
if not IsWindowVisible(h) then continue;
if not GetWindowRect(h,R2) then continue;
(*Writeln(f,'Left: ' + IntToStr(R2.Left) + '; ' +
        'Right: ' + IntToStr(R2.Right) + '; ' +
        'Top: ' + IntToStr(R2.Top) + '; ' +
        'Left: ' + IntToStr(R2.Bottom) + ' ' +
         {GetWindowTitle(h) +} ' ' + IntToStr(GetParent(h)));
*)
if R2.Left = R2.Right then continue;
if R2.Top = R2.Bottom then continue;
if R1.Left > R2.Right then continue;
if R1.Right < R2.Left then continue;
if R1.Top > R2.Bottom then continue;
if R1.Bottom < R2.Top then continue;
if IsApplicationForm(h) then continue;
Result := True; exit;
until False;
{finally
CloseFile(f);
end;}
end;
{$ENDIF Windows}

procedure TForm1.TrayIcon1DblClick(Sender: TObject);
begin
 TrayIconClicked := True;
end;

procedure TForm1.TrayIcon1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
if Button = mbLeft then TrayIconClicked := True;
end;

procedure TForm1.TrayIcon1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
if Button <> mbLeft then exit;
if TrayIconClicked then
 begin
  TrayIconClicked := False;
  if (WindowState = wsMinimized)
     {$IFDEF Windows}
     or Overlapped
     {$ENDIF Windows}
  then
   ShowApp(True)
  else
//   Application.Minimize;
   DoMinimize;
 end;
end;

{$IFDEF Windows}
function WndCallback(Ahwnd: HWND; uMsg: UINT; wParam: WParam; lParam: LParam):LRESULT; stdcall;
var
 r:_RECT;
begin
 case uMsg of
 WM_NCHITTEST:
  begin
   if GetWindowRect(Ahwnd,r) and MoveWin.Touche(smallint(lParam and $FFFF) - r.Left,smallint(lParam shr 16) - r.Top) then
    Result := HTCAPTION
   else
    Result := DefWindowProc(Ahwnd,uMsg,wParam,lParam);
   exit;
  end;
 MM_MIXM_CONTROL_CHANGE:
  begin
   if (HMIXER(WParam) = SysVolumeParams.MixerHandle) and
      (LParam = SysVolumeParams.ControlID) then
    GetSysVolume;
  end;
 MM_MCINOTIFY:
  begin
   if not CheckCDNum(CurCDNum) then exit;
   if LParam <> integer(CDIDs[CurCDNum]) then exit;
   if WParam = MCI_NOTIFY_SUCCESSFUL then
    //WMPLAYNEXTITEM(Msg);
    PostMessage(Ahwnd,WM_PLAYNEXTITEM,0,0);
  end;
 end;
 Result:=CallWindowProc(PrevWndProc,Ahwnd, uMsg, WParam, LParam);
end;

function AWndCallback(Ahwnd: HWND; uMsg: UINT; wParam: WParam; lParam: LParam):LRESULT; stdcall;
begin
 case uMsg of
 WM_SYSCOMMAND:
  case (WParam and $FFF0) of
  SC_MINIMIZE:
   if TrayMode = 1 then ShowWindow(GetParent(Form1.Handle), SW_HIDE); //LCL cannot minimize app tool window
  SC_RESTORE:
   if TrayMode = 1 then ShowWindow(GetParent(Form1.Handle), SW_SHOWNA); //LCL cannot restore hidden app tool window
  end;
 end;
 Result:=CallWindowProc(PrevAWndProc,Ahwnd, uMsg, WParam, LParam);
end;
{$ENDIF Windows}

procedure TForm1.FormCreate(Sender: TObject);

//{$define CreateRgn}

{$ifdef CreateRgn}
 function AddRoundRectRgnR(a,b,c,d,e,f:integer):HRGN;
 begin
  Result := CreateRoundRectRgn(a,b,c,d,e,f);
  CombineRgn(MyFormRgn,MyFormRgn,Result,RGN_OR);
 end;

 procedure AddRoundRectRgn(a,b,c,d,e,f:integer);
 begin
  DeleteObject(AddRoundRectRgnR(a,b,c,d,e,f));
 end;
{$endif CreateRgn}

var
 i:integer;
// x,y:integer;
// l:integer;
// tf:textfile;
// pt:char;
{$ifndef CreateRgn}
 hr:HRGN;
{$endif CreateRgn}
// f:file;
// lpr:array of byte;

const
 RegionVolPoints:array[0..2] of tagPOINT =
  ((x:237+70-18;y:21+11+1),(x:237+70;y:21+11+1),(x:237+70;y:21+1));
 RegionProgrPoints:array[0..11] of tagPOINT =
  ((x:96;y:84),(x:100;y:84),(x:100;y:83),(x:112;y:83),(x:112;y:84),
   (x:116;y:84),(x:116;y:92),(x:112;y:92),(x:112;y:93),(x:100;y:93),
   (x:100;y:92),(x:96;y:92));

{$ifndef CreateRgn}
{$i rgn.inc}
{$endif CreateRgn}

begin

Randomize;

{$ifdef CreateRgn}
MyFormRgn := CreateRectRgn(51,1,311,114);
AddRoundRectRgn(0,0,115,115,115,115);
AddRoundRectRgn(358-115,0,358,115,115,115);
{$endif CreateRgn}
RgnLoop := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(62-10,110-10,62+11,110+11,21,21);
RgnBack := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(80,96,80+35,123,14,14);
RgnPlay := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(119,96,119+35,123,14,14);
RgnPause := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(158,96,158+35,123,14,14);
RgnStop := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(197,96,197+35,123,14,14);
RgnNext := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(235,96,235+35,123,14,14);
RgnOpen := {$ifdef CreateRgn}AddRoundRectRgnR{$else CreateRgn}CreateRoundRectRgn{$endif CreateRgn}(275,96,275+35,123,14,14);
RgnMixer := CreateRoundRectRgn(318,21,318+26,21+26,26,26);
RgnPList := CreateRoundRectRgn(310,77,310+26,77+26,26,26);
RgnTools := CreateRoundRectRgn(322,50,322+26,50+26,26,26);
RgnMin := CreateRoundRectRgn(282,6,282+16,6+16,16,16);
RgnClose := CreateRoundRectRgn(304,6,304+16,6+16,16,16);
RgnVol := CreatePolygonRgn(RegionVolPoints,3,ALTERNATE);
RgnProgr := CreatePolygonRgn(RegionProgrPoints,12,ALTERNATE);

SensSpa := TSensZone.Create(@SensSpa,spa_x,spa_y,spa_width,spa_height,@ButSpaClick);
SensAmp := TSensZone.Create(@SensAmp,amp_x,amp_y,amp_width,amp_height,@ButAmpClick);
SensTime := TSensZone.Create(@SensTime,time_x,time_y,time_width,time_height,@ButTimeClick);
MoveWin := TMoveZone.Create(@MoveWin,84,5,279-84,22-5,0,22-5,0,@DoMovingWindow);
MoveVol := TMoveZone.Create(@MoveVol,237,21+1,70,12,4,8,RgnVol,@DoMovingVol);
MoveProgr := TMoveZone.Create(@MoveProgr,96,83,255-96,10,2,5,RgnProgr,@DoMovingProgr);
MoveScr := TMoveZone.Create(@MoveScr,scr_x,scr_y,scr_width,scr_height,0,scr_height,0,@DoMovingScroll);

BMP_DBuffer := TBitmap.Create; BMP_DBuffer.Width := MWWidth; BMP_DBuffer.Height := MWHeight;

FileMode := 0;

LoadSkin('',True);

VolumeCtrl := MoveVol.zw - MoveVol.Bm1w;
VolumeCtrlMax := VolumeCtrl;
MoveVol.PosX := VolumeCtrl;
{$IFDEF Windows}
SysVolumeParams.Title := 'Not selected';
SysVolumeParams.MixerNumber := -1;
SysVolumeParams.DestNumber := -1;
SysVolumeParams.CtrlNumber := -1;
SysVolumeParams.MixerID := -1;
SysVolumeParams.Opened := False;
{$ENDIF Windows}

ProgrWidth := MoveProgr.zw - MoveProgr.Bm1w;

Led_AY.State := True;

{AssignFile(tf,'_rgn.txt');
Rewrite(tf);
for y := 0 to Form1.Height-1 do
 begin
  for x := 0 to Form1.Width-1 do
   Write(tf,Ord(PtInRegion(MyFormRgn,x,y)));
  Writeln(tf);
 end;
CloseFile(tf);}

{DeleteObject(MyFormRgn);
MyFormRgn := CreateRectRgn(0,0,0,0);
AssignFile(tf,'_rgn.txt');
Reset(tf);
for y := 0 to Form1.Height-1 do
 begin
  for x := 0 to Form1.Width-1 do
   begin
    Read(tf,pt);
    if Pt = '1' then
     begin
      hr := CreateRectRgn(x,y,x+1,y+1);
      CombineRgn(MyFormRgn,MyFormRgn,hr,RGN_OR);
      DeleteObject(hr);
     end;
   end;
  Readln(tf);
 end;
CloseFile(tf);}

{i := GetRegionData(MyFormRgn,0,nil);
SetLength(lpr,i);
i := GetRegionData(MyFormRgn,i,@lpr[0]);
AssignFile(f,'_rgn');
Rewrite(f,1);
BlockWrite(f,lpr[0],i);
CloseFile(f);}

{$ifndef CreateRgn}
with rgn[0] do
 MyFormRgn := CreateRectRgn(x,y,x+w,y+h);
for i := 1 to nrects do
 with rgn[i] do
  begin
   hr := CreateRectRgn(x,y,x+w,y+h);
   CombineRgn(MyFormRgn,MyFormRgn,hr,RGN_OR);
   DeleteObject(hr);
  end;
(*AssignFile(f,'_rgn');
Reset(f,1);
i := FileSize(f);
SetLength(lpr,i);
BlockRead(f,lpr[0],i);
CloseFile(f);
//SetWindowRgn(Handle,ExtCreateRegion(nil,i,lpRgnData(@lpr[0])^),True);

l := (i-32) div 16 - 1;

{MyFormRgn := CreateRectRgnIndirect(PRect(@lpr[32])^);
for i := 1 to l do
 begin
  hr := CreateRectRgnIndirect(PRect(@lpr[32+i*16])^);
  CombineRgn(MyFormRgn,MyFormRgn,hr,RGN_OR);
  DeleteObject(hr);
 end;}

AssignFile(tf,'rgn.inc');
Rewrite(tf);
Writeln(tf,'//region once created in Windows 7 and used for other OSes to avoid differences in CreateRoundRectRgn');
Writeln(tf,'nrects = ',l,';');
Writeln(tf,'rgn:packed array[0..nrects] of packed record');
Writeln(tf,' x,w:smallint;');
Writeln(tf,' y,h:shortint;');
Writeln(tf,'end = (');
for i := 0 to l do
 begin
  Write(tf,'  (x:',PRect(@lpr[32+i*16])^.Left,';w:',PRect(@lpr[32+i*16])^.Right-PRect(@lpr[32+i*16])^.Left,
            ';y:',PRect(@lpr[32+i*16])^.Top,';h:',PRect(@lpr[32+i*16])^.Bottom-PRect(@lpr[32+i*16])^.Top,')');
  if i < l then Writeln(tf,',') else Writeln(tf);
 end;
Writeln(tf,' );');
CloseFile(tf);*)

{with LPRGNDATA(@lpr[0])^ do
 begin
  MyFormRgn := CreateRectRgnIndirect(LPRECT(@lpr[rdh.dwSize])^);
  for i := 1 to rdh.nCount do
   begin
    hr := CreateRectRgnIndirect(LPRECT(@lpr[rdh.dwSize+i*16])^);
    CombineRgn(MyFormRgn,MyFormRgn,hr,RGN_OR);
    DeleteObject(hr);
   end;
 end;}
{$endif CreateRgn}

{$IFDEF Windows}
SetWindowRgn(Handle,MyFormRgn,True);
{$ENDIF Windows}

PSpa_prev := @Spa_prev;
PSpa_piks := @Spa_piks;
Synthesizer := @Synthesizer_Stereo16;
Application.OnRestore := @AppRestore;
Application.OnMinimize := @AppMinimize;
Application.OnEndSession := @AppEndSession;
Application.TaskBarBehavior := tbSingleButton;
//Application.MainFormOnTaskBar:=False;


//FIDO_Descriptor_Filename := ExtractFileDir(GetProcessFileName);
//{$IFDEF Windows}
//FIDO_Descriptor_Filename := ValidateDirName(UTF8Decode(FIDO_Descriptor_Filename));
//{$ENDIF Windows}
FIDO_Descriptor_Filename := GetAppConfigDirUTF8(False) + 'aystatus.txt';

for i := 0 to spa_num - 1 do Spa_piks[i] := 0;

BMP_Sources := TBitmap.Create; BMP_Sources.Width:=max_src; BMP_Sources.Height:=max_height;

BMP_Time := TBitmap.Create; BMP_Time.Width:=time_width; BMP_Time.Height:=time_height;
{$IFDEF Windows}
BMP_Time.Canvas.Font.Name:='MS Sans Serif'; //todo
{$ENDIF Windows}
BMP_Time.Canvas.Font.Height:=-20;
BMP_Time.Canvas.Font.Bold:=True;
BMP_Time.Canvas.Font.Color:=$464646;
BMP_Time.Canvas.Brush.Style:=bsClear;

BMP_Vis := TBitmap.Create; BMP_Vis.Width:=max_width2; BMP_Vis.Height:=max_height2;
BMP_Vis.Canvas.Pen.Color:=$464646;
BMP_Vis.Canvas.Pen.Width:=3;

BMP_VScroll := TBitmap.Create; BMP_VScroll.Width:=scr_width; BMP_VScroll.Height:=scr_lineheight*3;
{$IFDEF Windows}
BMP_VScroll.Canvas.Font.Name:='MS Sans Serif'; //todo
{$ENDIF Windows}
BMP_VScroll.Canvas.Font.Height:=-20;
BMP_VScroll.Canvas.Font.Color:=$606060;
//BMP_VScroll.Canvas.Brush.Color:=clWhite;

BMP_Scroll := TBitmap.Create; BMP_Scroll.Width:=scr_width; BMP_Scroll.Height:=scr_lineheight;

BMP_VScroll.Canvas.Brush.Color:=clWhite;
BMP_VScroll.Canvas.FillRect(0,0,scr_width,scr_lineheight*3);
CopyBmpSources;

VisTimer := TTimer.Create(Self);
VisTimer.Interval:=VisTimerPeriod;
VisTimer.OnTimer:=@VisTimerEvent;
VisTimer.Enabled:=True;

{$IFDEF Windows}
PrevWndProc:=Windows.WNDPROC(SetWindowLong(Handle,GWL_WNDPROC,PtrInt(@WndCallback)));
PrevAWndProc:=Windows.WNDPROC(SetWindowLong(GetParent(Handle),GWL_WNDPROC,PtrInt(@AWndCallback)));
{$ENDIF Windows}

end;

procedure SetScrollString(const scrstr:string);
begin
ss := scrstr;
GetStringWnJ(ss,sw,sj);
if scr_lineheight*(Scroll_Distination - Item_Displayed + 1) - Scroll_Offset = 0 then
 begin
  if scr_width < sw then
   begin
    if HorScrl_Offset > sw - scr_width then
     HorScrl_Offset := sw - scr_width - 1
   end
  else
   begin
    HorScrl_Offset := 0;
    BMP_VScroll.Canvas.FillRect(Rect(0,scr_lineheight,scr_width,scr_lineheight*2));
   end;
  RedrawScroll;
 end;
end;

procedure ReprepareScroll;
begin
if Item_Displayed > 0 then
 begin
  ss1 := GetPlayListString(PlaylistItems[Item_Displayed - 1]);
  GetStringWnJ(ss1,sw1,sj1)
 end;
if Item_Displayed < Length(PlaylistItems) - 1 then
 begin
  ss2 := GetPlayListString(PlaylistItems[Item_Displayed + 1]);
  GetStringWnJ(ss2,sw2,sj2)
 end;
if (Item_Displayed >= 0) and (Item_Displayed < Length(PlaylistItems)) then
 SetScrollString(GetPlayListString(PlaylistItems[Item_Displayed]));
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);

 procedure TryClick(Bt:TButtZone);
 begin
  if Bt.Clicked = 2 then
   begin
    Bt.Clicked := 0;
    Bt.Action(Sender);
   end;
 end;

begin
case Key of
byte('T'):
 ButTimeClick(Sender);
byte('1'):
 ButAmpClick(Sender);
byte('2'):
 ButSpaClick(Sender);
byte('P'):
 TryClick(ButTools);
byte('E'):
 TryClick(ButList);
byte('G'):
 TryClick(ButMixer);
byte('R'):
 TryClick(ButLoop);
byte('X'):
 TryClick(ButPlay);
VK_NUMPAD5:
 if not IsPlaying then
  TryClick(ButPlay)
 else
  TryClick(ButPause);
byte('V'):
 TryClick(ButStop);
byte('C'):
 TryClick(ButPause);
byte('B'),VK_NUMPAD6:
 TryClick(ButNext);
byte('Z'),VK_NUMPAD4:
 TryClick(ButPrev);
byte('L'),VK_NUMPAD0:
 TryClick(ButOpen);
end;
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);

 procedure UnClickAllButButt(Butt:TButtZone);
 var
  p:PButtZone;
 begin
  p := ButtZoneRoot;
  while p <> nil do
   begin
    if p <> @Butt then
     if p^.Clicked <> 0 then
      begin
       p^.Clicked := 0;
       if not p^.Is_On then p^.UnPush;
      end;
    p := p^.Next;
   end;
 end;

 procedure Push(Bt:TButtZone);
 begin
  if Bt.Clicked = 0 then
   begin
    UnClickAllButButt(Bt);
    Bt.Clicked := 2;
    Bt.Push;
   end;
 end;

begin
case Key of
byte('P'):
 Push(ButTools);
byte('J'):
 begin
  UnClickAllButButt(nil);
  JumpToTime
 end;
byte('E'):
 Push(ButList);
byte('G'):
 Push(ButMixer);
byte('R'):
 Push(ButLoop);
byte('X'):
 Push(ButPlay);
VK_NUMPAD5:
 if not IsPlaying then
  Push(ButPlay)
 else
  Push(ButPause);
byte('V'):
 Push(ButStop);
byte('C'):
 Push(ButPause);
byte('B'),VK_NUMPAD6:
 Push(ButNext);
byte('Z'),VK_NUMPAD4:
 Push(ButPrev);
byte('L'),VK_NUMPAD0:
 Push(ButOpen);
VK_UP,VK_NUMPAD8:
 VolUp;
VK_DOWN,VK_NUMPAD2:
 VolDown;
VK_LEFT:
 begin
  UnClickAllButButt(nil);
  if Time_ms > 0 then
   Rewind(CurrTime_Rasch - 5000,Time_ms);
 end;
VK_RIGHT:
 begin
  UnClickAllButButt(nil);
  if Time_ms > 0 then
   Rewind(CurrTime_Rasch + 5000,Time_ms);
 end;
VK_F1:
 begin
  UnClickAllButButt(nil);
  CallHelp;
 end;
VK_ESCAPE:
 DoMinimize;
// Application.Minimize; 
end;
end;

procedure TForm1.VolUp;
begin
if MoveVol.PosX < MoveVol.zw - MoveVol.Bm1w then
 begin
  MoveVol.Clicked := False;
  MoveVol.HideBmp;
  Inc(MoveVol.PosX);
  OffsetRgn(MoveVol.RgnHandle,1,0);
  MoveVol.Redraw(False);
  MoveVol.Action(Self);
 end
end;

procedure TForm1.VolDown;
begin
if MoveVol.posX > 0 then
 begin
  MoveVol.Clicked := False;
  MoveVol.HideBmp;
  Dec(MoveVol.PosX);
  OffsetRgn(MoveVol.RgnHandle,-1,0);
  MoveVol.Redraw(False);
  MoveVol.Action(Self);
 end;
end;

procedure TForm1.AddTrayIcon;
begin
TrayIcon1.Icon.LoadFromResourceName(hInstance,Format('ICON%.2u',[TrayIconNumber]));
if not FileAvailable then
 TrayIcon1.Hint := 'AY Emulator';
TrayIcon1.Show;
end;

procedure TForm1.RemoveTrayIcon;
begin
TrayIcon1.Hide;
end;

function TForm1.LoadSkin(FName:string;First:boolean):boolean;
var
 Buffer:array of byte;
 Author,Comment:string;
 rs:TResourceStream;
 s:string;
 i:integer;
 tl,mx,pl,ls,pa,l1,l2,l3,lp:boolean;
 URHandle:integer;
begin
Result := False;
try
 if FName = '' then
  begin
   rs := TResourceStream.Create(HInstance,'DEFAULTSKIN',RT_RCDATA);
   UniReadInit(URHandle,URMemory,'',rs.Memory,rs.Size);
   Compressed_Size := rs.Size - SkinIdLen - 4;
  end
 else
  begin
   UniReadInit(URHandle,URFile,FName,nil,-1);
   Compressed_Size := UniReadersData[URHandle]^.UniFileSize - SkinIdLen - 4;
  end;
 try
  SetLength(s,SkinIdLen);
  UniRead(URHandle,@s[1],SkinIdLen);
  if s <> SkinId then
   begin
    if Russian_Interface then
     ShowMessage('Файл ' + FName +
       ' не является файлом шаблона для эмулятора версии 2.0')
    else
     ShowMessage('File ' + FName +
       ' is not AY-3-8910/12 Emulator v2.0 Skin File');
    exit;
   end;
  UniRead(URHandle,@Original_Size,4);
  UniAddDepacker(URHandle,UDLZH);
  SetLength(Buffer,Original_Size);
  UniRead(URHandle,@Buffer[0],Original_Size)
 finally
  UniReadClose(URHandle);
  if FName = '' then rs.Free;
 end;

 Author := '';
 i := 0;
 while (i < Original_Size) and (Buffer[i] <> 0) do
  begin
   Author := Author + char(Buffer[i]);
   Inc(i);
  end;
 Author := CPToUTF8(Author); //todo - skins only utf8 encoding
 Comment := '';
 Inc(i);
 while (i < Original_Size) and (Buffer[i] <> 0) do
  begin
   Comment := Comment + char(Buffer[i]);
   Inc(i);
  end;
 Comment := CPToUTF8(Comment);
 Inc(i);

 if not First then
  begin
   tl := ButTools.Is_On;
   mx := ButMixer.Is_On;
   ls := ButList.Is_On;
   pa := ButPause.Is_Pushed;
   pl := ButPlay.Is_Pushed;
   lp := ButLoop.Is_Pushed;
   l1 := Led_AY.State;
   l2 := Led_YM.State;
   l3 := Led_Stereo.State;
   BmpFree;
   MoveVol.Bmp1.Free;
   MoveVol.Bmp2.Free;
   MoveVol.Bmps := False;
   MoveProgr.Bmp1.Free;
   MoveProgr.Bmp2.Free;
   MoveProgr.Bmps := False;
   SetMainBmp(@Buffer[i],Original_Size - i);
   CopyBmpSources;
   ButTools.Is_On := tl;
   ButTools.Is_Pushed := tl;
   ButMixer.Is_On := mx;
   ButMixer.Is_Pushed := mx;
   ButList.Is_On := ls;
   ButList.Is_Pushed := ls;
   ButPause.Is_Pushed := pa;
   ButPlay.Is_Pushed := pl;
   ButLoop.Is_Pushed := lp;
   ButLoop.Is_On := lp;
   Led_AY.State := l1;
   Led_YM.State := l2;
   Led_Stereo.State := l3;
   if ButTools.Is_On then
    begin
     Form6.Edit1.Text := Author;
     Form6.Edit2.Text := Comment;
     Form6.Edit3.Text := FName;
    end;
  end
 else
  SetMainBmp(@Buffer[i],Original_Size - i);
except
 ShowException(ExceptObject, ExceptAddr);
 exit;
end;
SkinAuthor := Author;
SkinComment := Comment;
SkinFileName := FName;
Result := True;
if FileAvailable then
 begin
  RedrawTime;
  RedrawScroll;
 end;
Refresh;
end;

procedure TForm1.SetMainBmp(p:pointer;size:integer);
var
 Stream:TStream;
 Bitmap:TBitmap;
begin
Stream := TMemoryStream.Create;
Stream.Write(p^,size);
Stream.Position := 0;
Bitmap := TBitmap.Create;
Bitmap.LoadFromStream(Stream);
Stream.Free;
BMP_DBuffer.Canvas.CopyMode:=cmSrcCopy;
BMP_DBuffer.Canvas.CopyRect(Rect(0,0,MWWidth,MWHeight),Bitmap.Canvas,Rect(0,0,MWWidth,MWHeight));
ButPlay := TButtZone.Create(@ButPlay,119,96,35,27,RgnPlay,
                            Bitmap,119,96,119,122,@PlayClick);
ButPrev := TButtZone.Create(@ButPrev,80,96,35,27,RgnBack,
                            Bitmap,80,96,80,122,@ButPrevClick);
ButNext := TButtZone.Create(@ButNext,235,96,35,27,RgnNext,
                            Bitmap,235,96,235,122,@ButNextClick);
ButOpen := TButtZone.Create(@ButOpen,275,96,35,27,RgnOpen,
                            Bitmap,275,96,275,122,@ButOpenClick);
ButStop := TButtZone.Create(@ButStop,197,96,35,27,RgnStop,
                            Bitmap,197,96,197,122,@ButStopClick);
ButPause := TButtZone.Create(@ButPause,158,96,35,27,RgnPause,
                             Bitmap,158,96,158,122,@ButPauseClick);
ButLoop := TButtZone.Create(@ButLoop,62-10,110-10,21,21,RgnLoop,
                             Bitmap,62-10,110-10,358-21,110-7,@ButLoopClick);
ButMixer := TButtZone.Create(@ButMixer,318,21,26,26,RgnMixer,
                             Bitmap,318,21,26*2,124,@ButMixerClick);
ButList := TButtZone.Create(@ButList,310,77,26,26,RgnPList,
                            Bitmap,310,77,26,124,@ButListClick);
ButTools := TButtZone.Create(@ButTools,322,50,26,26,RgnTools,
                             Bitmap,322,50,0,124,@ButToolsClick);
ButMinimize := TButtZone.Create(@ButMinimize,282,6,16,16,RgnMin,
                                Bitmap,282,6,0,0,@ButMinClick);
ButClose := TButtZone.Create(@ButClose,304,6,16,16,RgnClose,
                             Bitmap,304,6,358-16,0,@ButCloseClick);
ButAbout := TButtZone.Create(@ButAbout,258,84,307-258,92-84,0,
                             Bitmap,258,84,0,123-(92-84),@ButAboutClick);
MoveVol.AddBitmaps(Bitmap,358-41,113,18,11,True);
MoveProgr.AddBitmaps(Bitmap,0,103,20,10,True);
Led_AY := TLedZone.Create(@Led_AY,99,26,144-99,33-26,
                          Bitmap,99,26,358-(144-99)-1,150-(33-26)-1);
Led_YM := TLedZone.Create(@Led_YM,144,26,190-144,33-26,
                          Bitmap,144,26,358-(190-144)-1,150-(33-26)*2-2);
Led_Stereo := TLedZone.Create(@Led_Stereo,190,26,234-190,33-26,
                              Bitmap,190,26,358-(234-190)-1,150-(33-26)*3-3);
Bitmap.Free;
end;

procedure TForm1.BmpFree;
var
 pppp,pppp1:PButtZone;
 ppp,ppp1:PLedZone;
begin
if ButtZoneRoot <> nil then
 begin
  pppp := ButtZoneRoot;
  ButtZoneRoot := nil;
  repeat
   pppp1 := pppp^.Next;
   pppp^.Free;
   pppp := pppp1;
  until pppp = nil;
 end;
if LedZoneRoot <> nil then
 begin
  ppp := LedZoneRoot;
  LedZoneRoot := nil;
  repeat
   ppp1 := ppp^.Next;
   ppp^.Free;
   ppp := ppp1;
  until ppp = nil;
 end;
end;

procedure TForm1.CopyBmpSources;
begin
BMP_Sources.Canvas.CopyMode:=cmSrcCopy;
BMP_Sources.Canvas.CopyRect(Bounds(spa_src,0,spa_width,spa_height),BMP_DBuffer.Canvas,Bounds(spa_x,spa_y,spa_width,spa_height));
BMP_Sources.Canvas.CopyRect(Bounds(amp_src,0,amp_width,amp_height),BMP_DBuffer.Canvas,Bounds(amp_x,amp_y,amp_width,amp_height));
BMP_Sources.Canvas.CopyRect(Bounds(time_src,0,time_width,time_height),BMP_DBuffer.Canvas,Bounds(time_x,time_y,time_width,time_height));
BMP_Sources.Canvas.CopyRect(Bounds(scr_src,0,scr_width,scr_height),BMP_DBuffer.Canvas,Bounds(scr_x,scr_y,scr_width,scr_height));
BMP_Time.Canvas.CopyMode:=cmSrcCopy;
BMP_Time.Canvas.CopyRect(Rect(0,0,time_width,time_height),BMP_Sources.Canvas,Bounds(time_src,0,time_width,time_height));
BMP_Scroll.Canvas.CopyMode:=cmSrcCopy;
BMP_Scroll.Canvas.CopyRect(Rect(0,0,scr_width,scr_height),BMP_Sources.Canvas,Bounds(scr_src,0,scr_width,scr_height));
end;


procedure TForm1.FormDropFiles(Sender: TObject; const FileNames: array of String
  );
var
 nFiles,i:integer;
// Filename:WideString;
 r1,r2:boolean;
begin
StopAndFreeAll;
ClearPlayList;
r1 := RecurseDirs; r2 := RecurseDirsDoDetect;
RecurseDirs := True; RecurseDirsDoDetect := False;
 try
  nFiles := Length(FileNames);
  for i := 0 to nFiles - 1 do
   begin
//    Filename := UTF8Decode(FileNames[i]);
    if not DirectoryExistsUTF8(FileNames[i]) then
     Form3.Add_File({ValidateFileName(PWideChar(Filename))}FileNames[i],RecurseDirsDoDetect)
    else
     Form3.SearchFilesInFolder({ValidateDirName(PWideChar(Filename))}FileNames[i],-1);
   end;
 finally
  RecurseDirs := r1; RecurseDirsDoDetect := r2;
  CalculateTotalTime(False);
  CreatePlayOrder;
 end;
PlayItem(0,0);
end;

procedure TForm1.FIDO_SaveStatus(Status:FIDO_Status);
var
 f:TextFile;
 s:string;

 procedure KillFile;
 begin
 try
  if FileExistsUTF8(FIDO_Descriptor_FileName) then
   begin
    DeleteFileUTF8(FIDO_Descriptor_FileName);
//    FIDO_Descriptor_String := s;
   end
 except
 end;
 end;

var
{$IFDEF Windows}
 i:integer;
{$ENDIF Windows}
 FN:string;
begin
if not FIDO_Descriptor_Enabled or Uninstall then exit;
case Status of
FIDO_Nothing:
        begin
         if FIDO_Descriptor_KillOnNothing then
          begin
           KillFile;
           exit;
          end;
         s := FIDO_Descriptor_Prefix + FIDO_Descriptor_Nothing
        end;
FIDO_Exit:
        begin
         if FIDO_Descriptor_KillOnExit then
          begin
           KillFile;
           exit;
          end;
         s := FIDO_Descriptor_Prefix + FIDO_Descriptor_Nothing
        end;
else    s := FIDO_Descriptor_Prefix +
                CurItem.PLStr + FIDO_Descriptor_Suffix
end;
if s <> FIDO_Descriptor_String then
 begin
  FIDO_Descriptor_String := s;
  s := UTF8ToSys(s);
  {$IFDEF Windows}
  for i := 1 to Length(s) do
   case s[i] of
    #205: s[i] := 'H';
    #240: s[i] := 'p'
   end;
  if not FIDO_Descriptor_WinEnc then
   AnsiToOemBuff(@s[1], @s[1], Length(s));
  {$ELSE Windows}
  { TODO: AnsiToOem platform independent}
   s := FIDO_Descriptor_String; //temporary UTF8
  {$ENDIF Windows}
  try
    {$IFDEF Windows}
    FN := GetTempFileName(FIDO_Descriptor_FileName);
    {$ELSE Windows}
    FN := FIDO_Descriptor_FileName;
    {$ENDIF Windows}
   AssignFile(f,FN);
   Rewrite(f);
   try
    Write(f,s);
   finally
    CloseFile(f);
    {$IFDEF Windows}
    RenameTempFile(FN,FIDO_Descriptor_FileName);
    {$ENDIF Windows}
   end;
  except;
  end;
 end;
end;

procedure TForm1.JumpToTime;

function TimeValid(stime:string;var time:integer):boolean;
var
 temp,t1:integer;
begin
Result := True;
Val(stime,time,temp);
if temp = 0 then exit;
if (temp > 1) and (temp < Length(stime)) and (stime[temp] = ':') then
 begin
  Val(Copy(stime,temp + 1,Length(stime) - temp),time,t1);
  if t1 = 0 then
   begin
    Val(Copy(stime,1,temp - 1),t1,temp);
    if temp = 0 then
     begin
      inc(time,t1*60);
      exit
     end;
   end;
 end;
Result := False
end;

var
 time:integer;
begin
if not IsPlaying then exit;
if Paused then exit;
with TForm8.Create(Self) do
 try
  Edit1.Text := TimeSToStr(round(CurrTime_Rasch / 1000));
  if Russian_Interface then
   begin
    Caption := 'Перейти на время';
    Label1.Caption := 'минуты:секунды';
    Label2.Caption := 'Длина трека ' + TimeSToStr(round(Time_ms / 1000));
    Button1.Caption := 'Перейти';
    Button2.Caption := 'Отмена';
   end
  else
   begin
    Caption := 'Jump to time';
    Label1.Caption := 'Minutes:Seconds';
    Label2.Caption := 'Track length ' + TimeSToStr(round(Time_ms / 1000));
    Button1.Caption := 'Jump';
    Button2.Caption := 'Cancel';
   end;
  if ShowModal = mrOK then
   if TimeValid(Edit1.Text,time) then
    Rewind(time*1000,Time_ms);
 finally
  Free;
 end;
end;

procedure TForm1.CallHelp;
var
 f:string;
begin
(*
if Russian_Interface then
 WinHelp(Handle,
  PChar(ValidateDirName(UTF8Decode(ExtractFilePath(GetProcessFileName)))+'Ay_Rus.hlp'),
                  HELP_FINDER, 0)
else
 WinHelp(Handle,
  PChar(ValidateDirName(UTF8Decode(ExtractFilePath(GetProcessFileName)))+'Ay_Eng.hlp'),
                  HELP_FINDER, 0);
*)
f := ExtractFilePath(GetProcessFileName);
if Russian_Interface then
 f := f+'Ay_Rus.chm'
else
 f := f+'Ay_Eng.chm';
if not OpenDocument(f) then
 ShowMessage('Can''t open '+ f);
end;

procedure TForm1.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
VolDown;
end;

procedure TForm1.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
VolUp;
end;

{$IFDEF Windows}
procedure TForm1.WMFINALIZEWO(var Msg: TMsg);
begin
WOThreadFinalization;
RestoreControls;
PostMessage(Form1.Handle,WM_PLAYNEXTITEM,0,0);
end;
{$ENDIF Windows}

procedure TForm1.FormDeactivate(Sender: TObject);
var
 p:PSensZone;
 p1:PMoveZone;
 p2:PButtZone;
begin
//{$IFDEF Windows}
//  ClipCursor(nil);
//{$ENDIF Windows}
  p := SensZoneRoot;
  while p <> nil do
   begin
    p^.Clicked := False;
    p := p^.Next;
   end;
  p2 := ButtZoneRoot;
  while p2 <> nil do
   begin
    if (p2^.Clicked = 1) and not p2^.Is_On then
     p2^.UnPush;
    p2^.Clicked := 0;
    p2 := p2^.Next;
   end;
  p1 := MoveZoneRoot;
  while p1 <> nil do
   begin
    p1^.Clicked := False;
    p1 := p1^.Next;
   end;
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
  MainWinRepaint;
end;

{$IFNDEF Windows}
procedure TForm1.DoSetRgn(Sender: TObject);
begin
  Timer1.Enabled := False;
  SetWindowRgn(Handle,MyFormRgn,True);
end;
{$ENDIF Windows}

procedure TForm1.FormShow(Sender: TObject);
begin
{$IFNDEF Windows} //GTK can set region after OnShow only :(
Timer1 := TTimer.Create(Self);
Timer1.Interval:=1;
Timer1.OnTimer:=@DoSetRgn;
{$ENDIF Windows}
end;

procedure StopPlaying;
begin
try
case CurFileType of
BASSFileMin..BASSFileMax{$IFNDEF Windows}, MinAYChipFile..MaxAYChipFile{$ENDIF Windows}:
 PlayFreeBASS;
{$IFDEF Windows}
CDAFile:
 StopCDDevice(CurCDNum);
MIDFileMin..MIDFileMax:
 StopMIDIThread;
else
 StopWOThread;
 {$ENDIF Windows}
end;
finally
IsPlaying := False;
{$IFDEF Windows}
Reseted := 0;
{$ENDIF Windows}
Paused := False;
RestoreControls;
end;
end;

{$IFDEF Windows}
procedure GetSysVolume;
var
 MCD:TMIXERCONTROLDETAILS;
 MDU:array of TMIXERCONTROLDETAILS_UNSIGNED;
 i,ps:integer;
begin
if SysVolumeParams.MixerID = -1 then exit;
SetLength(MDU,SysVolumeParams.Chans);
FillChar(MCD,sizeof(TMIXERCONTROLDETAILS),0);
MCD.cbStruct := sizeof(TMIXERCONTROLDETAILS);
MCD.dwControlID := SysVolumeParams.ControlID;
MCD.cChannels := SysVolumeParams.Chans;
MCD.cbDetails := sizeof(TMIXERCONTROLDETAILS_UNSIGNED);
MCD.paDetails := @MDU[0];
if mixerGetControlDetails(SysVolumeParams.MixerID,@MCD,
                               MIXER_GETCONTROLDETAILSF_VALUE or
                               MIXER_OBJECTF_MIXER) = MMSYSERR_NOERROR then
 begin
  ps := 0;
  for i := 0 to SysVolumeParams.Chans - 1 do
   if SysVolumeParams.Vals[i].dwValue <> MDU[i].dwValue then
    begin
     ps := 1;
     break
    end;
  if ps <> 0 then
   begin
    for i := 0 to SysVolumeParams.Chans - 1 do
     SysVolumeParams.Vals[i].dwValue := MDU[i].dwValue;
    ps := SysVolumeParams.Pos;
    SysVolumeParams.Pos := 0;
    for i := 0 to SysVolumeParams.Chans - 1 do
     if DWORD(SysVolumeParams.Pos) < MDU[i].dwValue then
      SysVolumeParams.Pos := MDU[i].dwValue;
//    if (SysVolumeParams.Pos > 0) and (ps = SysVolumeParams.Pos) then
     for i := 0 to SysVolumeParams.Chans - 1 do
      SysVolumeParams.Balans[i].dwValue := MDU[i].dwValue
   end
 end
else
 SysVolumeParams.Pos := SysVolumeParams.Max;
if VolLinear then
 VolumeCtrl := round((SysVolumeParams.Pos - SysVolumeParams.Min) /
                    (SysVolumeParams.Max - SysVolumeParams.Min) * VolumeCtrlMax)
else
 VolumeCtrl := round(
            ln((SysVolumeParams.Pos - SysVolumeParams.Min) /
                        (SysVolumeParams.Max - SysVolumeParams.Min)
//                         + 1) / ln(2)  //closer to linear version
           * (exp(1) - 1) + 1)
                      * VolumeCtrlMax);
RedrawVolume;
end;
{$ENDIF Windows}

procedure SetSysVolume;
var
 v:Single;
{$IFDEF Windows}
 MCD:TMIXERCONTROLDETAILS;
 i,ps:integer;
{$ENDIF Windows}
begin
if VolLinear then
 v := VolumeCtrl / VolumeCtrlMax
else
 //(exp(VolumeCtrl / VolumeCtrlMax * ln(2)) - 1) //closer to linear version
 v := (exp((VolumeCtrl / VolumeCtrlMax)) - 1) / (exp(1) - 1);
{$IFDEF Windows}
if SysVolumeParams.MixerID = -1 then exit;
ps := SysVolumeParams.Min;
for i := 0 to SysVolumeParams.Chans - 1 do
 if DWORD(ps) < SysVolumeParams.Balans[i].dwValue then
  ps := SysVolumeParams.Balans[i].dwValue;
SysVolumeParams.Pos := SysVolumeParams.Min +
                round(v * (SysVolumeParams.Max - SysVolumeParams.Min));
for i := 0 to SysVolumeParams.Chans - 1 do
 if ps > 0 then
  SysVolumeParams.Vals[i].dwValue := SysVolumeParams.Min +
   round((SysVolumeParams.Pos - SysVolumeParams.Min)/ps*
             (SysVolumeParams.Balans[i].dwValue - DWORD(SysVolumeParams.Min)))
 else
  SysVolumeParams.Vals[i].dwValue := SysVolumeParams.Pos;
FillChar(MCD,sizeof(TMIXERCONTROLDETAILS),0);
MCD.cbStruct := sizeof(TMIXERCONTROLDETAILS);
MCD.dwControlID := SysVolumeParams.ControlID;
MCD.cChannels := SysVolumeParams.Chans;
MCD.cbDetails := sizeof(TMIXERCONTROLDETAILS_UNSIGNED);
MCD.paDetails := @SysVolumeParams.Vals[0];
mixerSetControlDetails(SysVolumeParams.MixerID,@MCD,
                                        MIXER_SETCONTROLDETAILSF_VALUE or
                                        MIXER_OBJECTF_MIXER);
{$ELSE Windows}
SetVolumeBASS(v);
{$ENDIF Windows}
RedrawVolume;
end;

procedure RedrawVolume;
begin
if VolumeCtrl = MoveVol.PosX then exit;
MoveVol.HideBmp;
OffsetRgn(MoveVol.RgnHandle,VolumeCtrl - MoveVol.PosX,0);
MoveVol.PosX := VolumeCtrl;
MoveVol.Redraw(False);
end;

procedure TForm1.WMPLAYERROR(var Msg: TMsg);
begin
ButStopClick(Self);
end;

procedure StopAndFreeAll;
begin
try
 StopPlaying
finally
 FreeBASS;
 UnloadBASS;
{$IFDEF Windows}
 try
  FreeAllCD
 except
 end;
{$ENDIF Windows}
end;
end;

procedure TForm1.SaveParams;

 procedure SaveDW(Nm:PChar; const Vl:integer);
 begin
 OptionsWrite(Nm,IntToStr(Vl));
 end;

 procedure SaveStr(Nm:PChar; const Vl:string);
 begin
 OptionsWrite(Nm,Vl);
 end;

begin
if Uninstall then exit;

if OptionsInit(True) then
try
 SaveDW('SampleRate',SampleRate);
 SaveDW('SampleBit',SampleBit);
 SaveDW('OutChansMono',Ord(Form2.RadioButton14.Checked));
 SaveDW('OutChansList',Ord(Form2.CheckBox8.Checked));
 SaveDW('BufLen_ms',BufLen_ms);
 SaveDW('NumberOfBuffers',NumberOfBuffers);
 SaveDW('Chip',Ord(not Form2.RadioButton1.Checked) + 1);
 SaveDW('ChipList',Ord(Form2.CheckBox2.Checked));
 SaveDW('FrqZ80',FrqZ80);
 SaveDW('FrqAY',Form2.FrqAYTemp);
 SaveDW('FrqAYList',Ord(Form2.CheckBox3.Checked));
 SaveDW('FrqPl',Form2.FrqPlTemp);
 SaveDW('FrqPlList',Ord(Form2.CheckBox9.Checked));
 SaveDW('IntOffset',IntOffset);
 SaveDW('MaxTStates',MaxTStates);
 SaveDW('VisAmpls',Ord(IndicatorChecked));
 SaveDW('VisSpectrum',Ord(SpectrumChecked));
 SaveDW('VisScroll',Ord(Do_Scroll));
 SaveDW('VisPeriod',VisTimerPeriod);
 SaveDW('Russian',Ord(Russian_Interface));
 SaveDW('Loop',Ord(Do_Loop));
 SaveDW('TrayMode',TrayMode);
 SaveDW('TimeMode',TimeMode);
 SaveStr('Skin',SkinFileName);
 SaveDW('MFPTimerMode',MFPTimerMode);
 SaveDW('MFPTimerFrq',MFPTimerFrq);
 SaveDW('AutoSaveDefDir',Ord(AutoSaveDefDir));
 SaveDW('AutoSaveWindowsPos',Ord(AutoSaveWindowsPos));
 SaveDW('AutoSaveVolumePos',Ord(AutoSaveVolumePos));
 SaveDW('BeeperMax',BeeperMax);
 SaveDW('ChanAL',Form2.TrackBar1.Position);
 SaveDW('ChanAR',Form2.TrackBar2.Position);
 SaveDW('ChanBL',Form2.TrackBar3.Position);
 SaveDW('ChanBR',Form2.TrackBar4.Position);
 SaveDW('ChanCL',Form2.TrackBar5.Position);
 SaveDW('ChanCR',Form2.TrackBar6.Position);
 SaveDW('ChansList',Ord(Form2.CheckBox1.Checked));
 SaveStr('EditorPath',VTPath);
 SaveDW('FIDO_Descriptor_Enabled',Ord(FIDO_Descriptor_Enabled));
 SaveDW('FIDO_Descriptor_KillOnExit',Ord(FIDO_Descriptor_KillOnExit));
 SaveDW('FIDO_Descriptor_KillOnNothing',Ord(FIDO_Descriptor_KillOnNothing));
 SaveDW('FIDO_Descriptor_WinEnc',Ord(FIDO_Descriptor_WinEnc));
 SaveStr('FIDO_Descriptor_FileName',FIDO_Descriptor_FileName);
 SaveStr('FIDO_Descriptor_Nothing',FIDO_Descriptor_Nothing);
 SaveStr('FIDO_Descriptor_Suffix',FIDO_Descriptor_Suffix);
 SaveStr('FIDO_Descriptor_Prefix',FIDO_Descriptor_Prefix);
 SaveStr('DefaultCodePage',CodePageDef);
 SaveDW('FilterQuality',FilterQuality);
 SaveDW('PreAmp',PreAmp);
 SaveDW('VolLinear',Ord(VolLinear));
 if AutoSaveVolumePos then
  SaveDW('Volume',VolumeCtrl);
 if AutoSaveWindowsPos then
  begin
   SaveDW('MainX',Form1.Left);
   SaveDW('MainY',Form1.Top);
   SaveDW('ListX',Form3.Left);
   SaveDW('ListY',Form3.Top);
   SaveDW('ListW',Form3.Width);
   SaveDW('ListH',Form3.Height);
   SaveDW('ListVis',Ord(Form3.Visible));
   SaveDW('MixerX',Form2.Left);
   SaveDW('MixerY',Form2.Top);
   if ButTools.Is_On then
    begin
     ToolsY := Form6.Top;
     ToolsX := Form6.Left
    end;
   SaveDW('ToolsX',ToolsX);
   SaveDW('ToolsY',ToolsY)
  end;
 SaveDW('ListItem',PlayingItem);
 SaveDW('AppIcon',AppIconNumber);
 SaveDW('TrayIcon',TrayIconNumber);
 SaveDW('MenuIcon',MenuIconNumber);
 SaveDW('MusIcon',MusIconNumber);
 SaveDW('SkinIcon',SkinIconNumber);
 SaveDW('ListIcon',ListIconNumber);
 SaveDW('BASSIcon',BASSIconNumber);
 SaveStr('SkinDirectory',SkinDirectory);
 SaveDW('PlayListDirection',Direction);
 SaveDW('PlayListLoop',Ord(ListLooped));
 SaveDW('PLColorBkSel',PLColorBkSel);
 SaveDW('PLColorBkPl',PLColorBkPl);
 SaveDW('PLColorBk',PLColorBk);
 SaveDW('PLColorPlSel',PLColorPlSel);
 SaveDW('PLColorPl',PLColorPl);
 SaveDW('PLColorSel',PLColorSel);
 SaveDW('PLColor',PLColor);
 SaveDW('PLColorErrSel',PLColorErrSel);
 SaveDW('PLColorErr',PLColorErr);
{$IFDEF Windows}
 SaveDW('WODevice',WODevice);
 SaveStr('WODeviceName',Form2.cbWODevice.Items[integer(WODevice) + 1]);
 SaveDW('MIDIDevice',MIDIDevice);
 SaveStr('MIDIDeviceName',Form2.cbMODevice.Items[integer(MIDIDevice) + 1]);
 SaveDW('MIDISeekToFirstNote',Ord(MIDISeekToFirstNote));
 SaveDW('Priority',Priority);
{$ENDIF Windows}
 SaveDW('BASSFFTType',BASSFFTType);
 SaveDW('BASSFFTNoWin',BASSFFTNoWin);
 SaveDW('BASSFFTRemDC',BASSFFTRemDC);
 SaveDW('BASSAmpMin',round(BASSAmpMin * 10000));
 SaveStr('BASSNetAgent',BASSNetAgent);
 SaveDW('BASSNetUseProxy',Ord(BASSNetUseProxy));
 SaveStr('BASSNetProxy',BASSNetProxy);
{$IFDEF Windows}
 SaveDW('VolMixerNumber',SysVolumeParams.MixerNumber);
 SaveDW('VolDestNumber',SysVolumeParams.DestNumber);
 SaveDW('VolCtrlNumber',SysVolumeParams.CtrlNumber);
 SaveDW('VolCtrlChans',SysVolumeParams.Chans);
 {$ENDIF Windows}
 if AutoSaveDefDir then SaveDefaultDir2;
finally
 OptionsDone;
end;
end;

procedure AdjustFormOnDesktop(Frm:TForm);
var
 i:integer;
begin
if Frm.Left >= Screen.DesktopLeft+Screen.DesktopWidth then
 begin
  i := Screen.DesktopLeft+Screen.DesktopWidth-Frm.Width; if i < Screen.DesktopLeft then i := Screen.DesktopLeft;
  Frm.Left := i;
 end
else if Frm.Left+Frm.Width < Screen.DesktopLeft then
 Frm.Left := Screen.DesktopLeft;
if Frm.Top >= Screen.DesktopTop+Screen.DesktopHeight then
 begin
  i := Screen.DesktopTop+Screen.DesktopHeight-Frm.Height; if i < Screen.DesktopTop then i := Screen.DesktopTop;
  Frm.Top := i;
 end
else if Frm.Top+Frm.Height < Screen.DesktopTop then
 Frm.Top := Screen.DesktopTop;
end;

procedure TForm1.CommandLineAndRegCheck;
var
 i,v,v1:integer;
 dir:string;
 CanReadOptions:boolean;
{$IFDEF Windows}
 Mixers:TSysMixers;
{$ENDIF Windows}

 function GetDW(Nm:PChar; var Vl:integer):boolean;
 var
  s:string;
 begin
 if OptionsRead(Nm,s) then
  try
   Vl := StrToInt(s);
   Result := True;
  except
   Result := False;
  end;
 end;

 function GetStr(Nm:PChar; var Vl:string):boolean;
 begin
 Result := OptionsRead(Nm,Vl);
 end;

begin
ClearParams;
CanReadOptions := OptionsInit(False);
SetDefault;
AppIconNumber := -1; SelectAppIcon(0);

{$IFDEF Windows}
if (integer(GetVersion) < 0) then //Win9x or Win32s
 SetPriority(HIGH_PRIORITY_CLASS);
{$ENDIF Windows}
try
try
if CanReadOptions then
 begin
  if GetDW('SampleRate',v) then Set_Sample_Rate2(v);
  if GetDW('SampleBit',v) then Set_Sample_Bit2(v);
  if GetDW('OutChansMono',v) then Set_Stereo2(v);
  if GetDW('OutChansList',v) then Form2.CheckBox8.Checked := v <> 0;
  if GetDW('BufLen_ms',v) then Set_BufLen_ms2(v);
  if GetDW('NumberOfBuffers',v) then Set_NumberOfBuffers2(v);
  if GetDW('Chip',v) then Set_Chip2(ChTypes(v));
  if GetDW('ChipList',v) then Form2.CheckBox2.Checked := v <> 0;
  if GetDW('FrqZ80',v) then Set_Z80_Frq2(v);
  if GetDW('FrqAY',v) then Set_Chip_Frq2(v);
  if GetDW('FrqAYList',v) then Form2.CheckBox3.Checked := v <> 0;
  if GetDW('FrqPl',v) then Set_Player_Frq2(v);
  if GetDW('FrqPlList',v) then Form2.CheckBox9.Checked := v <> 0;
  if GetDW('MaxTStates',v) then Set_N_Tact2(v);
  if GetDW('IntOffset',v) then Set_IntOffset2(v);
  if GetDW('VisAmpls',v) then IndicatorChecked := v <> 0;
  if GetDW('VisSpectrum',v) then SpectrumChecked := v <> 0;
  if GetDW('VisScroll',v) then Do_Scroll := v <> 0;
  if GetDW('VisPeriod',v) then SetVisTimerPeriod(v);
  if GetDW('FilterQuality',v) then SetFilter2(v);
  if GetDW('Russian',v) then Set_Language2(v <> 0);
  if GetDW('Loop',v) then Set_Loop2(v <> 0);
  if GetDW('TimeMode',v) then if v in [0..2] then TimeMode := v;
  SkinDirectory := '';
  if GetStr('Skin',dir) then if dir <> '' then
   if LoadSkin(dir,False) then SkinDirectory := ExtractFileDir(dir);
  if GetStr('SkinDirectory',dir) then if dir <> '' then
   SkinDirectory := dir;
  v1 := MFPTimerMode;
  if GetDW('MFPTimerMode',v) then v1 := v;
  if v1 = 0 then
   Set_MFP_Frq2(0,0)
  else
   begin
    v1 := MFPTimerFrq;
    if GetDW('MFPTimerFrq',v) then v1 := v;
    Set_MFP_Frq2(1,v1)
   end;
  if GetDW('AutoSaveDefDir',v) then SetAutoSaveDefDir2(v <> 0);
  if GetDW('AutoSaveWindowsPos',v) then SetAutoSaveWindowsPos2(v <> 0);
  if GetDW('AutoSaveVolumePos',v) then SetAutoSaveVolumePos2(v <> 0);
  if GetDW('ChanAL',v) then SetChan2(v,0);
  if GetDW('ChanAR',v) then SetChan2(v,1);
  if GetDW('ChanBL',v) then SetChan2(v,2);
  if GetDW('ChanBR',v) then SetChan2(v,3);
  if GetDW('ChanCL',v) then SetChan2(v,4);
  if GetDW('ChanCR',v) then SetChan2(v,5);
  if GetDW('BeeperMax',v) then SetChan2(v,6);
  if GetDW('PreAmp',v) then SetChan2(v,7);
  if GetDW('ChansList',v) then Form2.CheckBox1.Checked := v <> 0;
  if GetStr('EditorPath',dir) then VTPath := dir;
  if GetDW('FIDO_Descriptor_Enabled',v) then FIDO_Descriptor_Enabled := v <> 0;
  if GetDW('FIDO_Descriptor_KillOnExit',v) then FIDO_Descriptor_KillOnExit := v <> 0;
  if GetDW('FIDO_Descriptor_KillOnNothing',v) then FIDO_Descriptor_KillOnNothing := v <> 0;
  if GetDW('FIDO_Descriptor_WinEnc',v) then FIDO_Descriptor_WinEnc := v <> 0;
  if GetStr('FIDO_Descriptor_FileName',dir) then FIDO_Descriptor_FileName := dir;
  if GetStr('FIDO_Descriptor_Nothing',dir) then FIDO_Descriptor_Nothing := dir;
  if GetStr('FIDO_Descriptor_Suffix',dir) then FIDO_Descriptor_Suffix := dir;
  if GetStr('FIDO_Descriptor_Prefix',dir) then FIDO_Descriptor_Prefix := dir;
  if GetStr('DefaultCodePage',dir) then CodePageDef := dir;
  if GetDW('PlayListDirection',v) then if v in [0..3] then Form3.SetDirection(v);
  if GetDW('PlayListLoop',v) then
   begin
    ListLooped := v <> 0;
    Form3.LoopListButton.Down := ListLooped
   end;
  if GetDW('PLColorBkSel',v) then PLColorBkSel := v;
  if GetDW('PLColorBkPl',v) then PLColorBkPl := v;
  if GetDW('PLColorBk',v) then PLColorBk := v;
  if GetDW('PLColorPlSel',v) then PLColorPlSel := v;
  if GetDW('PLColorPl',v) then PLColorPl := v;
  if GetDW('PLColorSel',v) then PLColorSel := v;
  if GetDW('PLColor',v) then PLColor := v;
  if GetDW('PLColorErrSel',v) then PLColorErrSel := v;
  if GetDW('PLColorErr',v) then PLColorErr := v;

{$IFDEF Windows}
  if not GetStr('WODeviceName',dir) then dir := '';
  if GetDW('WODevice',v) then Set_WODevice2(v,dir);
  if not GetStr('MIDIDeviceName',dir) then dir := '';
  if GetDW('MIDIDevice',v) then Set_MIDIDevice2(v,dir);
  if GetDW('MIDISeekToFirstNote',v) then
   MIDISeekToFirstNote := v <> 0;
{$ENDIF Windows}
  if GetDW('BASSFFTType',v) then
   if (DWORD(v) >= BASS_DATA_FFT256) and (DWORD(v) <= BASS_DATA_FFT32768) then
    BASSFFTType := v;
  if GetDW('BASSFFTNoWin',v) then
   if v in [0,BASS_DATA_FFT_NOWINDOW] then
    BASSFFTNoWin := v;
  if GetDW('BASSFFTRemDC',v) then
   if v in [0,BASS_DATA_FFT_REMOVEDC] then
    BASSFFTRemDC := v;
  if GetDW('BASSAmpMin',v) then
   if (v >= 1) and (v <= 200) then
    BASSAmpMin := v / 10000;
  if GetStr('BASSNetAgent',dir) then BASSNetAgent := dir;
  if GetDW('BASSNetUseProxy',v) then
   BASSNetUseProxy := v <> 0;
  if GetStr('BASSNetProxy',dir) then BASSNetProxy := dir;

  if GetDW('VolLinear',v) then VolLinear := v <> 0;
  {$IFDEF Windows}
  if GetDW('VolMixerNumber',v) then SysVolumeParams.MixerNumber := v;
  if GetDW('VolDestNumber',v) then SysVolumeParams.DestNumber := v;
  if GetDW('VolCtrlNumber',v) then SysVolumeParams.CtrlNumber := v;
  if GetDW('VolCtrlChans',v) then SysVolumeParams.Chans := v;
  GetSystemMixers(Mixers);
  if not SelectMixerControl(Mixers,SysVolumeParams.MixerNumber,
                                   SysVolumeParams.DestNumber,
                                   SysVolumeParams.CtrlNumber) then
   DetectVolumeCtrl2(Mixers);
  Mixers := nil;
  {$ENDIF Windows}
  if AutoSaveVolumePos then
   begin
    if GetDW('Volume',v) then
     if v < VolumeCtrlMax then
      begin
       VolumeCtrl := v;
       SetSysVolume;
      end
   end;
  {$IFDEF Windows}
  if GetDW('Priority',v) then SetPriority2(v);
  {$ENDIF Windows}
  DefaultDirectory := '';
  if GetStr('DefaultDirectory',dir) then DefaultDirectory := dir;
 end
{$IFDEF Windows}
else
 DetectVolumeCtrl;
{$ENDIF Windows}
finally
 Form2.SetMixerParams;
end;
LastTimeComLine := GetTickCount - CLFast;
try
 if ParamCount <> 0 then
  CommandLineInterpreter('"' + GetCurrentDirUTF8 + '" ' + GetCommandLine,True);
 for i := 0 to Length(AfterScan) - 1 do
  CommandLineInterpreter(AfterScan[i],True);
except
 ShowException(ExceptObject, ExceptAddr);
end;
AfterScan := nil;
if Length(PlayListItems) = 0 then
  if TryLoadDefaultPL then
   if CanReadOptions then
    if GetDW('ListItem',v) then
     if (v >= 0) and (v < Length(PlayListItems)) then
      PlayingItem := v;
CreatePlayOrder;
CalculateTotalTime(False);
dir := ExtractFileDir(GetProcessFileName);
if DefaultDirectory = '' then DefaultDirectory := dir;
if SetCurrentDirUTF8(DefaultDirectory) then Form1.OpenDialog1.InitialDir := DefaultDirectory;
if CanReadOptions then
 begin
  if AutoSaveWindowsPos then
   begin
    Position := poDesigned;
    if GetDW('MainX',v) then Left := v;
    if GetDW('MainY',v) then Top := v;
    AdjustFormOnDesktop(Form1);
    Form3.Position := poDesigned;
    if GetDW('ListX',v) then Form3.Left := v;
    if GetDW('ListY',v) then Form3.Top := v;
    if GetDW('ListW',v) then Form3.Width := v;
    if GetDW('ListH',v) then Form3.Height := v;
    AdjustFormOnDesktop(Form3);
    if GetDW('ListVis',v) then if v <> 0 then
     begin
      ButList.Switch_On;
      Form3.Visible := True;
     end;
    Form2.Position := poDesigned;
    if GetDW('MixerX',v) then Form2.Left := v;
    if GetDW('MixerY',v) then Form2.Top := v;
    AdjustFormOnDesktop(Form2);
    if GetDW('ToolsX',v) then ToolsX := v;
    if GetDW('ToolsY',v) then ToolsY := v;
   end;
  if GetDW('AppIcon',v) then SelectAppIcon(v);
  if GetDW('TrayIcon',v) then SelectTrayIcon(v);
  if GetDW('MenuIcon',v) then MenuIconNumber := v;
  if GetDW('MusIcon',v) then MusIconNumber := v;
  if GetDW('SkinIcon',v) then SkinIconNumber := v;
  if GetDW('ListIcon',v) then ListIconNumber := v;
  if GetDW('BASSIcon',v) then BASSIconNumber := v;
  if GetDW('TrayMode',v) then Set_TrayMode2(v);
 end;
finally
if CanReadOptions then OptionsDone;
end;
FIDO_SaveStatus(FIDO_Nothing);
if FileAvailable then
 PlayCurrent
else
 PlayItem(PlayingOrderItem,-1);
InitialScan := True;
end;

procedure TForm1.SetBuffers(len,num:integer);
begin
{$IFDEF Windows}
if WOThreadActive then exit;
{$ENDIF Windows}
if (num < 2) or (num > 10) then exit;
if (len < 5) or (len > 2000) then exit;
BufLen_ms := len;
NumberOfBuffers := num;
{$IFDEF Windows}
SetLength(waveOutBuffers,NumberOfBuffers);
{$ENDIF Windows}
BufferLength := round(BufLen_ms * SampleRate / 1000);
VisPosMax := round(BufferLength * NumberOfBuffers / VisStep) + 1;
VisTickMax := VisStep * VisPosMax;
SetLength(VisPoints,VisPosMax);
end;

{$IFDEF Windows}
procedure TForm1.Set_WODevice2(WOD:integer;NM:string);
var
 l,j:integer;
begin
if WOThreadActive or (WOD < -1) then exit;
l := Form2.cbWODevice.Items.Count; if WOD >= l - 1 then exit;
if (NM <> '') and (Form2.cbWODevice.Items[WOD + 1] <> NM) then
 begin
  j := 0;
  while (j < l) and (Form2.cbWODevice.Items[j] <> NM) do inc(j);
  if j < l then
   WOD := j - 1
  else
   WOD := -1;
 end;
if WODevice <> DWORD(WOD) then
 begin
  WODevice := WOD;
  Form2.cbWODevice.ItemIndex := WOD + 1
 end
end;

procedure TForm1.Set_MIDIDevice2(MD:integer;NM:string);
var
 l,j:integer;
begin
if MIDIThreadActive or (MD < -1) then exit;
l := Form2.cbMODevice.Items.Count; if MD >= l - 1 then exit;
if (NM <> '') and (Form2.cbMODevice.Items[MD + 1] <> NM) then
 begin
  j := 0;
  while (j < l) and (Form2.cbMODevice.Items[j] <> NM) do inc(j);
  if j < l then
   MD := j - 1
  else
   MD := -1;
 end;
if MIDIDevice <> DWORD(MD)  then
 begin
  MIDIDevice := MD;
  Form2.cbMODevice.ItemIndex := MD + 1;
 end;
end;
{$ENDIF Windows}

procedure TForm1.Set_BufLen_ms2(BL:integer);
begin
 if BL <> BufLen_ms then
  begin
   SetBuffers(BL,NumberOfBuffers);
   Form2.TrackBar8.Position := BufLen_ms;
  end;
end;

procedure TForm1.Set_NumberOfBuffers2(NB:integer);
begin
if NB <> NumberOfBuffers then
 begin
  SetBuffers(BufLen_ms,NB);
  Form2.TrackBar9.Position := NumberOfBuffers;
 end;
end;

procedure TForm1.Set_Chip2(Ch:ChTypes);
begin
if (Ch <> ChType) and (Ch in [AY_Chip,YM_Chip]) then
 begin
  ChType := Ch;
  Calculate_Level_Tables;
  Form2.CheckBox5.Checked := False;
  Form2.CheckBox4.Checked := False;
  case Ch of
  AY_Chip:
   begin
    Form2.RadioButton1.Checked := True;
    Led_AY.State := False;
    Led_YM.State := True;
    Led_AY.Redraw(False);
    Led_YM.Redraw(False);
    if IsPlaying then
     Form2.CheckBox4.Checked := True;
   end;
  YM_Chip:
   begin
    Form2.RadioButton2.Checked := True;
    Led_AY.State := True;
    Led_YM.State := False;
    Led_AY.Redraw(False);
    Led_YM.Redraw(False);
    if IsPlaying then
     Form2.CheckBox5.Checked := True;
   end;
  end;
 end;
end;

procedure TForm1.Set_IntOffset2(InO:integer);
begin
if (InO <> IntOffset) and (InO >= 0) and (InO < MaxTStates) then
 begin
  IntOffset := InO;
  Form2.FTact.Text := IntToStr(InO);
 end;
end;

procedure TForm1.Set_Language2(Rus:boolean);
begin
if Rus = Russian_Interface then exit;
SwapLan;
if ButTools.Is_On then
 if Russian_Interface then
  Form6.RadioButton6.Checked := True
 else
  Form6.RadioButton7.Checked := True;
end;

procedure TForm1.Set_Loop2(Lp:boolean);
begin
if Do_Loop = Lp then exit;
Do_Loop := Lp;
case Lp of
True:ButLoop.Switch_On;
else ButLoop.Switch_Off;
end;
end;

procedure TForm1.Set_TrayMode2(TM:integer);
begin
if (TrayMode = TM) or (DWORD(TM) > 2) then exit;
TrayMode := TM;
if TM = 2 then
 if WindowState <> wsMinimized then TM := 0 else TM := 1;
case TM of
0:
 begin
  RemoveTrayIcon;
  AddTaskbarButton;
 end;
1:
 begin
  AddTrayIcon;
  RemoveTaskbarButton;
 end;
end;
if ButTools.Is_On then
 case TrayMode of
 0:Form6.RadioButton8.Checked := True;
 1:Form6.RadioButton9.Checked := True;
 2:Form6.RadioButton10.Checked := True;
end;
end;

procedure TForm1.Set_MFP_Frq2(Md, Fr: integer);
begin
if (Md = MFPTimerMode) and (Fr = MFPTimerFrq) then exit;
Set_MFP_Frq(Md,Fr);
Form2.FrqMFPTemp := MFPTimerFrq;
Form2.Set_MFPFrqs
end;

procedure TForm1.SetAutoSaveDefDir2(ASD:boolean);
begin
AutoSaveDefDir := ASD;
if ButTools.Is_On then Form6.CheckBox38.Checked := ASD
end;

procedure TForm1.SetAutoSaveWindowsPos2(ASW:boolean);
begin
AutoSaveWindowsPos := ASW;
if ButTools.Is_On then Form6.CheckBox40.Checked := ASW
end;

procedure TForm1.SetAutoSaveVolumePos2(ASV:boolean);
begin
AutoSaveVolumePos := ASV;
Form2.CheckBox39.Checked := ASV;
end;

{$IFDEF Windows}
procedure TForm1.SetPriority2(NP:integer);
begin
if not (NP in
    [IDLE_PRIORITY_CLASS,NORMAL_PRIORITY_CLASS,HIGH_PRIORITY_CLASS]) then exit;
SetPriority(NP);
if ButTools.Is_On then
 case Priority of
 IDLE_PRIORITY_CLASS:   Form6.RadioButton3.Checked := True;
 NORMAL_PRIORITY_CLASS: Form6.RadioButton4.Checked := True;
 HIGH_PRIORITY_CLASS:   Form6.RadioButton5.Checked := True;
 end;
end;
{$ENDIF Windows}

procedure TForm1.SetChan2(u,i:integer);
begin
if DWORD(u) > 255 then exit;
with Form2 do
 case i of
 0:Change_Show(TrackBar1,Edit1,Edit12,u,Index_AL);
 1:Change_Show(TrackBar2,Edit2,Edit13,u,Index_AR);
 2:Change_Show(TrackBar3,Edit3,Edit14,u,Index_BL);
 3:Change_Show(TrackBar4,Edit4,Edit15,u,Index_BR);
 4:Change_Show(TrackBar5,Edit5,Edit16,u,Index_CL);
 5:Change_Show(TrackBar6,Edit6,Edit17,u,Index_CR);
 6:Change_Show2(Form2.TrackBar7,Form2.Edit20,u,BeeperMax);
 7:Change_Show2(TrackBar13,Edit30,u,PreAmp);
 end;
end;

procedure TForm1.CalcFiltKoefs;
const
 MaxF = 9200;
var
 i:integer;
 K,F,C,i2,Filt_M2:double;
 FKt:array of double;
 s:string;
begin
//Work range [0..MaxF)
//Range [MaxF..SampleRate / 2) is easy cut-off from 0 to -53 dB
//Cut-off range is [SampleRate / 2.. AY_Freq div 8 / 2] (-53 dB)
//for Ay_Freq = 1773400 Hz:
(*
Полезная область - 0..11083,75 Гц (10)
221675->44100 - 67 (коэффициентов)
221675->48000 - 57
221675->96000 - 20
221675->110000 - 17

Полезная область - 0..10076,14 (11)
221675->22050 - 771

Полезная область - 0..9236,46 (12)
221675->22050 - 409

Полезная область - 0..8525,96 (13)
221675->22050 - 293
*)
IsFilt := 0;
C := 22050; if SampleRate >= 44100 then
 begin
  C := SampleRate / 2;
  inc(IsFilt);
 end;
Filt_M := round(3.3/(C - MaxF) * (AY_Freq div 8));
if AY_Freq * Filt_M > 3500000 * 50 then //90% of usage for my Celeron 850 MHz
 begin
  Filt_M := round(3500000 * 50 / AY_Freq);
  IsFilt := 0;
 end;
C := Pi * (MaxF + C) / (AY_Freq div 8);
SetLength(FKt,Filt_M);
Filt_M2 := (Filt_M - 1) / 2;
K := 0;
for i := 0 to Filt_M - 1 do
 begin
  i2 := i - Filt_M2;
  if i2 = 0 then
   F := C
  else
   F := sin(C * i2) / i2 * (0.54 + 0.46 * cos(2 * Pi / Filt_M * i2));
  FKt[i] := F;
  K := K + F;
 end;
SetLength(Filt_K,Filt_M);
for i := 0 to Filt_M - 1 do
 Filt_K[i] := round(FKt[i] / K * $1000000);
s := 'FIR (' + IntToStr(Filt_M) + ' pts)';
if IsFilt = 0 then s := s + ' + averager';
Form2.Label13.Caption := s;
dec(Filt_M);
end;

procedure TForm1.SetFilter(FQ: integer);
begin
SuspendPlaying;
try
FilterQuality := FQ;
if (FQ = 0) or (SampleRate >= AY_Freq div 8) then
 begin
  IsFilt := -1;
  Filt_K := nil;
  Filt_XL := nil;
  Filt_XR := nil;
  Form2.Label13.Caption := 'Averager';
  exit;
 end;
CalcFiltKoefs;
SetLength(Filt_XL,Filt_M + 1);
SetLength(Filt_XR,Filt_M + 1);
FillChar(Filt_XL[0],(Filt_M + 1) * 4,0);
FillChar(Filt_XR[0],(Filt_M + 1) * 4,0);
Filt_I := 0;
finally
{$IFDEF Windows}
WOUnresetPlaying;
{$ENDIF Windows}
end;
end;

procedure TForm1.SetFilter2(FQ: integer);
begin
if (FilterQuality = FQ) or (FQ > 6) then exit;
SetFilter(FQ);
end;

procedure TForm1.SaveAllParams;
var
 Tmp:boolean;
begin
try
 Tmp := FIDO_Descriptor_Enabled;
 FIDO_Descriptor_Enabled := False;
 StopAndFreeAll;
 FIDO_Descriptor_Enabled := Tmp;
 FIDO_SaveStatus(FIDO_Exit);
 FreePlayingResourses;
 SaveParams;
except
 ShowException(ExceptObject, ExceptAddr);
end;
end;

procedure TForm1.DoCloseActions;
var
 p,p1:PSensZone;
 pp,pp1:PMoveZone;
begin
if CloseActionsDone then exit;
CloseActionsDone := True;
SaveAllParams;
TrySaveDefaultPL;
{$IFDEF Windows}
if SysVolumeParams.Opened then
 mixerClose(SysVolumeParams.MixerHandle);
{$ENDIF Windows}
RemoveTrayIcon;
BmpFree;

if SensZoneRoot <> nil then
 begin
  p := SensZoneRoot;
  SensZoneRoot := nil;
  repeat
   p1 := p^.Next;
   p^.Free;
   p := p1;
  until p = nil;
 end;
if MoveZoneRoot <> nil then
 begin
  pp := MoveZoneRoot;
  MoveZoneRoot := nil;
  repeat
   pp1 := pp^.Next;
   pp^.Free;
   pp := pp1;
  until pp = nil;
 end;

DeleteObject(RgnProgr);
DeleteObject(RgnVol);
DeleteObject(RgnClose);
DeleteObject(RgnMin);
DeleteObject(RgnTools);
DeleteObject(RgnPList);
DeleteObject(RgnMixer);
DeleteObject(RgnOpen);
DeleteObject(RgnNext);
DeleteObject(RgnStop);
DeleteObject(RgnPause);
DeleteObject(RgnPlay);
DeleteObject(RgnBack);
DeleteObject(RgnLoop);
DeleteObject(MyFormRgn);

VisTimer.Free;

BMP_Scroll.Free;
BMP_VScroll.Free;
BMP_Vis.Free;
BMP_Time.Free;
BMP_Sources.Free;
BMP_DBuffer.Free;
{$IFDEF Windows}
SetPriority(NORMAL_PRIORITY_CLASS);
{$ENDIF Windows}
end;

procedure TForm1.HideMinimize(var Msg: TMsg);
begin
//Application.Minimize;
DoMinimize;
end;

procedure TForm1.SelectAppIcon(n:integer);
begin
if AppIconNumber = n then exit;
AppIconNumber := n;
Application.Icon.LoadFromResourceName(hInstance,Format('ICON%.2u',[n]));
end;

procedure TForm1.SelectTrayIcon(n:integer);
begin
if TrayIconNumber = n then exit;
TrayIconNumber := n;
TrayIcon1.Icon.LoadFromResourceName(hInstance,Format('ICON%.2u',[n]));
end;

Procedure Ay_Emul_ShowException (Msg : ShortString);
begin
ShowMessage(SysToUTF8(Msg));
end;

procedure TForm1.Ay_Emul_ShowExceptionA(Sender : TObject; E : Exception);
begin
ShowMessage(SysToUTF8(E.Message));
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
DoCloseActions;
end;

procedure TForm1.AppEndSession(Sender: TObject);
begin
DoCloseActions;
end;

procedure AYVisualisation(smp:DWORD);
var
 CurVisPos:DWORD;
 T,E,A,i:integer;
 TE:boolean;
begin
VProgrPos := BaseSample + smp;
CurrTime_Rasch := trunc(VProgrPos / SampleRate * 1000);
CurVisPos := smp mod VisTickMax div VisStep;

if SpectrumChecked or IndicatorChecked then
 with VisPoints[CurVisPos] do
  begin
   if Calc = 0 then
    begin
     Calc := 1;
     for i := 0 to 1 do
      with VisPoints[CurVisPos].R[i] do
       begin
        case EnvT of
        8,12 : E := 28;
        10,14: E := 26;
        else
         begin
          E := AmpE - 1;
          if E < 0 then E := 0
         end
        end;
        T := TnA;
        if AmpA and 16 = 0 then
         AmpA := AmpA * 2
        else if not (EnvT in [8,10,12,14]) then
         AmpA := E
        else
         begin
          A := E;
          TE := Mix and 1 = 0;
          if (T <= 3) and TE then
           Dec(A,6)
          else if TE then
           A := 30;
          AmpA := A;
          if (T <= 3) or not TE then
           if EnvT in [8,12] then
            T := EnvP * 16
           else
            T := EnvP * 32;
         end;
        TnA := T;
        T := TnB;
        if AmpB and 16 = 0 then
         AmpB := AmpB * 2
        else if not (EnvT in [8,10,12,14]) then
         AmpB := E
        else
         begin
          A := E;
          TE := Mix and 2 = 0;
          if (T <= 3) and TE then
           Dec(A,6)
          else if TE then
           A := 30;
          AmpB := A;
          if (T <= 3) or not TE then
           if EnvT in [8,12] then
            T := EnvP * 16
           else
            T := EnvP * 32;
         end;
        TnB := T;
        T := TnC;
        if AmpC and 16 = 0 then
         AmpC := AmpC * 2
        else if not (EnvT in [8,10,12,14]) then
         AmpC := E
        else
         begin
          A := E;
          TE := Mix and 4 = 0;
          if (T <= 3) and TE then
           Dec(A,6)
          else if TE then
           A := 30;
          AmpC := A;
          if (T <= 3) or not TE then
           if EnvT in [8,12] then
            T := EnvP * 16
           else
            T := EnvP * 32;
         end;
        TnC := T;
        if not TSMode then break;
       end;
    end;
   T := R[0].AmpA; E := R[0].AmpB; A := R[0].AmpC;
   if TSMode then
    begin
     if R[1].AmpA > T then T := R[1].AmpA;
     if R[1].AmpB > E then E := R[1].AmpB;
     if R[1].AmpC > A then A := R[1].AmpC;
    end;
   RedrawVisChannels(T,E,A,30);
   RedrawVisSpectrum(@VisPoints[CurVisPos],31);
  end;
  ShowProgress(VProgrPos);
end;

end.


