{
This is part of AY Emulator project
AY-3-8910/12 Emulator
Version 2.9 for Windows and Linux
Author Sergey Vladimirovich Bulba
(c)1999-2016 S.V.Bulba
}

unit WaveOutAPI;

{$mode objfpc}{$H+}

interface

uses Classes, LCLIntf, LCLType, MMSystem, SysUtils, Forms, StdCtrls;

type
//Digital sound data buffer
 TWaveOutBuffer = packed array of byte;

 EMultiMediaError = class(Exception);

 TWOThread = class(TThread)
     protected
       procedure Execute; override;
     end;

const
 WODeviceDef = WAVE_MAPPER;

var
 WODevice:DWORD;
 Reseted:integer = 0;
 ResetCrSec:TCriticalSection;
 HWO:HWAVEOUT;
 waveOutBuffers:array of record
  Buf:TWaveOutBuffer;
  WH:WAVEHDR;
 end;

procedure WaveEnumDevices(cb:TComboBox);
procedure StartWOThread;
procedure StopWOThread;
procedure WOThreadFinalization;
procedure WOCheck(Res:MMRESULT);
function WOThreadActive:boolean;
procedure WOResetPlaying(Res:boolean);
procedure WOUnresetPlaying;
procedure WOPauseRestart;
procedure WOVisualisation;

implementation

uses MainWin, Players, AY;

var
 WOEvent:PRTLEvent;
 WOThread:TWOThread = nil;

 WOCS:TCriticalSection;

procedure WaveEnumDevices(cb:TComboBox);
var
 outcaps:WAVEOUTCAPS;
 i:integer;
begin
for i := 0 to integer(waveOutGetNumDevs) - 1 do
 if waveOutGetDevCaps(i, @outcaps, sizeof(outcaps)) = MMSYSERR_NOERROR then
  cb.Items.Add(AnsiToUTF8(outcaps.szPname))
 else
  cb.Items.Add('Unknown WaveOut device')
end;

procedure WOCheck(Res:MMRESULT);
var
 ErrMsg:array[0..255] of Char;
begin
if Res <> 0 then
 begin
  EnterCriticalSection(WOCS);
  waveOutGetErrorText(Res,ErrMsg,SizeOf(ErrMsg));
  LeaveCriticalSection(WOCS);
  raise EMultiMediaError.Create(ErrMsg);
 end;
end;

procedure WOVisualisation;
var
 MMTIME1:MMTime;
begin
if not WOThreadActive or Paused then exit;
    MMTIME1.wType := TIME_SAMPLES;
    EnterCriticalSection(WOCS);
    waveOutGetPosition(HWO,@MMTIME1,sizeof(MMTIME1));
    LeaveCriticalSection(WOCS);
    if MMTIME1.sample <> 0 then //if woReseted then don't redraw
     AYVisualisation(MMTIME1.sample);
end;

procedure WaitForWOThreadExit;
begin
if WoThread = nil then exit;
WOThread.Terminate;
WOThread.Free;
WOThread := nil;
end;

procedure StopWOThread;
var
 msg:TMsg;
begin
if WOThreadActive then
 begin
  IsPlaying := False;
  WOResetPlaying(True);
  WOUnresetPlaying;

  RtlEventSetEvent(WOEvent);

  WaitForWOThreadExit;
  while not PeekMessage(msg,Form1.Handle,WM_FINALIZEWO,WM_FINALIZEWO,PM_REMOVE) do Sleep(0);
  WOThreadFinalization;
 end;
end;

procedure TWOThread.Execute;

 function AllBuffersDone:boolean;
 var
  i:integer;
 begin
  Result := False;
  for i := 0 to NumberOfBuffers - 1 do
   if waveOutBuffers[i].WH.dwFlags and WHDR_DONE = 0 then exit;
  Result := True;
 end;

var
 i,SampleSize:integer;
 mut:boolean;
begin
SampleSize := (SampleBit div 8) * NumberOfChannels;
mut := False;
try
repeat
 EnterCriticalSection(ResetCrSec);
 try
 mut := True;
 if not Real_End_All then
  begin
   for i := 0 to NumberOfBuffers - 1 do
    with waveOutBuffers[i] do
     begin
      if Reseted > 0 then break;
      if not IsPlaying then break;
      if WH.dwFlags and WHDR_DONE <> 0 then
       begin
        MakeBuffer(WH.lpdata);
        if Reseted > 0 then break;
        if not IsPlaying then break;
        if BuffLen = 0 then
         begin
          if AllBuffersDone then break
         end
        else
         begin
          WH.dwBufferLength := BuffLen * SampleSize;
          WH.dwFlags := WH.dwFlags and not WHDR_DONE;
          EnterCriticalSection(WOCS);
          try
          WOCheck(waveOutWrite(HWO,@WH,sizeof(WAVEHDR)))
          finally
          LeaveCriticalSection(WOCS)
          end
         end
       end
     end
  end;
 if Real_End_All and (Reseted = 0) and AllBuffersDone then break;

 finally
 LeaveCriticalSection(ResetCrSec);
 mut := False;
 end;

 if not IsPlaying then break;

 RtlEventWaitFor(WOEvent,BufLen_ms);

until Terminated or not IsPlaying;
finally
if mut then
 LeaveCriticalSection(ResetCrSec);
PostMessage(Form1.Handle,WM_FINALIZEWO,0,0);
end;
end;

procedure StartWOThread;
var
 pwfx:pcmwaveformat;
 i,bl:integer;
begin
if WOThreadActive then exit;
with pwfx.wf do
 begin
  wFormatTag := 1;
  nChannels := NumberOfChannels;
  nSamplesPerSec := SampleRate;
  nBlockAlign := (SampleBit div 8) * NumberOfChannels;
  nAvgBytesPerSec := SampleRate * nBlockAlign
 end;
pwfx.wBitsPerSample := SampleBit;
WOCheck(waveOutOpen(@HWO,WODevice,@pwfx,THandle(WOEvent),0,CALLBACK_EVENT));
try
bl := BufferLength * pwfx.wf.nBlockAlign;
for i := 0 to NumberOfBuffers - 1 do
 with waveOutBuffers[i] do
  begin
   SetLength(Buf,bl);
   with WH do
    begin
     lpdata := @Buf[0];
     dwBufferLength := bl;
     dwFlags := 0;
     dwUser := 0;
     dwLoops := 0;
    end;
   WOCheck(waveOutPrepareHeader(HWO,@WH,sizeof(WAVEHDR)));
   WH.dwFlags := WH.dwFlags or WHDR_DONE;
  end
except
 WOCheck(waveOutClose(HWO));
 raise;
end;

IsPlaying := True;
Reseted := 0;
Paused := False;

WOThread := TWOThread.Create(False);

WOThread.Priority := tpHigher;
end;

procedure WOThreadFinalization;
var
 i:integer;
begin
WaitForWOThreadExit;

try
 WOCheck(waveOutReset(HWO));
 for i := 0 to NumberOfBuffers - 1 do
  with waveOutBuffers[i] do
   begin
    while WH.dwFlags and WHDR_DONE = 0 do Sleep(0);
    if WH.dwFlags and WHDR_PREPARED <> 0 then
     WOCheck(waveOutUnprepareHeader(HWO,@WH,sizeof(WAVEHDR)));
    Buf := nil;
   end;
 WOCheck(waveOutClose(HWO));
except
 ShowException(ExceptObject, ExceptAddr);
end;

IsPlaying := False;
Reseted := 0;
end;

function WOThreadActive:boolean;
begin
Result := WOThread <> nil;
end;

procedure WOResetPlaying(Res:boolean);
var
 i:integer;
begin
inc(Reseted);
if Reseted > 1 then exit;
EnterCriticalSection(ResetCrSec);
if Res then
 begin
  WOCheck(waveOutReset(HWO));
  MkVisPos := 0;
  VisPoint := 0;
  NOfTicks := 0;
  for i := 0 to NumberOfBuffers - 1 do
   with waveOutBuffers[i] do
    while WH.dwFlags and WHDR_DONE = 0 do Sleep(0);
 end;
end;

procedure WOUnresetPlaying;
begin
dec(Reseted);
if Reseted = 0 then
 begin
  RtlEventSetEvent(WOEvent);
  LeaveCriticalSection(ResetCrSec);
 end;
end;

procedure WOPauseRestart;
begin
EnterCriticalSection(WOCS);
try
if Paused then
 begin
  WOCheck(waveOutRestart(HWO));
  Paused := False
 end
else
 begin
  WOCheck(waveOutPause(HWO));
  Paused := True
 end 
finally
 LeaveCriticalSection(WOCS);
end
end;

initialization

WOEvent := RTLEventCreate;

InitializeCriticalSection(WOCS);
InitializeCriticalSection(ResetCrSec);

finalization

DeleteCriticalSection(ResetCrSec);

DeleteCriticalSection(WOCS);
RTLeventdestroy(WOEvent);

end.
