{
lightBASS
---------
(c)2003,2016 S.V.Bulba
http://bulba.untergrund.net/
svbulba@gmail.com

Description:
------------
Dinamycally loads/unloads BASS.DLL (BASSWMA.DLL, BASS_APE.DLL, BASSFLAC.DLL,
BASSWV.DLL, BASS_AC3.DLL)
Performs all required checks before calling BASS DLLs
Uses minimal set of constants, types and declarations from
original BASS.PAS, BASSWMA.PAS, BASS_APE.PAS, BASSFLAC.PAS, BASSWV.PAS,
BASS_AC3.PAS

Linux version is supported too

Written for using with BASS version 2.4
}

unit lightBASS;

{$mode objfpc}{$H+}

interface

uses
 LCLIntf,LCLType,{$IFDEF MSWINDOWS}Windows,{$ELSE}dl,{$ENDIF}SysUtils;

type
 HMUSIC = DWORD;
 HSAMPLE = DWORD;
 HSTREAM = DWORD;
 HSYNC = DWORD;
// BOOL = LongBool;
// QWORD = int64;
 HPLUGIN = DWORD;

 EBASSError = class(Exception);

 STREAMPROC = function(handle: HSTREAM; buffer: Pointer; length: DWORD; user: Pointer): DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 DOWNLOADPROC = procedure(buffer: Pointer; length: DWORD; user: Pointer); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 SYNCPROC = procedure(handle: HSYNC; channel, data: DWORD; user: Pointer); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};

  BASS_CHANNELINFO = record
    freq: DWORD;
    chans: DWORD;
    flags: DWORD;
    ctype: DWORD;
    origres: DWORD;
    plugin: HPLUGIN;
    sample: HSAMPLE;
    filename: PChar;
  end;

 TBassDll = (BASSDLL,BASSWMADLL,BASSAPEDLL,BASSFLACDLL,BASSWVDLL,BASSAC3DLL);

{
Next procedures and functions checks some flags and handles
and if all OK calls BASS.DLL functions. During calling
some errors can be ocurred, all of them are correctly translated
into DELPHI's exceptions with error messages
}

procedure RaiseLastBASSError;

procedure LoadBASS(DLL:TBassDll); //If BASS.DLL is not loaded then loads BASS.DLL,
                        //checks version and gets some procs addresses
                        //if DLL <> BASSDLL then also loads BASS plug-in

procedure UnloadBASS;   //Unload BASS*.DLL if it was loaded

procedure InitBASS(device: Integer; freq, flags: DWORD; win: HWND);
                        //calls BASS_Init if BASS was not initialized

procedure FreeBASS;     //calls BASS_Free if BASS was initialized

procedure PlayBASS(FileName:PChar;Stream:boolean;DLL:TBassDll;StartTime,TimeLenMs:integer);
                        //Tries start playing file: Stream = True - as stream,
                        //                          Stream = False - as module.
                        //If successed then set sync for end of music by
                        //WM_PLAYNEXTITEM message

procedure PlayFreeBASS;     //if PlayBASS was OK, stops playing and removes sync

function BASS_StreamCreateFile2(DLL:TBassDll; f: Pointer; flags: DWORD): HSTREAM;

procedure SwitchPause;  //during playing pauses/resumes playback

function GetLengthBASS:QWORD;
                        //returns max position for using with
                        //BASS_ChannelGetPosition.

procedure SetSync;    //Set SYNC_END and POS message
procedure SetMetaSync;//Set SYNC_META message
procedure RemoveSync; //Remove SYNC_END, POS and META message

{$IFNDEF Windows}
procedure SetVolumeBASS(v:single); //if BASS loaded set global volume
{$ENDIF Windows}


const
 BASS_DEFAULTDEVICE = -1;
 BASS_NOSOUNDDEVICE = 0;
 BASS_FIRSTSOUNDDEVICE = 1;

//Some consts from BASS*.PAS
 BASS_CONFIG_NET_AGENT     = 16;
 BASS_CONFIG_NET_PROXY     = 17;

 BASS_TAG_ID3   = 0;
 BASS_TAG_ID3V2 = 1;
 BASS_TAG_OGG   = 2;
 BASS_TAG_HTTP  = 3;
 BASS_TAG_ICY   = 4;
 BASS_TAG_META  = 5;
 BASS_TAG_APE   = 6;
 BASS_TAG_WMA   = 8;

 BASS_TAG_RIFF_INFO  = $100;
 BASS_TAG_MUSIC_NAME = $10000;
 BASS_TAG_MUSIC_MESSAGE = $10001;
 BASS_TAG_MUSIC_INST = $10100;
 BASS_TAG_MUSIC_SAMPLE = $10300;

 BASS_STREAM_PRESCAN     = $20000;
 BASS_STREAM_DECODE      = $200000;

 BASS_SAMPLE_LOOP        = 4;

 BASS_MUSIC_STOPBACK     = $80000;
 BASS_MUSIC_PRESCAN      = BASS_STREAM_PRESCAN;
 BASS_MUSIC_CALCLEN      = BASS_MUSIC_PRESCAN;
 BASS_MUSIC_NOSAMPLE     = $100000;

 BASS_POS_BYTE           = 0;

 BASS_DATA_FFT256   = $80000000;
 BASS_DATA_FFT512   = $80000001;
 BASS_DATA_FFT1024  = $80000002;
 BASS_DATA_FFT2048  = $80000003;
 BASS_DATA_FFT4096  = $80000004;
 BASS_DATA_FFT8192  = $80000005;
 BASS_DATA_FFT16384 = $80000006;
 BASS_DATA_FFT32768 = $80000007;
 BASS_DATA_FFT_NOWINDOW = $20;
 BASS_DATA_FFT_REMOVEDC = $40;

var
 BASSErrorString:string;

 //Some BASS.DLL function addresses (see description in original BASS.PAS)
 BASS_GetVersion:function: DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ErrorGetCode:function: LongInt; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_GetConfigPtr:function (option: DWORD): Pointer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_SetConfigPtr:function (option: DWORD; value: Pointer): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 {$IFDEF MSWINDOWS}
 BASS_Init:function (device: LongInt; freq, flags: DWORD; win: HWND; clsid: PGUID): BOOL; stdcall;
 {$ELSE}
 BASS_Init:function (device: LongInt; freq, flags: DWORD; win: Pointer; clsid: Pointer): BOOL; cdecl;
 {$ENDIF}
 BASS_Free:function: BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_StreamCreateFile:function (mem: BOOL; f: Pointer; offset, length: QWORD; flags: DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_StreamCreateURL:function (url: PChar; offset: DWORD; flags: DWORD; proc: DOWNLOADPROC; user: Pointer):HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_StreamCreate:function  (freq, chans, flags: DWORD; proc: STREAMPROC; user: Pointer): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_StreamFree:function (handle: HSTREAM): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelPlay:function (handle: DWORD; restart: BOOL): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelSetSync:function (handle: DWORD; type_: DWORD; param: QWORD; proc: SYNCPROC; user: Pointer): HSYNC; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelRemoveSync:function (handle: DWORD; sync: HSYNC): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelPause:function (handle: DWORD): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelStop:function (handle: DWORD): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetLength:function (handle, mode: DWORD): QWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetPosition:function (handle, mode: DWORD): QWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelSetPosition:function (handle: DWORD; pos: QWORD; mode: DWORD): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetLevel:function (handle: DWORD): DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetData:function (handle: DWORD; buffer: Pointer; length: DWORD): DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 {$IFNDEF Windows}
 BASS_ChannelSetAttribute: function (handle, attrib: DWORD; value: Single): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 {$ENDIF Windows}
 BASS_ChannelGetAttribute:function (handle, attrib: DWORD; var value: Single): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelBytes2Seconds:function (handle: DWORD; pos: QWORD): Double; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelSeconds2Bytes:function (handle: DWORD; pos: Double): QWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelFlags:function (handle, flags, mask: DWORD): DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetInfo:function (handle: DWORD; var info: BASS_CHANNELINFO):BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_ChannelGetTags:function (handle: HSTREAM; tags: DWORD): PAnsiChar; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_MusicLoad:function (mem: BOOL; f: Pointer; offset: QWORD; length, flags, freq: DWORD): HMUSIC; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_MusicFree:function (handle: HMUSIC): BOOL; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};

 BASS_WMA_StreamCreateFile:function (mem:BOOL; fl:pointer; offset,length:QWORD; flags:DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_APE_StreamCreateFile:function (mem: BOOL; f: Pointer; offset, length: QWORD; flags: DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_FLAC_StreamCreateFile:function (mem:BOOL; f:Pointer; offset,length:QWORD; flags:DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_WV_StreamCreateFile:function (mem:BOOL; fl:pointer; offset,length:QWORD; flags:DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
 BASS_AC3_StreamCreateFile:function (mem:BOOL; f:Pointer; offset,length:QWORD; flags:DWORD): HSTREAM; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};


 MusicHandle:DWORD = 0;  //handle to stream or module,
                           //0 => no music loaded

 MusicIsStream:boolean;    //True => Music is stream, otherwise module
 StreamIsUser:boolean;    //True => Stream is user filled, otherwise file-stream

 BASSPaused:boolean;       //pause flag, used by SwitchPause

 BASSInitialized:boolean = False; //True => BASS_Init was called successfully

 {$IFNDEF Windows}
 BASSVolume:single = 1; //global volume 0..1
 {$ENDIF Windows}

 BASSDevice:integer;

implementation

uses MainWin, Players, WinVersion, BASScode;

const
//Some consts from BASS.PAS
 BASS_SYNC_POS           = 0;
 BASS_SYNC_END           = 2;
 BASS_SYNC_META          = 4;
{$IFNDEF Windows}
 BASS_STREAMPROC_END = $80000000;
 BASS_SAMPLE_8BITS       = 1;
 BASS_ATTRIB_VOL         = 2;
{$ENDIF Windows}

// BASS_MP3_SETPOS         = BASS_STREAM_PRESCAN;

 LibsInfo:array[TBassDll] of record
   Hnd:PtrInt; //handle to BASS*.DLL instance,
               //0 => BASS*.DLL is not loaded
   FileName,FileProcName:string;
 end =
 ((Hnd:0;FileName:{$IFDEF MSWINDOWS}'bass.dll'{$ELSE}'libbass.so'{$ENDIF};FileProcName:''),
  (Hnd:0;FileName:{$IFDEF MSWINDOWS}'basswma.dll'{$ELSE}'libbasswma.so'{$ENDIF};FileProcName:'BASS_WMA_StreamCreateFile'),
  (Hnd:0;FileName:{$IFDEF MSWINDOWS}'bass_ape.dll'{$ELSE}'libbass_ape.so'{$ENDIF};FileProcName:'BASS_APE_StreamCreateFile'),
  (Hnd:0;FileName:{$IFDEF MSWINDOWS}'bassflac.dll'{$ELSE}'libbassflac.so'{$ENDIF};FileProcName:'BASS_FLAC_StreamCreateFile'),
  (Hnd:0;FileName:{$IFDEF MSWINDOWS}'basswv.dll'{$ELSE}'libbasswv.so'{$ENDIF};FileProcName:'BASS_WV_StreamCreateFile'),
  (Hnd:0;FileName:{$IFDEF MSWINDOWS}'bass_ac3.dll'{$ELSE}'libbass_ac3.so'{$ENDIF};FileProcName:'BASS_AC3_StreamCreateFile'));

var
 hsEND:HSYNC = 0; //sync handler, used for end of music message (WM_PLAYNEXTITEM)
                  //if <> 0 then sync is set
 hsPOS:HSYNC = 0;                 
 hsMETA:HSYNC = 0;
 SyncTime,SyncRestart:QWORD;

 CallbackWindow:HWND;

procedure SyncEndProc(handle: HSYNC; channel, data: DWORD; user: Pointer); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
begin
if (SyncTime = QWORD(-1)) or not Do_Loop or not BASS_ChannelSetPosition(MusicHandle,SyncRestart,BASS_POS_BYTE) then
 PostMessage(CallbackWindow,WM_PLAYNEXTITEM,0,0);
end;

procedure SyncMetaProc(handle: HSYNC; channel, data: DWORD; user: Pointer); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
begin
PostMessage(CallbackWindow,WM_BASSMETADATA,0,0);
end;

procedure RaiseLastBASSError;
const
 BASSErCodes:array[0..46] of string =
 ('All is OK',
  'Memory error',
  'Can''t open the file',
  'Can''t find a free sound driver',
  'The sample buffer was lost',
  'Invalid handle',
  'Unsupported sample format',
  'Invalid position',
  'BASS_Init has not been successfully called',
  'BASS_Start has not been successfully called',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Already initialized/paused/whatever',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Can''t get a free channel',
  'An illegal type was specified',
  'An illegal parameter was specified',
  'No 3D support',
  'No EAX support',
  'Illegal device number',
  'Not playing',
  'Illegal sample rate',
  'Unknown error',
  'The stream is not a file stream',
  'Unknown error',
  'No hardware voices available',
  'Unknown error',
  'The MOD music has no sequence data',
  'No internet connection could be opened',
  'Couldn''t create the file',
  'Effects are not enabled',
  'Unknown error',
  'Unknown error',
  'Requested data is not available',
  'The channel is/isn''t a "decoding channel"',
  'A sufficient DirectX version is not installed',
  'Connection timedout',
  'Unsupported file format',
  'Unavailable speaker',
  'Invalid BASS version',
  'Codec is not available/supported',
  'The channel/file has ended',
  'The device is busy');
var
 ErCode:DWORD;
begin
ErCode := BASS_ErrorGetCode();
if ErCode > 46 then ErCode := 26;
BASSErrorString := BASSErCodes[ErCode];
raise EBASSError.Create(BASSErCodes[ErCode]);
end;

function TryGet(const p:pointer):pointer;
begin
Result := p;
if p = nil then
 begin
  BASSErrorString := '';
  RaiseLastOSError;
 end;
end;

procedure LoadBASSPlugIn(DLL:TBassDll);

 function TryGetProcAddress:pointer;
 begin
 {$IFDEF MSWINDOWS}
 Result := TryGet(GetProcAddress(LibsInfo[DLL].Hnd,pchar(LibsInfo[DLL].FileProcName)));
 {$ELSE}
 Result := TryGet(dlsym(LibsInfo[DLL].Hnd,pchar(LibsInfo[DLL].FileProcName)));
 {$ENDIF}
 end;

begin
if LibsInfo[DLL].Hnd <> 0 then exit;
{$IFDEF MSWINDOWS}
LibsInfo[DLL].Hnd := LoadLibrary(PChar(LibsInfo[DLL].FileName));
{$ELSE}
LibsInfo[DLL].Hnd := PtrInt(dlopen(pchar(extractfiledir(paramstr(0))+'/'+LibsInfo[DLL].FileName),RTLD_LAZY));
{$ENDIF}
if LibsInfo[DLL].Hnd = 0 then
 begin
//  UnloadBASS;
  BASSErrorString := LibsInfo[DLL].FileName + ' 2.4 required for playing extra files';
  raise EBASSError.Create(BASSErrorString);
 end;
try
  case DLL of
  BASSWMADLL:
   pointer(BASS_WMA_StreamCreateFile) := TryGetProcAddress;
  BASSAPEDLL:
   pointer(BASS_APE_StreamCreateFile) := TryGetProcAddress;
  BASSFLACDLL:
   pointer(BASS_FLAC_StreamCreateFile) := TryGetProcAddress;
  BASSWVDLL:
   pointer(BASS_WV_StreamCreateFile) := TryGetProcAddress;
  BASSAC3DLL:
   pointer(BASS_AC3_StreamCreateFile) := TryGetProcAddress;
  end;
except
 UnloadBASS;
 raise;
end;
end;

procedure CheckVersion;
const
 errs = 'Sorry, BASS version 2.4 required';
begin
if BASS_GetVersion() and $FFFF0000 <> $02040000 then
 begin
  BASSErrorString := errs;
  raise EBASSError.Create(errs);
 end;
end;

procedure LoadBASS(DLL:TBassDll);

 function TryGetProcAddress(const prcname:string):pointer;
 begin
 {$IFDEF MSWINDOWS}
 Result := TryGet(GetProcAddress(LibsInfo[BASSDLL].Hnd,pchar(prcname)));
 {$ELSE}
 Result := TryGet(dlsym(LibsInfo[BASSDLL].Hnd,pchar(prcname)));
 {$ENDIF}
 end;

begin
if LibsInfo[BASSDLL].Hnd = 0 then
 begin
  {$IFDEF MSWINDOWS}
  LibsInfo[BASSDLL].Hnd := LoadLibrary(PChar(LibsInfo[BASSDLL].FileName));
  {$ELSE}
  LibsInfo[BASSDLL].Hnd := PtrInt(dlopen(pchar(extractfiledir(paramstr(0))+'/'+LibsInfo[BASSDLL].FileName),RTLD_LAZY or RTLD_GLOBAL));
  {$ENDIF}
  if LibsInfo[BASSDLL].Hnd = 0 then
   begin
    BASSErrorString := LibsInfo[BASSDLL].FileName+' 2.4 by Ian Luck required for playing extra file types';
    raise EBASSError.Create(BASSErrorString);
   end;
  try
   pointer(BASS_GetVersion) := TryGetProcAddress('BASS_GetVersion');
   CheckVersion;
   pointer(BASS_ErrorGetCode) := TryGetProcAddress('BASS_ErrorGetCode');
   pointer(BASS_GetConfigPtr) := TryGetProcAddress('BASS_GetConfigPtr');
   GetNetConfig;
   pointer(BASS_SetConfigPtr) := TryGetProcAddress('BASS_SetConfigPtr');
   pointer(BASS_Init) := TryGetProcAddress('BASS_Init');
   pointer(BASS_Free) := TryGetProcAddress('BASS_Free');
   pointer(BASS_StreamCreateFile) := TryGetProcAddress('BASS_StreamCreateFile');
   pointer(BASS_StreamCreateURL) := TryGetProcAddress('BASS_StreamCreateURL');
   pointer(BASS_StreamCreate) := TryGetProcAddress('BASS_StreamCreate');
   pointer(BASS_StreamFree) := TryGetProcAddress('BASS_StreamFree');
   pointer(BASS_ChannelPlay) := TryGetProcAddress('BASS_ChannelPlay');
   pointer(BASS_ChannelSetSync) := TryGetProcAddress('BASS_ChannelSetSync');
   pointer(BASS_ChannelRemoveSync) := TryGetProcAddress('BASS_ChannelRemoveSync');
   pointer(BASS_ChannelPause) := TryGetProcAddress('BASS_ChannelPause');
   pointer(BASS_ChannelStop) := TryGetProcAddress('BASS_ChannelStop');
   pointer(BASS_ChannelGetLength) := TryGetProcAddress('BASS_ChannelGetLength');
   pointer(BASS_ChannelGetPosition) := TryGetProcAddress('BASS_ChannelGetPosition');
   pointer(BASS_ChannelSetPosition) := TryGetProcAddress('BASS_ChannelSetPosition');
   pointer(BASS_ChannelGetLevel) := TryGetProcAddress('BASS_ChannelGetLevel');
   pointer(BASS_ChannelGetData) := TryGetProcAddress('BASS_ChannelGetData');
   {$IFNDEF Windows}
   pointer(BASS_ChannelSetAttribute) := TryGetProcAddress('BASS_ChannelSetAttribute');
   {$ENDIF Windows}
   pointer(BASS_ChannelGetAttribute) := TryGetProcAddress('BASS_ChannelGetAttribute');
   pointer(BASS_ChannelBytes2Seconds) := TryGetProcAddress('BASS_ChannelBytes2Seconds');
   pointer(BASS_ChannelSeconds2Bytes) := TryGetProcAddress('BASS_ChannelSeconds2Bytes');
   pointer(BASS_ChannelFlags) := TryGetProcAddress('BASS_ChannelFlags');
   pointer(BASS_ChannelGetInfo) := TryGetProcAddress('BASS_ChannelGetInfo');
   pointer(BASS_ChannelGetTags) := TryGetProcAddress('BASS_ChannelGetTags');
   pointer(BASS_MusicLoad) := TryGetProcAddress('BASS_MusicLoad');
   pointer(BASS_MusicFree) := TryGetProcAddress('BASS_MusicFree');
  except
   UnloadBASS;
   raise;
  end;
 end;
LoadBASSPlugIn(DLL);
end;

procedure UnloadBASS;
var
 DLL:TBassDll;
begin
for DLL := High(TBassDll) downto Low(TBassDll) do
 if LibsInfo[DLL].Hnd <> 0 then
   begin
    {$IFDEF MSWINDOWS}
    FreeLibrary(LibsInfo[DLL].Hnd);
    {$ELSE}
    dlClose(LibsInfo[DLL].Hnd);
    {$ENDIF}
    LibsInfo[DLL].Hnd := 0;
   end;
end;

procedure InitBASS(device: Integer; freq, flags: DWORD; win: HWND);
begin
if BASSInitialized and (BASSDevice = device) then exit;
FreeBASS;
BASSDevice := device;
CallbackWindow := win;
BASSInitialized := BASS_Init(device,freq,flags,{$IFDEF MSWINDOWS}win{$ELSE}nil{$ENDIF},nil);
if not BASSInitialized then RaiseLastBASSError;
end;

procedure FreeBASS;
begin
if BASSInitialized then
 begin
  BASSInitialized := False;
  BASS_Free;
 end;
end;

{$IFNDEF Windows}
function StreamProcEmu(handle: HSTREAM; buffer: Pointer; length: DWORD; user: Pointer): DWORD; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
var
 bl:integer;
begin
Result := 0;
if handle = MusicHandle then
 begin
  bl := BufferLength;
  BufferLength := DivMul(length,8,Int64(NumberOfChannels) * Int64(SampleBit));
  MakeBuffer(buffer);
  BufferLength := bl;
  Result := BuffLen * NumberOfChannels * SampleBit div 8;
  if Real_End_All then
   Result := Result or BASS_STREAMPROC_END;
 end;
end;
{$ENDIF Windows}

procedure PlayBASS(FileName:PChar;Stream:boolean;DLL:TBassDll;StartTime,TimeLenMs:integer);
var
 fl:DWORD;
 Restart:boolean;
begin
if MusicHandle <> 0 then exit;
MusicIsStream := Stream;
StreamIsUser := FileName = '';
fl := 0; if Do_Loop then fl := BASS_SAMPLE_LOOP;
if Stream then
 begin
 //too slow
{  if Mpeg and ((Length(FileName) <= 2) or (FileName[1] <> '\') or
   (FileName[2] <> '\')) then //exclude prescan if network path
   f := f or BASS_STREAM_PRESCAN;}
  {$IFNDEF Windows}
  if FileName <> nil then
  {$ENDIF Windows}
   MusicHandle := BASS_StreamCreateFile2(DLL,FileName,fl)
  {$IFNDEF Windows}
  else
   begin
    if SampleBit = 8 then
     fl := BASS_SAMPLE_8BITS
    else
     fl := 0;
    MusicHandle := BASS_StreamCreate(SampleRate,NumberOfChannels,fl,@StreamProcEmu,nil);
   end
  {$ENDIF Windows}
   ;
 end
else
 MusicHandle := BASS_MusicLoad(False,FileName,0,0,
  BASS_MUSIC_STOPBACK or BASS_MUSIC_PRESCAN {or BASS_MUSIC_POSRESETEX} or fl,0);
if MusicHandle = 0 then RaiseLastBASSError;
BASSPaused := False;
Restart := True;
SyncTime := QWORD(-1);
if StartTime >= 0 then
 begin
  SyncRestart := BASS_ChannelSeconds2Bytes(MusicHandle,StartTime / 1000);
  if not BASS_ChannelSetPosition(MusicHandle,SyncRestart,BASS_POS_BYTE) then
   RaiseLastBASSError
  else
   begin
    Restart := False;
    SyncTime := BASS_ChannelSeconds2Bytes(MusicHandle,(StartTime + TimeLenMs) / 1000);
    SetSync;
   end;
 end;
if Restart and not Do_Loop then SetSync;
if Stream and not StreamIsUser then SetMetaSync;
{$IFNDEF Windows}
BASS_ChannelSetAttribute(MusicHandle,BASS_ATTRIB_VOL,BASSVolume);
{$ENDIF Windows}
if not BASS_ChannelPlay(MusicHandle,Restart) then RaiseLastBASSError;
end;

procedure PlayFreeBASS;
begin
if MusicHandle = 0 then exit;
try
 RemoveSync;
 BASS_ChannelStop(MusicHandle);
finally
 if MusicIsStream then
  BASS_StreamFree(MusicHandle)
 else
  BASS_MusicFree(MusicHandle);
 MusicHandle := 0;
end;
end;

(*var
 aFile:THandle=0;

procedure MyDownloadProc(buffer:pointer;length:DWORD;user:pointer);{$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
const
 CrLf = #13#10;
begin
if aFile = 0 then aFile := FileCreate(ParamStr(0)+'.download.mp3',fmOpenWrite);
if buffer = nil then
 begin
  FileClose(aFile);
  aFile := 0;
 end
else
 if length <> 0 then
  FileWrite(aFile,buffer^,length)
 else
  begin
   FileWrite(aFile,PChar(buffer)^,StrLen(PChar(buffer)));
   FileWrite(aFile,CrLf,2);
  end;
end;*)

function BASS_StreamCreateFile2(DLL:TBassDll; f: Pointer; flags: DWORD): HSTREAM;
begin
if FileIsURL(PChar(f)) then
 begin
  BASS_SetConfigPtr(BASS_CONFIG_NET_AGENT,PChar(BASSNetAgent));
  if not BASSNetUseProxy then
   BASS_SetConfigPtr(BASS_CONFIG_NET_PROXY,nil)
  else
   BASS_SetConfigPtr(BASS_CONFIG_NET_PROXY,PChar(BASSNetProxy));
  Result := BASS_StreamCreateURL(f,0,flags(* or $800000{BASS_STREAM_STATUS}*),nil{@MyDownloadProc},nil);
  exit;
 end;
case DLL of
BASSWMADLL:
 Result := BASS_WMA_StreamCreateFile(False,f,0,0,flags);
BASSAPEDLL:
 Result := BASS_APE_StreamCreateFile(False,f,0,0,flags);
BASSFLACDLL:
 Result := BASS_FLAC_StreamCreateFile(False,f,0,0,flags);
BASSWVDLL:
 Result := BASS_WV_StreamCreateFile(False,f,0,0,flags);
BASSAC3DLL:
 Result := BASS_AC3_StreamCreateFile(False,f,0,0,flags);
else
 Result := BASS_StreamCreateFile(False,f,0,0,flags);
end; 
end;

procedure SwitchPause;
begin
if MusicHandle = 0 then exit;
if not BASSPaused then
 begin
  BASSPaused := BASS_ChannelPause(MusicHandle);
  if not BASSPaused then RaiseLastBASSError;
 end
else
 begin
  BASSPaused := not BASS_ChannelPlay(MusicHandle,False);
  if BASSPaused then RaiseLastBASSError;
 end
end;

procedure SetSync;
begin
hsEND := BASS_ChannelSetSync(MusicHandle,BASS_SYNC_END,0,@SyncEndProc,nil);
if SyncTime <> QWORD(-1) then
 hsPOS := BASS_ChannelSetSync(MusicHandle,BASS_SYNC_POS,SyncTime,@SyncEndProc,nil);
end;

procedure SetMetaSync;
begin
hsMETA := BASS_ChannelSetSync(MusicHandle,BASS_SYNC_META,0,@SyncMetaProc,nil);
end;

procedure RemoveSync;
begin
 if hsEND <> 0 then
  begin
   if not BASS_ChannelRemoveSync(MusicHandle,hsEND) then RaiseLastBASSError;
   hsEND := 0;
  end;
 if hsPOS <> 0 then
  begin
   if not BASS_ChannelRemoveSync(MusicHandle,hsPOS) then RaiseLastBASSError;
   hsPOS := 0;
  end;
 if hsMETA <> 0 then
  begin
   if not BASS_ChannelRemoveSync(MusicHandle,hsMETA) then RaiseLastBASSError;
   hsMETA := 0;
  end;
end;

function GetLengthBASS:QWORD;
begin
Result := 0;
if MusicHandle = 0 then exit;
Result := BASS_ChannelGetLength(MusicHandle,BASS_POS_BYTE);
if int64(Result) = - 1 then RaiseLastBASSError;
end;

{$IFNDEF Windows}
procedure SetVolumeBASS(v:single);
begin
BASSVolume := v;
if MusicHandle = 0 then exit;
BASS_ChannelSetAttribute(MusicHandle,BASS_ATTRIB_VOL,v);
end;
{$ENDIF Windows}

end.
