{
This is part of AY Emulator project
AY-3-8910/12 Emulator
Version 2.9 for Windows and Linux
Author Sergey Vladimirovich Bulba
(c)1999-2016 S.V.Bulba
}

unit Options;

{$mode objfpc}{$H+}

interface

uses
  {$IFDEF Windows}Windows,{$ENDIF Windows}
  WinVersion, Classes, SysUtils, FileUtil, Forms, lazutf8;

function OptionsInit(PrepareToWrite:boolean):boolean;
procedure OptionsDone;
function OptionsRead(const OptionName:string;var OptionValue:string):boolean;
procedure OptionsWrite(const OptionName,OptionValue:string);
function CheckRegistration(S1:PChar;Ind:integer):boolean;
procedure SaveDefaultDir3;
procedure SaveDefaultDir2;
procedure DeleteOptions;
{$IFDEF Windows}
procedure RegisterType(n:integer);
function RegisterFile(S1:PChar;Ind:integer):boolean;
function UnRegisterFile(S1:PChar;Ind:integer):boolean;
function CheckRegPath(a:integer):boolean;
procedure SetIfRegPath;
procedure CheckRegError(Index:integer);

var
 RegAudio:boolean;
 RegPlaylist:boolean;
 RegSkin:boolean;
 RegBASS:boolean;
{$ENDIF Windows}

implementation

uses
 MainWin;

const
 fn = 'Ay_Emul.cfg';

var
  OPath:string='';
  OpArr:array of record
   Name,Value:string;
  end;

{$IFDEF Windows}
const
MyKeys:array[0..3]of string =
('AY_Emul Audio File','AY_Emul Playlist File',
 'AY_Emul Skin File','AY_Emul BASS File');

type
 ERegistryError = class(Exception);

procedure CheckRegError(Index:integer);
var
 Strg:PChar;
begin
if Index <> ERROR_SUCCESS then
 begin
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ALLOCATE_BUFFER,
                nil,Index,0,@Strg,0,nil);
  try
   raise ERegistryError.Create(Strg);
  finally
   LocalFree(HLOCAL(Strg));
  end;
 end;
end;
{$ENDIF Windows}

procedure AddOption(const Name,Value:string);
var
 i,n:integer;
begin
n := Length(OpArr);
if Name <> '' then
  for i := 0 to n-1 do
   if OpArr[i].Name = Name then
    begin
     OpArr[i].Value:=Value;
     exit;
    end;
SetLength(OpArr,n+1);
OpArr[n].Name:=Name;
OpArr[n].Value:=Value;
end;

function OptionsInit(PrepareToWrite:boolean):boolean;
var
 f:TextFile;

 procedure LoadOptions;
 var
  s,s1:string;
  i:integer;
 begin
  System.Reset(f);
  try
   while not eof(f) do
    begin
     Readln(f,s1);
     s := Trim(s1); if s = '' then
      begin
       AddOption('','');
       continue;
      end;
     if s[1] in [';','#'] then
      begin
       AddOption('',s1);
       continue;
      end;
     i := Pos('=',s1); if i in [0,1] then continue;
     AddOption(Copy(s1,1,i-1),Copy(s1,i+1,Length(s1)-i));
    end;
  finally
   CloseFile(f);
  end;
 end;

  function CheckAndRead(const path:string):boolean;
  var
   fname:string;
  begin
  Result := False;
  fname := path+fn;
  if FileExistsUTF8(fname) then
   begin
    {$IFDEF Windows}
    fname := ValidateFileName(UTF8Decode(fname));
    {$ENDIF Windows}
    AssignFile(f,fname);
    try
     LoadOptions;
    except
     exit;
    end;
    OPath := path;
    Result := True;
   end;
  end;

 function CheckWriteAccess(const path:string):boolean;
 var
  fname:string;
  h:THandle;
 begin
 Result := False;
 fname := path+fn;
 if not FileExistsUTF8(fname) then
  if not ForceDirectoriesUTF8(path) then exit;
 h := FileCreateUTF8(fname); if integer(h) = -1 then exit;
 FileClose(h);
 OPath := path;
 Result := True;
 end;

begin
Result := False;
if OPath = '' then
 begin
  Application.Title := 'Ay_Emul'; //avoid undesired behavior of GetAppConfigDirUTF8
  if not CheckAndRead(ExtractFilePath(GetProcessFileName)) then
   if not CheckAndRead(GetAppConfigDirUTF8(False)) then
    if not PrepareToWrite then exit;
 end
else
 if not CheckAndRead(OPath) then
  if not PrepareToWrite then exit;
if PrepareToWrite then
 begin
  if OPath = '' then
   begin
    if not CheckWriteAccess(GetAppConfigDirUTF8(False)) then
     if not CheckWriteAccess(ExtractFilePath(GetProcessFileName)) then
      begin
       OpArr := nil;
       exit;
      end;
   end
  else if not CheckWriteAccess(OPath) then
   begin
    OpArr := nil;
    exit;
   end;
 end;
Result := True;
end;

procedure OptionsDone;
var
 i,n:integer;
 f:TextFile;
 fname:string;
begin
n := Length(OpArr); if n = 0 then exit;
fname := OPath+fn;
{$IFDEF Windows}
fname := ValidateFileName(UTF8Decode(fname));
{$ENDIF Windows}
AssignFile(f,fname);
try
 Rewrite(f);
 try
  for i := 0 to n-1 do
   if OpArr[i].Name <> '' then
    Writeln(f,OpArr[i].Name,'=',OpArr[i].Value)
   else
    Writeln(f,OpArr[i].Value);
 finally
  CloseFile(f);
 end;
except
end;
OpArr := nil;
end;

function OptionsRead(const OptionName:string;var OptionValue:string):boolean;
var
 i,n:integer;
begin
Result := False;
n := Length(OpArr); if n = 0 then exit;
for i := 0 to n-1 do
 if OpArr[i].Name = OptionName then
  begin
   OptionValue := OpArr[i].Value;
   Result := True;
   exit;
  end;
end;

procedure OptionsWrite(const OptionName,OptionValue:string);
begin
AddOption(OptionName,OptionValue);
end;

{$IFDEF Windows}
function CheckRegPath(a:integer):boolean;
var
 DataStr:string;
 Str1,Str2:string;
 i:integer;
 subKeyHnd1:HKey;
begin
Result := False;
DataStr := MyKeys[a] + '\shell\open\command'#0;
i := RegOpenKeyEx(HKEY_CLASSES_ROOT,PChar(DataStr),0,{KEY_ALL_ACCESS}KEY_READ,subKeyHnd1);
if i <> ERROR_SUCCESS then
 begin
  if i <> ERROR_FILE_NOT_FOUND then CheckRegError(i);
  exit;
 end;
try
 Str1 := UTF8LowerCase(RegGetUTF8Value(subKeyHnd1));
 Str2 := UTF8LowerCase(GetProcessFileName);
 Result := Str1 = '"' + Str2 + '"';
finally
 RegCloseKey(subKeyHnd1);
end;
end;
{$ENDIF Windows}

procedure SaveDefaultDir2;
{$IFDEF Windows}
var
 DefDir:string;
{$ENDIF Windows}
begin
if Form1.DefaultDirectory = '' then exit;
{$IFDEF Windows}
DefDir := ExtractFileDrive(Form1.DefaultDirectory);
if not ((DefDir <> '') and (DefDir[1] in ['a'..'z','A'..'Z']) and
        (GetDriveType(PChar(DefDir + '\')) = DRIVE_FIXED)) then exit;
{$ENDIF Windows}
OptionsWrite('DefaultDirectory',Form1.DefaultDirectory);
end;

{$IFDEF Windows}
procedure RegisterType(n:integer);
const
 RTypes:array[boolean,0..3] of string =
 (('AY/YM Audio File',
   'AY Emulator Playlist File',
   'AY Emulator Skin File',
   'AY_Emul&BASS Audio File'),
  ('Аудио файл AY/YM',
   'Список проигрывания эмулятора AY',
   'Файл обшивки эмулятора AY',
   'Аудио файл AY_Emul&BASS'));
 Cmds:array[0..2] of record
  Name:string;
  VisName:array[boolean] of string;
 end =
 ((Name:'open';VisName:('&Open','&Открыть')),
  (Name:'add';VisName:('Add','Добавить')),
  (Name:'addplay';VisName:('Add''n''play','Добавить и проиграть')));
var
 i:integer;
 DataStr:string;
 subKeyHnd1,subKeyHnd2,subKeyHnd3,subKeyHnd4:HKey;
 CreateStatus:longword;

procedure AddShellCommand(cm:integer);
begin
 i := 0;
 i := RegCreateKeyEx(subKeyHnd1,pchar('shell\'+Cmds[cm].Name),0,@i,
        REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd2,@CreateStatus);
 CheckRegError(i);
 try
  DataStr := Cmds[cm].VisName[Russian_Interface];
  RegSetUTF8Value(subKeyHnd2,DataStr);
  i := 0;
  i := RegCreateKeyEx(subKeyHnd2,'command',0,@i,
         REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd3,@CreateStatus);
  CheckRegError(i);
  try
   DataStr := '"' + GetProcessFileName + '"';
   RegSetUTF8Value(subKeyHnd3,DataStr);
  finally
   RegCloseKey(subKeyHnd3);
  end;
  i := 0;
  i := RegCreateKeyEx(subKeyHnd2,'ddeexec',0,@i,
         REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd3,@CreateStatus);
  CheckRegError(i);
  try
   case cm of
   0:
    begin
     DataStr := '"%1"';
     if n = 2 then DataStr := '/p' + DataStr;
    end;
   1:
     DataStr := '/add "%1"';
   2:
     DataStr := '/adp "%1"';
   end;
   RegSetUTF8Value(subKeyHnd3,DataStr);
   i := 0;
   i := RegCreateKeyEx(subKeyHnd3,'Application',0,@i,
          REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd4,@CreateStatus);
   CheckRegError(i);
   try
    RegSetUTF8Value(subKeyHnd4,DdeServiceName);
   finally
    RegCloseKey(subKeyHnd4);
   end;
  finally
   RegCloseKey(subKeyHnd3);
  end;
 finally
  RegCloseKey(subKeyHnd2);
 end;
end;

begin
DataStr := MyKeys[n] + #0;
i := 0;
i := RegCreateKeyEx(HKEY_CLASSES_ROOT,PChar(DataStr),0,@i,
       REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd1,@CreateStatus);
CheckRegError(i);
try
 DataStr := RTypes[Russian_Interface,n];
 RegSetUTF8Value(subKeyHnd1,DataStr);
 i := 0;
 i := RegCreateKeyEx(subKeyHnd1,'DefaultIcon',0,@i,
        REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd2,@CreateStatus);
 CheckRegError(i);
 try
  DataStr := GetProcessFileName + ',';
  case n of
  0:
   DataStr := DataStr + IntToStr(MusIconNumber);
  1:
   DataStr := DataStr + IntToStr(ListIconNumber);
  2:
   DataStr := DataStr + IntToStr(SkinIconNumber);
  3:
   DataStr := DataStr + IntToStr(BASSIconNumber)
  end;
  RegSetUTF8Value(subKeyHnd2,DataStr);
 finally
  RegCloseKey(subKeyHnd2);
 end;
 AddShellCommand(0);
 if n <> 2 then
  begin
   AddShellCommand(1);
   AddShellCommand(2);
  end;
 case n of
 0:
  RegAudio := True;
 1:
  RegPlaylist := True;
 2:
  RegSkin := True;
 3:
  RegBASS := True;
 end;
finally
 RegCloseKey(subKeyHnd1);
end;
end;

procedure SetIfRegPath;
begin
RegAudio := CheckRegPath(0);
RegPlaylist := CheckRegPath(1);
RegSkin := CheckRegPath(2);
RegBASS := CheckRegPath(3);
end;

{$ENDIF Windows}

function CheckRegistration(S1:PChar;Ind:integer):boolean;
{$IFDEF Windows}
var
 i:integer;
 RStr:PChar;
 subKeyHnd1:HKey;
begin
Result := False;
if Ind = 0 then if not RegAudio then exit;
if Ind = 1 then if not RegPlaylist then exit;
if Ind = 2 then if not RegSkin then exit;
if Ind = 3 then if not RegBASS then exit;
i := RegOpenKeyEx(HKEY_CLASSES_ROOT,S1,0,KEY_READ,subKeyHnd1);
if i <> ERROR_SUCCESS then
 begin
  if i <> ERROR_FILE_NOT_FOUND then CheckRegError(i);
  exit;
 end;
if RegQueryValueEx(subKeyHnd1,nil,nil,nil,nil,@i) = ERROR_SUCCESS then
 begin
  GetMem(RStr,i);
  if RegQueryValueEx(subKeyHnd1,nil,nil,nil,PByte(RStr),@i) = ERROR_SUCCESS then
   Result := (MyKeys[Ind] = RStr);
  FreeMem(RStr);
 end;
RegCloseKey(subKeyHnd1);
{$ELSE Windows}
begin
 Result := False;
{$ENDIF Windows}
end;

{$IFDEF Windows}
function RegisterFile(S1:PChar;Ind:integer):boolean;
var
 i:integer;
 subKeyHnd1:HKey;
 CreateStatus:longword;
 RStr:PChar;
begin
try
 i := 0;
 CheckRegError(RegCreateKeyEx(HKEY_CLASSES_ROOT,S1,0,@i,
       REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,nil,subKeyHnd1,@CreateStatus));
 try
  if RegQueryValueEx(subKeyHnd1,nil,nil,nil,nil,@i) = ERROR_SUCCESS then
   begin
    GetMem(RStr,i);
    try
     if RegQueryValueEx(subKeyHnd1,nil,nil,nil,PByte(RStr),@i) = ERROR_SUCCESS then
      if (StrComp(PChar(MyKeys[Ind] + #0),RStr) <> 0) and (RStr[0] <> #0) then
       CheckRegError(RegSetValueEx(subKeyHnd1,'Before AY_Emul',0,REG_SZ,RStr,i))
    finally
     FreeMem(RStr);
    end;
   end;
  CheckRegError(RegSetValueEx(subKeyHnd1,nil,0,REG_SZ,PChar(MyKeys[Ind] + #0),
                 Length(MyKeys[Ind]) + 1));
  Result := True;
 finally
  RegCloseKey(subKeyHnd1);
 end;
except
 Result := False;
end;
end;

function UnRegisterFile(S1:PChar;Ind:integer):boolean;
var
 i,j:integer;
 subKeyHnd1:HKey;
 RStr1,RStr2:PChar;
begin
Result := False;
i := RegOpenKeyEx(HKEY_CLASSES_ROOT,S1,0,KEY_ALL_ACCESS,subKeyHnd1);
if i <> ERROR_SUCCESS then
 begin
  if i <> ERROR_FILE_NOT_FOUND then CheckRegError(i);
  exit
 end;
try
 j := RegQueryValueEx(subKeyHnd1,nil,nil,nil,nil,@i);
 if j in [ERROR_SUCCESS,ERROR_FILE_NOT_FOUND] then
  begin
   if j = ERROR_SUCCESS then
    begin
     GetMem(RStr1,i);
     try
      if RegQueryValueEx(subKeyHnd1,nil,nil,nil,PByte(RStr1),@i) = ERROR_SUCCESS then
       if StrComp(RStr1,PChar(MyKeys[Ind] + #0)) = 0 then
        begin
         if RegQueryValueEx(subKeyHnd1,'Before AY_Emul',nil,nil,nil,@i) = ERROR_SUCCESS then
          begin
           GetMem(RStr2,i);
           try
            try
             if RegQueryValueEx(subKeyHnd1,'Before AY_Emul',nil,nil,PByte(RStr2),@i) = ERROR_SUCCESS then
              CheckRegError(RegSetValueEx(subKeyHnd1,nil,0,REG_SZ,RStr2,i));
             exit;
            except
            end;
           finally
            FreeMem(RStr2);
            RegDeleteValue(subKeyHnd1,'Before AY_Emul');
           end
          end;
         try
          CheckRegError(RegDeleteValue(subKeyHnd1,nil));
         except
          Result := True;
         end;
        end;
     finally
      FreeMem(RStr1);
     end;
    end;
   CheckRegError(RegQueryInfoKey(subKeyHnd1,nil,nil,nil,@i,nil,nil,@j,nil,nil,nil,nil));
   if (i = 0) and (j = 0) then
    begin
     RegCloseKey(subKeyHnd1);
     subKeyHnd1 := 0;
     CheckRegError(RegDeleteKey(HKEY_CLASSES_ROOT,S1));
    end;
  end;
finally
 if subKeyHnd1 <> 0 then RegCloseKey(subKeyHnd1);
end;
end;
{$ENDIF Windows}

procedure DeleteOptions;
begin
if FileExistsUTF8(FIDO_Descriptor_FileName) then
 DeleteFileUTF8(FIDO_Descriptor_FileName);
if OPath = '' then exit;
if FileExistsUTF8(OPath+fn) then
 DeleteFileUTF8(OPath+fn);
end;

procedure SaveDefaultDir3;
begin
OptionsInit(True);
try
 SaveDefaultDir2;
finally
 OptionsDone;
end;
end;

end.

